/*
 * File:        $RCSfile$
 * Package:     SAMRAI application
 * Copyright:   (c) 1997-2005 The Regents of the University of California
 * Revision:    $Revision: 173 $
 * Modified:    $Date: 2005-01-19 09:09:04 -0800 (Wed, 19 Jan 2005) $
 * Description: Numerical routines for example FAC Poisson solver
 */

#ifndef included_FACPoisson
#define included_FACPoisson


#include "CartesianVizamraiDataWriter.h"
#include "CellPoissonFACSolver.h"
#include "CellVariable.h"
#include "tbox/Database.h"
#include "Box.h"
#include "LocationIndexRobinBcCoefs.h"
#include "Patch.h"
#include "PatchHierarchy.h"
#include "PatchLevel.h"
#include "tbox/Pointer.h"
#include "SideVariable.h"
#include "StandardTagAndInitStrategy.h"
#include "VariableContext.h"
#include "VisDerivedDataStrategy.h"
#include "VisItDataWriter.h"


namespace SAMRAI {

/*!
 * @brief Class to solve a sample Poisson equation on a SAMR grid.
 *
 * This class demonstrates how use the FAC Poisson solver
 * class to solve Poisson's equation on a single level
 * within a hierarchy.
 *
 * We set up and solve the following problem:
 *
 *   2d: div(grad(u)) = -2 (pi^2) sin(pi x) sin(pi y)
 *
 *   3d: div(grad(u)) = -3 (pi^2) sin(pi x) sin(pi y) sin(pi z)
 *
 * which has the exact solution
 *
 *   2d: u = sin(pi x) sin(pi y)
 *
 *   3d: u = sin(pi x) sin(pi y) sin(pi z)
 *
 * This class inherits and implements virtual functions from
 * - mesh::StandardTagAndInitStrategy<NDIM> to initialize data
 *   on the SAMR grid.
 * - appu::VisDerivedDataStrategy<NDIM> to write out certain data
 *   in a vis file, such as the error of the solution.
 *
 * Inputs:  The only input parameter for this class is
 * "fac_poisson", the input database for the solv::CellPoissonFACSolver<NDIM>
 * object.  See the documentation for solv::CellPoissonFACSolver
 * for its input parameters.
 */
class FACPoisson :
   public mesh::StandardTagAndInitStrategy<NDIM> ,
   public appu::VisDerivedDataStrategy<NDIM>
{

public:

   /*!
    * @brief Constructor.
    *
    * If you want standard output and logging,
    * pass in valid pointers for those streams.
    *
    * @param object_name Ojbect name
    * @param database Input database (may be NULL)
    */
   FACPoisson( const string &object_name,
               tbox::Pointer<tbox::Database> database=NULL);

   virtual ~FACPoisson();



   //@{ @name mesh::StandardTagAndInitStrategy<NDIM> virtuals


   /*!
    * @brief Allocate and initialize data for a new level
    * in the patch hierarchy.
    *
    * This is where you implement the code for initialize data on
    * the grid.  All the information needed to initialize the grid
    * are in the arguments.
    *
    * @see mesh::StandardTagAndInitStrategy<NDIM>::initializeLevelData()
    */
   virtual void initializeLevelData (
      const tbox::Pointer<hier::BasePatchHierarchy<NDIM> > hierarchy ,
      const int level_number ,
      const double init_data_time ,
      const bool can_be_refined ,
      const bool initial_time ,
      const tbox::Pointer<hier::BasePatchLevel<NDIM> >
         old_level=tbox::Pointer<hier::PatchLevel<NDIM> >((0)) ,
      const bool allocate_data = true );

   /*!
    * @brief Reset any internal hierarchy-dependent information.
    */
   virtual void resetHierarchyConfiguration (
      tbox::Pointer<hier::BasePatchHierarchy<NDIM> > new_hierarchy ,
      int coarsest_level ,
      int finest_level );

   //@}


   //@{ @name appu::VisDerivedDataStrategy<NDIM> virtuals

   virtual bool packDerivedDataIntoDoubleBuffer(
      double* buffer ,
      const hier::Patch<NDIM> &patch ,
      const hier::Box<NDIM> &region ,
      const string &variable_name ,
      int depth_id );

   //@}



   /*!
    * @brief Solve using HYPRE Poisson solver
    *
    * Set up the linear algebra problem and use a
    * solv::CellPoissonFACSolver<NDIM> object to solve it.
    * -# Set initial guess
    * -# Set boundary conditions
    * -# Specify Poisson equation parameters
    * -# Call solver
    */
   int solvePoisson();

   /*!
    * @brief Set up external plotter to plot internal
    * data from this class.
    *
    * After calling this function, the external
    * data writer may be used to write the
    * viz file for this object.
    *
    * The internal hierarchy is used and must be
    * established before calling this function.
    * (This is commonly done by building a hierarchy
    * with the mesh::StandardTagAndInitStrategy<NDIM> virtual
    * functions implemented by this class.)
    *
    * @param viz_writer Vizramrai writer
    */
   int setupPlotter(
      appu::CartesianVizamraiDataWriter<NDIM> &plotter ) const;

   /*!
    * @brief Set up external plotter to plot internal
    * data from this class.
    *
    * After calling this function, the external
    * data writer may be used to write the
    * viz file for this object.
    *
    * The internal hierarchy is used and must be
    * established before calling this function.
    * (This is commonly done by building a hierarchy
    * with the mesh::StandardTagAndInitStrategy<NDIM> virtual
    * functions implemented by this class.)
    *
    * @param viz_writer VisIt writer
    */
   int setupPlotter(
      appu::VisItDataWriter<NDIM> &plotter ) const;



private:

  string d_object_name;

  tbox::Pointer<hier::PatchHierarchy<NDIM> > d_hierarchy;

  //@{
  /*!
   * @name Major algorithm objects.
   */

  /*!
   * @brief FAC poisson solver.
   */
  solv::CellPoissonFACSolver<NDIM> d_poisson_fac_solver;

  /*!
   * @brief Boundary condition coefficient implementation.
   */
  solv::LocationIndexRobinBcCoefs<NDIM> d_bc_coefs;

  //@}

  //@{

  /*!
   * @name Private state variables for solution.
   */

  /*!
   * @brief Context owned by this object.
   */
  tbox::Pointer<hier::VariableContext> d_context;

  /*!
   * @brief Descriptor indices of internal data.
   *
   * These are initialized in the constructor and never change.
   */
  int d_comp_soln_id, d_exact_id, d_rhs_id;

  //@}

};


}

#endif	// included_FACPoisson
