//
// File:        $RCSfile$
// Package:     SAMRAI application
// Copyright:   (c) 1997-2005 The Regents of the University of California
// Revision:    $Revision: 173 $
// Modified:    $Date: 2005-01-19 09:09:04 -0800 (Wed, 19 Jan 2005) $
// Description: Class containing numerical routines for modified Bratu problem
//

#include "ModifiedBratuProblem.h"
#include "ModifiedBratuFort.h"

#if defined(HAVE_PETSC) && defined(HAVE_KINSOL) && defined(HAVE_HYPRE)

#include <iostream>
#include <iomanip>
#include <fstream>

#ifndef LACKS_SSTREAM
#ifndef included_sstream
#define included_sstream
#include <sstream>
#endif
#else
#ifndef included_strstream
#define included_strstream
#include <strstream>
#endif
#endif

using namespace SAMRAI;

#include <stdlib.h>
#include <stdio.h>
#include <math.h>

#include <float.h>

#include "tbox/Array.h"
#include "BoundaryBox.h"
#include "BoxArray.h"
#include "CartesianPatchGeometry.h"
#include "CellData.h"
#include "CellIterator.h"
#include "CoarsenOperator.h"
#include "FaceData.h"
#include "HierarchyCellDataOpsReal.h"
#include "Index.h"
#include "OutersideData.h"
#include "PatchCellDataOpsReal.h"
#include "RefineSchedule.h"
#include "tbox/RestartManager.h"
#include "SideData.h"
#include "SideIndex.h"
#include "tbox/Timer.h"
#include "tbox/TimerManager.h"
#include "tbox/Utilities.h"
#include "tbox/IEEE.h"
#include "VariableDatabase.h"

#include "PVodeTrio_SAMRAIVector.h"
#include "PETSc_SAMRAIVectorReal.h"


#define MODIFIED_BRATU_PROBLEM (1)

#ifdef DEBUG_CHECK_ASSERTIONS
#ifndef included_assert
#define included_assert
#include <assert.h>
#endif
#endif

// Define for number of ghost cells on solution quantity
#define NUM_GHOSTS_U       (1)

/*
*************************************************************************
*                                                                       *
* Constructor and destructor for Modified Bratu Problem class.          *
*                                                                       *
* The problem we solve is:                                              *
*                                                                       *
*     du/dt = div( D(x,t)*grad(u) ) + lambda * exp(u) + f(u,x,t).       *
*                                                                       *
* The unknown u is a cell-centered variable.  The time discretization   *
* uses the backward Euler strategy.  We use a standard 7-point stencil  *
* (based on the finite volume meethod) for the div(D*grad(u)) term.     *
*                                                                       *
* The variables used here to manage data for the discrete problem are:  *
*                                                                       *
*    solution ........... unknown quantity "u"                          *
*    source_term ........ source term "f"                               *
*    exponential_term ... product "lambda * exp(u)"                     *
*    diffusion_coef ..... diffusion coefficient "D"                     *
*                                                                       *
* Other quantities used in the solution process:                        *
*                                                                       *
*    weight.............. weights for solution vector entries on grid   *
*    flux................ side-centered fluxes "D * grad(u)"            *
*    coarse_fine_flux.... fluxes at coarse-fine interfaces              *
*    jacobian_a.......... Jacobian entries for FAC solver               *
*    jacobian_b.......... Jacobian entries for FAC solver               *
*                                                                       *
* The constructor creates these variables to represent the solution     *
* and other quantities on the patch hierarchy.  The solution quantity   *
* is managed using three variable contexts, "CURRENT", "NEW", and       *
* "SCRATCH".  CURRENT and NEW represent the current and new values of   *
* the solution, respectively.  They have no ghost cells.  The SCRATCH   *
* context is where most of the computations involving the spatial       *
* discretization stencil occur.  This storage has a ghost cell width    *
* of one.  All other quantities are managed using only the SCRATCH      *
* context.  However, they all have zero ghost cell widths.              *
*                                                                       *
*************************************************************************
*/

ModifiedBratuProblem::ModifiedBratuProblem(
   const string& object_name, 
   tbox::Pointer<tbox::Database> input_db, 
   tbox::Pointer<geom::CartesianGridGeometry<NDIM> > grid_geometry,
   tbox::Pointer<appu::CartesianVizamraiDataWriter<NDIM> > viz_writer,
   tbox::Pointer<appu::VisItDataWriter<NDIM> > visit_writer)
   : d_jacobian_a_id(-1),
     d_jacobian_b_id(-1),
     d_precond_a_id(-1),
     d_precond_b_id(-1)
{
#ifdef DEBUG_CHECK_ASSERTIONS
   assert(!object_name.empty());
   assert(!input_db.isNull());
   assert(!grid_geometry.isNull());
#endif

   d_object_name = object_name;
   tbox::RestartManager::getManager()->registerRestartItem(d_object_name, this);

   d_grid_geometry = grid_geometry;

   /*
    * Set default values for data members.
    */
   d_current_time =
   d_current_dt =
   d_new_time =
   d_lambda =
   d_input_dt =
   tbox::IEEE::getSignalingNaN();

   d_max_precond_its = tbox::IEEE::getINT_MAX();
   d_precond_tol = tbox::IEEE::getSignalingNaN();

   getFromInput( input_db,
                 false );

   d_FAC_solver = (solv::CellPoissonFACSolver<NDIM>*)NULL;

   hier::VariableDatabase<NDIM>* variable_db = hier::VariableDatabase<NDIM>::getDatabase();

   d_current = variable_db->getContext("CURRENT");
   d_new     = variable_db->getContext("NEW");
   d_scratch = variable_db->getContext("SCRATCH");

   /*
    * Create variables used in the discrete problem.  Register variables
    * and contexts with variable database to retrieve their descriptor
    * indices for data management.
    */

   d_nghosts = hier::IntVector<NDIM>(NUM_GHOSTS_U);

   d_solution = new pdat::CellVariable<NDIM,double>(d_object_name+"solution",1);

   int soln_id = variable_db->registerVariableAndContext(d_solution,
                                                         d_current,
                                                         hier::IntVector<NDIM>(0));
   if ( !viz_writer.isNull() ) {
      viz_writer->registerPlotScalar("U", soln_id);
   }
   if ( !visit_writer.isNull() ) {
      visit_writer->registerPlotQuantity("U", "SCALAR", soln_id);
   }
   int soln_new_id = variable_db->registerVariableAndContext(d_solution,
                                                             d_new,
                                                             hier::IntVector<NDIM>(0));
   d_soln_scratch_id = variable_db->registerVariableAndContext(d_solution, 
                                                               d_scratch, 
                                                               d_nghosts);

   /*
    *hier::Variable<NDIM> to weight solution vector entries on a composite grid.
    */

   d_weight = new pdat::CellVariable<NDIM,double>(d_object_name+"weight",1);

   d_weight_id = variable_db->registerVariableAndContext(d_weight,
                                                         d_scratch,
                                                         hier::IntVector<NDIM>(0));

   /*
    * Other variables used in discrete problem.
    */

   d_source_term = new pdat::CellVariable<NDIM,double>(d_object_name+"source_term",1);

   int source_id = variable_db->registerVariableAndContext(d_source_term,
                                                           d_scratch,
                                                           hier::IntVector<NDIM>(0)); 

   d_exponential_term =
      new pdat::CellVariable<NDIM,double>(d_object_name+"exponential_term",1);

   int exp_id = variable_db->registerVariableAndContext(d_exponential_term, 
                                                        d_scratch,
                                                        hier::IntVector<NDIM>(0));

   d_diffusion_coef =
      new pdat::SideVariable<NDIM,double>(d_object_name+"diffusion_coef",1);

   int diffcoef_id = variable_db->registerVariableAndContext(d_diffusion_coef, 
                                                             d_scratch,
                                                             hier::IntVector<NDIM>(0));

   d_flux = new pdat::SideVariable<NDIM,double>(d_object_name+"flux",1);

   d_flux_id = variable_db->registerVariableAndContext(d_flux,
                                                       d_scratch,
                                                       hier::IntVector<NDIM>(0));

   d_coarse_fine_flux =
      new pdat::OutersideVariable<NDIM,double>(d_object_name+"coarse_fine_flux",1);

   d_coarse_fine_flux_id = 
      variable_db->registerVariableAndContext(d_coarse_fine_flux,
                                              d_scratch,
                                              hier::IntVector<NDIM>(0));

    /*
     * Variables for A(x)z=r preconditioning.
     */
    d_precond_a = new pdat::CellVariable<NDIM,double>(d_object_name+"precond_a",1);
    d_precond_a_id = variable_db->registerVariableAndContext(d_precond_a,
							     d_scratch,
							     hier::IntVector<NDIM>(0));
    d_precond_b = new pdat::FaceVariable<NDIM,double>(d_object_name+"precond_b",1);
    d_precond_b_id = variable_db->registerVariableAndContext(d_precond_b,
							     d_scratch,
							     hier::IntVector<NDIM>(0));

    /*
     * Variables for A(x)*v operations.
     */
   d_jacobian_a =
      new pdat::CellVariable<NDIM,double>(d_object_name+":jacobian_a",1);
   d_jacobian_a_id = variable_db->registerVariableAndContext(d_jacobian_a, 
							     d_scratch,
							     hier::IntVector<NDIM>(0));
   d_jacobian_b =
      new pdat::FaceVariable<NDIM,double>(d_object_name+":jacobian_b",1);
   d_jacobian_b_id = variable_db->registerVariableAndContext(d_jacobian_b, 
							     d_scratch,
							     hier::IntVector<NDIM>(0));

   /*
    * ComponentSelectors are used to allocate and deallocate storage for
    * variables on the patch hierarchy in groups.  This reduces the 
    * complexity of tracking each individual data index.   Storage 
    * for the time evolving solution is managed collectively and
    * auxiliary quantities used in nonlinear function evaluation.  These
    * latter quantities could be placed in a separate hier::ComponentSelector
    * and allocated only when needed, but we chose not to add that
    * for this simple problem.  Data needed to correctly compute
    * fluxes at coarse/fine interfaces is also managed collectively
    * and is only allocated when needed.  Finally data to form the 
    * Jacobian is also grouped together.
    */

   d_new_patch_problem_data.setFlag(soln_id);

   d_problem_data.setFlag(soln_new_id);
   d_problem_data.setFlag(d_soln_scratch_id);

   d_problem_data.setFlag(d_weight_id);

   d_problem_data.setFlag(source_id);
   d_problem_data.setFlag(exp_id);
   d_problem_data.setFlag(diffcoef_id);

   d_jacobian_data.setFlag(d_jacobian_a_id);
   d_jacobian_data.setFlag(d_jacobian_b_id);

   d_precond_data.setFlag(d_precond_a_id);
   d_precond_data.setFlag(d_precond_b_id);

   /*
    * Refine and coarsen algorithms used to solve the problem.
    *
    *    d_fill_new_level :
    *
    *       When new levels are created in the hierarchy (e.g., after 
    *       regridding), the "CURRENT" solution state is initialized in 
    *       new fine regions by using linear interpolation from coarser 
    *       levels.  
    *
    *    d_flux_coarsen:
    *
    *       Used to coarsen fluxes at coarse-fine interfaces during 
    *       nonlinear function evaluation.  
    */

   d_soln_refine_op = 
      grid_geometry->lookupRefineOperator(d_solution,
                                          "LINEAR_REFINE");

   d_fill_new_level.registerRefine(soln_id,
                                   soln_id,
                                   d_soln_scratch_id,
                                   d_soln_refine_op); 

   d_soln_coarsen_op =
      grid_geometry->lookupCoarsenOperator(d_solution,
                                          "CONSERVATIVE_COARSEN");

   tbox::Pointer<CoarsenOperator<NDIM> > flux_coarsen_op =
      grid_geometry->lookupCoarsenOperator(d_coarse_fine_flux,
                                           "CONSERVATIVE_COARSEN");

   /*
    * Register the transfer algorithms so we can create
    * schedules when we have info on the hierarchy (in
    * resetHierarchyConfiguration).
    */

   d_scratch_soln_coarsen.registerCoarsen(d_soln_scratch_id,
                                          d_soln_scratch_id,
                                          d_soln_coarsen_op);

   d_soln_coarsen.registerCoarsen(soln_id,
                                  soln_id,
                                  d_soln_coarsen_op);

   d_flux_coarsen.registerCoarsen(d_flux_id,
                                    d_coarse_fine_flux_id,
                                    flux_coarsen_op);

   d_soln_fill.registerRefine(d_soln_scratch_id, 
                              soln_new_id,
                              d_soln_scratch_id, 
                              d_soln_refine_op);

}

/*
*************************************************************************
*                                                                       *
* The destructor does nothing interesting.                              *
*                                                                       *
*************************************************************************
*/

ModifiedBratuProblem::~ModifiedBratuProblem()
{
   if (d_FAC_solver) {
      delete d_FAC_solver;
   }
}

/*
*************************************************************************
*                                                                       *
* The vector used in the nonlinear iteration to advance the data        *
* includes only the "NEW" solution values.  In other words, we          *
* know the "CURRENT" solution and we advance to the "NEW" solution.     *
*                                                                       *
*************************************************************************
*/

void ModifiedBratuProblem::setupSolutionVector(
   tbox::Pointer< solv::SAMRAIVectorReal<NDIM,double> > solution)
{
#ifdef DEBUG_CHECK_ASSERTIONS
   assert(!solution.isNull());
#endif

   d_solution_vector = solution;

   solution->addComponent(d_solution,
                          hier::VariableDatabase<NDIM>::getDatabase()->
                          mapVariableAndContextToIndex(d_solution, d_new),
                          d_weight_id);
}

/*
 * Vector weights are used to calculate inner products and norms.
 * In cells not covered by finer cells, the weights are set to the
 * the cell volume.  In those covered by finer cells, the weights 
 * are set to 0.  This latter assignment ensures that data in 
 * covered cells contribute to these sums only once. 
 */

void ModifiedBratuProblem::setVectorWeights(
   tbox::Pointer<hier::PatchHierarchy<NDIM> > hierarchy)
{
   for ( int amr_level = hierarchy->getFinestLevelNumber();
             amr_level >= 0;
             amr_level-- ) {

     /*
      * On every level, first assign cell volume to vector weight.
      */

      tbox::Pointer<hier::PatchLevel<NDIM> > level = hierarchy->getPatchLevel(amr_level);
      for (hier::PatchLevel<NDIM>::Iterator p(level); p; p++) {
         tbox::Pointer<hier::Patch<NDIM> > patch = level->getPatch(p());
         tbox::Pointer<geom::CartesianPatchGeometry<NDIM> > patch_geometry =
            patch->getPatchGeometry();
         const double* dx = patch_geometry->getDx();
         double cell_vol = dx[0];
#if (NDIM > 1)
         cell_vol *= dx[1];
#endif
#if (NDIM > 2)
         cell_vol *= dx[2];
#endif
         tbox::Pointer< pdat::CellData<NDIM,double> > w =
            patch->getPatchData(d_weight_id);
         w->fillAll(cell_vol); 
      }

      /*
       * On all but the finest level, assign 0 to vector
       * weight to cells covered by finer cells.
       */

      if (amr_level < hierarchy->getFinestLevelNumber()) {

        /*
	 * First get the boxes that describe index space of the next finer 
	 * level and coarsen them to describe corresponding index space 
	 * at this level.
	 */

         tbox::Pointer<hier::PatchLevel<NDIM> > next_finer_level = 
            hierarchy->getPatchLevel(amr_level+1);
         hier::BoxArray<NDIM> coarsened_boxes = next_finer_level->getBoxes(); 
         hier::IntVector<NDIM> coarsen_ratio = next_finer_level->getRatio();
         coarsen_ratio /= level->getRatio();
         coarsened_boxes.coarsen(coarsen_ratio);

         /*
	  * Then set vector weight to 0 wherever there is
	  * a nonempty intersection with the next finer level.
	  * Note that all assignments are local.
	  */

         for (hier::PatchLevel<NDIM>::Iterator p(level); p; p++) {

            tbox::Pointer<hier::Patch<NDIM> > patch = level->getPatch(p());
            for ( int i = 0; i < coarsened_boxes.getNumberOfBoxes(); i++ ) {

               hier::Box<NDIM> coarse_box = coarsened_boxes(i);
               hier::Box<NDIM> intersection = coarse_box*patch->getBox();
               if ( !intersection.empty() ) {
                  tbox::Pointer< pdat::CellData<NDIM,double> > w =
                     patch->getPatchData(d_weight_id);
                  w->fillAll(0.0, intersection);

               }  // assignment only in non-empty intersection
            }  // loop over coarsened boxes from finer level
         }  // loop over patches in level
      }  // all levels except finest
   }  // loop over levels
}

/*
*************************************************************************
*                                                                       *
* Set the initial guess for the nonlinear iteration.  If we are at the  *
* first step on the current hierarchy configuration, we initialize the  *
* the new solution with the current soltuion values.  Note that we do   *
* the same if we are not at the first step on the current hierarchy     *
* configuration.  In the future, we may employ a more sophisticated     *
* strategy involving some sort of extrapolation.                        *
*                                                                       *
*************************************************************************
*/

void ModifiedBratuProblem::setInitialGuess(
   const bool first_step, 
   const double current_time,
   const double current_dt,
   const double old_dt) 
{

   d_current_time = current_time;
   d_current_dt   = current_dt;
   d_new_time     = d_current_time + d_current_dt;

   tbox::Pointer<hier::PatchHierarchy<NDIM> > hierarchy = d_solution_vector->getPatchHierarchy();

   if (first_step) {

      if (d_FAC_solver) {
         delete d_FAC_solver;
      }

      d_FAC_solver =
         new solv::CellPoissonFACSolver<NDIM>(d_object_name+":FAC_solver");

   }

   for (int ln = hierarchy->getFinestLevelNumber(); ln >= 0; ln-- ) {
      hierarchy->getPatchLevel(ln)->
         allocatePatchData(d_jacobian_data);
      hierarchy->getPatchLevel(ln)->
         allocatePatchData(d_precond_data);
   }

   if ( first_step ) {

      for (int amr_level = 0;
           amr_level < hierarchy->getNumberLevels();
           amr_level++ ) {

         tbox::Pointer<hier::PatchLevel<NDIM> > patch_level = hierarchy->getPatchLevel(amr_level);
         for (hier::PatchLevel<NDIM>::Iterator p(patch_level); p; p++) {
            tbox::Pointer<hier::Patch<NDIM> > patch = patch_level->getPatch(p());

            tbox::Pointer< pdat::CellData<NDIM,double> > y_cur = 
               patch->getPatchData(d_solution, d_current);
            tbox::Pointer< pdat::CellData<NDIM,double> > y_new = 
               patch->getPatchData(d_solution, d_new);
            y_new->copy(*y_cur);

            y_new->setTime(d_new_time);

            patch->getPatchData(d_solution, d_scratch)->setTime(d_new_time);

            tbox::Pointer<geom::CartesianPatchGeometry<NDIM> > patch_geometry =
               patch->getPatchGeometry();
            const double* dx   = patch_geometry->getDx();
            const double* xlo  = patch_geometry->getXLower();
            const double* xhi  = patch_geometry->getXUpper();
            const hier::Index<NDIM> ifirst = patch->getBox().lower();
            const hier::Index<NDIM> ilast  = patch->getBox().upper();  

            tbox::Pointer< pdat::SideData<NDIM,double> > diffusion =
               patch->getPatchData(d_diffusion_coef, d_scratch);

#if (NDIM == 1)
            FORT_EVALDIFFUSION(ifirst(0), ilast(0),
                               dx, xlo, xhi,
                               diffusion->getPointer(0));
#endif
#if (NDIM == 2)
            FORT_EVALDIFFUSION(ifirst(0), ilast(0),
                               ifirst(1), ilast(1),
                               dx, xlo, xhi,
                               diffusion->getPointer(0),
                               diffusion->getPointer(1));
#endif
#if (NDIM == 3)
            FORT_EVALDIFFUSION(ifirst(0), ilast(0),
                               ifirst(1), ilast(1),
                               ifirst(2), ilast(2),
                               dx, xlo, xhi,
                               diffusion->getPointer(0),
                               diffusion->getPointer(1),
                               diffusion->getPointer(2));
#endif
         }
      }     
   } else {

      for (int amr_level = 0;
           amr_level < hierarchy->getNumberLevels();
           amr_level++ ) {

         tbox::Pointer<hier::PatchLevel<NDIM> > patch_level = hierarchy->getPatchLevel(amr_level);
         for (hier::PatchLevel<NDIM>::Iterator p(patch_level); p; p++) {
            tbox::Pointer<hier::Patch<NDIM> > patch = patch_level->getPatch(p());

            tbox::Pointer< pdat::CellData<NDIM,double> > y_cur = 
               patch->getPatchData(d_solution, d_current);
            tbox::Pointer< pdat::CellData<NDIM,double> > y_new = 
               patch->getPatchData(d_solution, d_new);
            y_new->copy(*y_cur);

            y_new->setTime(d_new_time);

            patch->getPatchData(d_solution, d_scratch)->setTime(d_new_time);

            tbox::Pointer<geom::CartesianPatchGeometry<NDIM> > patch_geometry =
               patch->getPatchGeometry();
            const double* dx   = patch_geometry->getDx();
            const double* xlo  = patch_geometry->getXLower();
            const double* xhi  = patch_geometry->getXUpper();
            const hier::Index<NDIM> ifirst = patch->getBox().lower();
            const hier::Index<NDIM> ilast  = patch->getBox().upper();

            tbox::Pointer< pdat::SideData<NDIM,double> > diffusion =
               patch->getPatchData(d_diffusion_coef, d_scratch);
#if (NDIM == 1)
            FORT_EVALDIFFUSION(ifirst(0), ilast(0),
                               dx, xlo, xhi,
                               diffusion->getPointer(0));
#endif
#if (NDIM == 2)
            FORT_EVALDIFFUSION(ifirst(0), ilast(0),
                               ifirst(1), ilast(1),
                               dx, xlo, xhi,
                               diffusion->getPointer(0),
                               diffusion->getPointer(1));
#endif
#if (NDIM == 3)
            FORT_EVALDIFFUSION(ifirst(0), ilast(0),
                               ifirst(1), ilast(1),
                               ifirst(2), ilast(2),
                               dx, xlo, xhi,
                               diffusion->getPointer(0),
                               diffusion->getPointer(1),
                               diffusion->getPointer(2));
#endif
         }
      }
   }
}

/*
*************************************************************************
*                                                                       *
* Return the time increment used for the first solution advance step.   *
*                                                                       *
*************************************************************************
*/

double ModifiedBratuProblem::getInitialDt()
{
   return(d_input_dt);
}

/*
*************************************************************************
*                                                                       *
* Compute and return the next time increment through which to advance   *
* the solution.  Note that good_solution is the value returned by the   *
* preceeding call to checkNewSolution().  The integer solver_retcode    *
* is the return code generated by the nonlinear solver.                 *
*                                                                       *
* In the future, we will have a more intelligent strategy.  When        *
* good_solution is true, we would like to take the largest timestep     *
* possible.   When good_solution is false, we will cut the timestep     *
* depending on the solver return code.  In this case, if the code       *
* indicates that that nonlinear iteration converged, we will cut the    *
* timestep by some amount since the temporal error is probably too      *
* large.  For a first order method like BE, determining a suitable      *
* timestep can be based on an estimate of the second time derivative    *
* of the solution.  See somments before checkNewSolution() below.       *
* If the nonlinear iteration did not converge, we will have to try      *
* another approach.                                                     *
*                                                                       *
* At this time, we always return a constant time step read from input.  *
*                                                                       *
*************************************************************************
*/

double ModifiedBratuProblem::getNextDt(
   const bool good_solution,
   const int solver_retcode)
{
   (void) good_solution;
   (void) solver_retcode;

   return(d_input_dt);
}

/*
*************************************************************************
*                                                                       *
* Check the computed solution and return a boolean value of true if it  *
* is acceptable; otherwise return false.  The integer solver_retcode    *
* is the return code generated by the nonlinear solver.  This value     *
* must be interpreted in a manner consistant with the solver in use.    *
*                                                                       *
* Ordinarily we would estimate the temporal error in the solution       *
* before accepting it.  For backward Euler, this is straightforward.    *
* For example, if we had both (d_solution, d_current) and (d_solution,  *
* d_old), we could evaluate dy/dt at these two timesteps by evaluating  *
* the rhs at these values.  Then the second derivative in the solution  *
* can be estimated by using finite differences and compared to some     *
* prescribed tolerance.  Other heuristics may be possible.              *
*                                                                       *
* For now, we accept every solution we compute.                         *
*                                                                       *
*************************************************************************
*/

bool ModifiedBratuProblem::checkNewSolution(const int solver_retcode)
{

   double new_time = d_current_time + d_current_dt;
   double maxerror = 0.0;
   tbox::Pointer<hier::PatchHierarchy<NDIM> > hierarchy = d_solution_vector->getPatchHierarchy();
   for (int amr_level = 0;
        amr_level < hierarchy->getNumberLevels();
        amr_level++ ) {
      tbox::Pointer<hier::PatchLevel<NDIM> > patch_level = hierarchy->getPatchLevel(amr_level);
      double levelerror = 0.0;
      double levell2error = 0.0;
      for (hier::PatchLevel<NDIM>::Iterator p(patch_level); p; p++) {
         tbox::Pointer<hier::Patch<NDIM> > patch = patch_level->getPatch(p());
         const hier::Index<NDIM> ifirst = patch->getBox().lower();
         const hier::Index<NDIM> ilast = patch->getBox().upper();
         tbox::Pointer<geom::CartesianPatchGeometry<NDIM> > patch_geom = patch->getPatchGeometry();
         const double* dx = patch_geom->getDx();
         const double* xlo = patch_geom->getXLower();
         const double* xhi = patch_geom->getXUpper();
         tbox::Pointer< pdat::CellData<NDIM,double> > u = patch->getPatchData(d_solution, d_new);
         tbox::Pointer< pdat::CellData<NDIM,double> > w = patch->getPatchData(d_weight_id);

#if (NDIM == 1)
         FORT_ERROR( ifirst(0), ilast(0),
                     u->getPointer(), w->getPointer(),
                     d_lambda,
                     xlo, xhi, dx,
                     new_time,
                     levelerror, levell2error );
#endif
#if (NDIM == 2)
         FORT_ERROR( ifirst(0), ilast(0),
                     ifirst(1), ilast(1),
                     u->getPointer(), w->getPointer(),
                     d_lambda,
                     xlo, xhi, dx,
                     new_time,
                     levelerror, levell2error );
#endif
#if (NDIM == 3)
         FORT_ERROR( ifirst(0), ilast(0),
                     ifirst(1), ilast(1),
                     ifirst(2), ilast(2),
                     u->getPointer(), w->getPointer(),
                     d_lambda,
                     xlo, xhi, dx,
                     new_time,
                     levelerror, levell2error );
#endif
      }
      maxerror = (levelerror > maxerror) ? levelerror : maxerror;
      tbox::pout << " At " << new_time 
           << " on level " << amr_level 
           << " max err is " << levelerror
           << " wtd l2 err is " << levell2error
           << endl;
   }

   tbox::pout << " At " << new_time << " err is " << maxerror << endl;

   return(true);
}

/*
*************************************************************************
*                                                                       *
* Update soltution quantities after computing an acceptable time        *
* advanced solution.   The new_time value is the new solution time.     *
*                                                                       *
* Since we have no dependent variables in this problem and our memory   *
* management is extremely simple, we just copy new into current.        *
* Then, we synchronize the times for each set of solution vlaues.       *
*                                                                       *
*************************************************************************
*/

void ModifiedBratuProblem::updateSolution(const double new_time)
{
   tbox::Pointer<hier::PatchHierarchy<NDIM> > hierarchy = d_solution_vector->getPatchHierarchy();

   d_new_time = d_current_time = new_time;

   for (int amr_level = 0;
        amr_level < hierarchy->getNumberLevels();
        amr_level++ ) {

      tbox::Pointer<hier::PatchLevel<NDIM> > patch_level = hierarchy->getPatchLevel(amr_level);
      for (hier::PatchLevel<NDIM>::Iterator p(patch_level); p; p++) {
         tbox::Pointer<hier::Patch<NDIM> > patch = patch_level->getPatch(p());

         tbox::Pointer< pdat::CellData<NDIM,double> > y_cur = 
            patch->getPatchData(d_solution, d_current);
         tbox::Pointer< pdat::CellData<NDIM,double> > y_new = 
            patch->getPatchData(d_solution, d_new);
         y_cur->copy(*y_new);

         y_cur->setTime(new_time);

         patch->getPatchData(d_solution, d_scratch)->setTime(new_time);

      }

      hierarchy->getPatchLevel(amr_level)->
         deallocatePatchData(d_jacobian_data);
      hierarchy->getPatchLevel(amr_level)->
         deallocatePatchData(d_precond_data);

   }

}

/*
*************************************************************************
*                                                                       *
* Allocate and initialize data for a new level in the patch hierarchy.  *
*                                                                       *
* At initial time only, we initialize the solution to zero.  At all     *
* other times the solution is either copied from an old level or        *
* interpolated from coarser levels.  The cell weights and diffusion     *
* coefficients are always set here since they live as long as a         *
* patch and are not time dependent.                                     *
*                                                                       *
*************************************************************************
*/

void ModifiedBratuProblem::initializeLevelData(
   const tbox::Pointer<hier::BasePatchHierarchy<NDIM> > hierarchy, 
   const int level_number,
   const double time,
   const bool can_be_refined,
   const bool initial_time,
   const tbox::Pointer<hier::BasePatchLevel<NDIM> > old_level,
   const bool allocate_data )
{
#ifdef DEBUG_CHECK_ASSERTIONS
   assert(!hierarchy.isNull());
   assert( (level_number >= 0)
           && (level_number <= hierarchy->getFinestLevelNumber()) );
   if ( !(old_level.isNull()) ) {
      assert( level_number == old_level->getLevelNumber() );
   }
   assert(!(hierarchy->getPatchLevel(level_number)).isNull());
#endif

   tbox::Pointer<hier::PatchLevel<NDIM> > level = hierarchy->getPatchLevel(level_number);

   level->allocatePatchData(d_new_patch_problem_data, time);

   if ((level_number > 0) || !old_level.isNull()) {
      d_fill_new_level.createSchedule(level, 
                                      old_level, 
                                      level_number-1, 
                                      hierarchy, 
                                      this)->fillData(time);

      if ( !old_level.isNull() ) {
         old_level->deallocatePatchData(d_new_patch_problem_data);
         old_level->deallocatePatchData(d_problem_data);
      }
   }

   level->allocatePatchData(d_problem_data);

   /*
    * Initialize data on patch interiors.  At initial time only, we 
    * initialize solution to zero.  At all other times the solution
    * is either copied from an old level, or interpolated from some
    * coarser level using the "d_fill_new_level" algorithm above.  
    */

   for (hier::PatchLevel<NDIM>::Iterator p(level); p; p++) {
      tbox::Pointer<hier::Patch<NDIM> > patch = level->getPatch(p());

      if ( initial_time ) {

          tbox::Pointer< pdat::CellData<NDIM,double> > u =
             patch->getPatchData(d_solution, d_current); 
          u->fillAll(0.0);

      }

      tbox::Pointer<geom::CartesianPatchGeometry<NDIM> > patch_geometry =
         patch->getPatchGeometry();
      const double* dx   = patch_geometry->getDx();
      const double* xlo  = patch_geometry->getXLower();
      const double* xhi  = patch_geometry->getXUpper();
      const hier::Index<NDIM> ifirst = patch->getBox().lower();
      const hier::Index<NDIM> ilast  = patch->getBox().upper();

      tbox::Pointer< pdat::SideData<NDIM,double> > diffusion =
         patch->getPatchData(d_diffusion_coef, d_scratch);
#if (NDIM == 1)
      FORT_EVALDIFFUSION(ifirst(0), ilast(0),
                               dx, xlo, xhi,
                               diffusion->getPointer(0));
#endif
#if (NDIM == 2)
      FORT_EVALDIFFUSION(ifirst(0), ilast(0),
                         ifirst(1), ilast(1),
                         dx, xlo, xhi,
                         diffusion->getPointer(0),
                         diffusion->getPointer(1));
#endif
#if (NDIM == 3)
      FORT_EVALDIFFUSION(ifirst(0), ilast(0),
                         ifirst(1), ilast(1),
                         ifirst(2), ilast(2),
                         dx, xlo, xhi,
                         diffusion->getPointer(0),
                         diffusion->getPointer(1),
                         diffusion->getPointer(2));
#endif
   }
}

/*
*************************************************************************
*                                                                       *
* After hierarchy levels whose numbers lie in the given range have      *
* been either added or changed in the hierarchy, this routine is        *
* called.  Typically, it is used to reset any data members that         *
* remain constant until the levels change again via adaptive gridding   *
* (e.g., communication schedules), thus allowing some efficiencies to   *
* be achieved.  Here, we do nothing, since the function evalutation     *
* and Jacobian-vector product routines have indeterminate data ids.     *
*                                                                       *
*************************************************************************
*/

void ModifiedBratuProblem::resetHierarchyConfiguration(
   const tbox::Pointer<hier::BasePatchHierarchy<NDIM> > hierarchy, 
   const int coarsest_level, 
   const int finest_level)
{
#ifdef DEBUG_CHECK_ASSERTIONS
   assert(!hierarchy.isNull());
   assert( (coarsest_level >= 0)
           && (coarsest_level <= finest_level)
           && (finest_level <= hierarchy->getFinestLevelNumber()) );
   for (int ln0 = 0; ln0 <= finest_level; ln0++) {
      assert(!(hierarchy->getPatchLevel(ln0)).isNull());
   }
#endif
   tbox::Pointer< hier::PatchHierarchy<NDIM> > patch_hierarchy = hierarchy;  
   (void) coarsest_level;  
   (void) finest_level;  
   d_flux_coarsen_schedule.resizeArray(patch_hierarchy->getNumberLevels());
   d_soln_fill_schedule.resizeArray(patch_hierarchy->getNumberLevels());
   d_soln_coarsen_schedule.resizeArray(patch_hierarchy->getNumberLevels());
   d_scratch_soln_coarsen_schedule.resizeArray(
      patch_hierarchy->getNumberLevels());
   int ln;
   /*
    * Rebuild schedules affected by the hierarchy change.
    * In the following loops, ln is the level number of
    * the destination level affected by the change.
    */
   {
      const int ln_beg = coarsest_level - (coarsest_level>0);
      const int ln_end = finest_level;
      for ( ln=ln_beg; ln<ln_end; ++ln ) {
         tbox::Pointer<hier::PatchLevel<NDIM> > level =
            patch_hierarchy->getPatchLevel(ln);
         tbox::Pointer<hier::PatchLevel<NDIM> > finer_level =
            patch_hierarchy->getPatchLevel(ln+1);
         d_flux_coarsen_schedule[ln] =
            d_flux_coarsen.createSchedule(level, finer_level);
      }
   }
   {
      const int ln_beg = coarsest_level;
      const int ln_end = finest_level;
      for ( ln=ln_beg; ln<=ln_end; ++ln ) {
         tbox::Pointer<hier::PatchLevel<NDIM> > level =
            patch_hierarchy->getPatchLevel(ln);
         d_soln_fill_schedule[ln] =
            d_soln_fill.createSchedule(level,
                                       ln-1,
                                       patch_hierarchy,
                                       (RefinePatchStrategy<NDIM> *)this);
      }
   }
   {
      const int ln_beg = coarsest_level - (coarsest_level>0);
      const int ln_end = finest_level;
      for ( ln=ln_beg; ln<ln_end; ++ln ) {
         tbox::Pointer<hier::PatchLevel<NDIM> > level =
            patch_hierarchy->getPatchLevel(ln);
         tbox::Pointer<hier::PatchLevel<NDIM> > finer_level =
            patch_hierarchy->getPatchLevel(ln+1);
         d_soln_coarsen_schedule[ln] =
            d_soln_coarsen.createSchedule(level, finer_level);
         d_scratch_soln_coarsen_schedule[ln] =
            d_scratch_soln_coarsen.createSchedule(level, finer_level);
      }
   }
   {
      const int ln_beg = coarsest_level - (coarsest_level>0);
      const int ln_end = finest_level;
      for ( ln=ln_beg; ln<ln_end; ++ln ) {
         tbox::Pointer<hier::PatchLevel<NDIM> > level =
            patch_hierarchy->getPatchLevel(ln);
         tbox::Pointer<hier::PatchLevel<NDIM> > finer_level =
            patch_hierarchy->getPatchLevel(ln+1);
      }
   }
}

/*
*************************************************************************
*                                                                       *
* KINSOL interface for user-supplied routines.                          *
*                                                                       *
*************************************************************************
*/

void ModifiedBratuProblem::evaluateNonlinearFunction(
   solv::PVodeTrioAbstractVector* soln,
   solv::PVodeTrioAbstractVector* fval)
{
#ifdef DEBUG_CHECK_ASSERTIONS
   assert(!(soln == (solv::PVodeTrioAbstractVector*)NULL));
   assert(!(fval == (solv::PVodeTrioAbstractVector*)NULL));
#endif

   tbox::Pointer< solv::SAMRAIVectorReal<NDIM,double> > x =
      solv::PVodeTrio_SAMRAIVector<NDIM>::getSAMRAIVector(soln);
   tbox::Pointer< solv::SAMRAIVectorReal<NDIM,double> > f =
      solv::PVodeTrio_SAMRAIVector<NDIM>::getSAMRAIVector(fval);

   evaluateBratuFunction(x, f);
}

int ModifiedBratuProblem::precondSetup(solv::PVodeTrioAbstractVector* soln,
                                       solv::PVodeTrioAbstractVector* soln_scale,
                                       solv::PVodeTrioAbstractVector* fval,
                                       solv::PVodeTrioAbstractVector* fval_scale,
                                       solv::PVodeTrioAbstractVector* vtemp1,
                                       solv::PVodeTrioAbstractVector* vtemp2,
                                       double mach_roundoff,
                                       int& num_feval)
{
   (void) soln_scale;
   (void) fval;
   (void) fval_scale;
   (void) vtemp1;
   (void) vtemp2;
   (void) mach_roundoff;
#ifdef DEBUG_CHECK_ASSERTIONS
   assert(!(soln == (solv::PVodeTrioAbstractVector*)NULL));
#endif

   num_feval += 0;

   tbox::Pointer< solv::SAMRAIVectorReal<NDIM,double> > x =
      solv::PVodeTrio_SAMRAIVector<NDIM>::getSAMRAIVector(soln);

   setupBratuPreconditioner(x); 

   return(0);
}

int ModifiedBratuProblem::precondSolve(solv::PVodeTrioAbstractVector* soln,
                                       solv::PVodeTrioAbstractVector* soln_scale,
                                       solv::PVodeTrioAbstractVector* fval,
                                       solv::PVodeTrioAbstractVector* fval_scale,
                                       solv::PVodeTrioAbstractVector* rhs,
                                       solv::PVodeTrioAbstractVector* vtemp,
                                       double mach_roundoff,
                                       int& num_feval)
{
   (void) soln;
   (void) soln_scale;
   (void) fval;
   (void) fval_scale;
   (void) vtemp;
   (void) mach_roundoff;
#ifdef DEBUG_CHECK_ASSERTIONS
   assert(!(rhs == (solv::PVodeTrioAbstractVector*)NULL));
#endif

   num_feval += 0;

   tbox::Pointer< solv::SAMRAIVectorReal<NDIM,double> > r =
      solv::PVodeTrio_SAMRAIVector<NDIM>::getSAMRAIVector(rhs);

   return(applyBratuPreconditioner(r, r));
}

int
ModifiedBratuProblem::jacobianTimesVector(solv::PVodeTrioAbstractVector* vector,
                                          solv::PVodeTrioAbstractVector* product,
                                          const bool soln_changed,
                                          solv::PVodeTrioAbstractVector* soln)
{
#ifdef DEBUG_CHECK_ASSERTIONS
   assert(!(vector == (solv::PVodeTrioAbstractVector*)NULL));
   assert(!(product == (solv::PVodeTrioAbstractVector*)NULL));
   assert(!(soln == (solv::PVodeTrioAbstractVector*)NULL));
#endif

   tbox::Pointer< solv::SAMRAIVectorReal<NDIM,double> > v =
      solv::PVodeTrio_SAMRAIVector<NDIM>::getSAMRAIVector(vector);
   tbox::Pointer< solv::SAMRAIVectorReal<NDIM,double> > Jv =
      solv::PVodeTrio_SAMRAIVector<NDIM>::getSAMRAIVector(product);

   if ( soln_changed ) {
      tbox::Pointer< solv::SAMRAIVectorReal<NDIM,double> > ucur =
	 solv::PVodeTrio_SAMRAIVector<NDIM>::getSAMRAIVector(soln);
      evaluateBratuJacobian(ucur);
   }
   return(jacobianTimesVector(v, Jv));
}

/*
*************************************************************************
*                                                                       *
* PETSc/SNES interface for user-supplied routines.                      *
*                                                                       *
*************************************************************************
*/

int ModifiedBratuProblem::evaluateNonlinearFunction(
   Vec xcur,
   Vec fcur)
{
#ifdef DEBUG_CHECK_ASSERTIONS
   assert(!(xcur == NULL));
   assert(!(fcur == NULL));
#endif

   tbox::Pointer< solv::SAMRAIVectorReal<NDIM,double> > x =
      solv::PETSc_SAMRAIVectorReal<NDIM,double>::getSAMRAIVector(xcur);
   tbox::Pointer< solv::SAMRAIVectorReal<NDIM,double> > f =
      solv::PETSc_SAMRAIVectorReal<NDIM,double>::getSAMRAIVector(fcur);

   evaluateBratuFunction(x, f);

   return(0);
}

int ModifiedBratuProblem::evaluateJacobian(
   Vec x)
{
#ifdef DEBUG_CHECK_ASSERTIONS
   assert(!(x == NULL));
#endif

   tbox::Pointer< solv::SAMRAIVectorReal<NDIM,double> > xvec =
      solv::PETSc_SAMRAIVectorReal<NDIM,double>::getSAMRAIVector(x);

   evaluateBratuJacobian(xvec); 

   return(0);
}

int ModifiedBratuProblem::jacobianTimesVector(
   Vec xin,
   Vec xout)
{
#ifdef DEBUG_CHECK_ASSERTIONS
   assert(xin != NULL);
   assert(xout != NULL);
#endif

   tbox::Pointer< solv::SAMRAIVectorReal<NDIM,double> > xinvec =
      solv::PETSc_SAMRAIVectorReal<NDIM,double>::getSAMRAIVector(xin);
   tbox::Pointer< solv::SAMRAIVectorReal<NDIM,double> > xoutvec =
      solv::PETSc_SAMRAIVectorReal<NDIM,double>::getSAMRAIVector(xout);

   jacobianTimesVector(xinvec, xoutvec);

   return(0);
}

int ModifiedBratuProblem::setupPreconditioner(Vec x)
{
  tbox::Pointer<solv::SAMRAIVectorReal<NDIM,double> > uvec =
    solv::PETSc_SAMRAIVectorReal<NDIM,double>::getSAMRAIVector(x);
  setupBratuPreconditioner(uvec);

   return(0);
}

int ModifiedBratuProblem::applyPreconditioner(
   Vec r, 
   Vec z)
{
#ifdef DEBUG_CHECK_ASSERTIONS
   assert(r != NULL);
   assert(z != NULL);
#endif

   tbox::Pointer< solv::SAMRAIVectorReal<NDIM,double> > rhs =
      solv::PETSc_SAMRAIVectorReal<NDIM,double>::getSAMRAIVector(r);
   tbox::Pointer< solv::SAMRAIVectorReal<NDIM,double> > soln =
      solv::PETSc_SAMRAIVectorReal<NDIM,double>::getSAMRAIVector(z);

   return(applyBratuPreconditioner(rhs, soln));
}

/*
*************************************************************************
*                                                                       *
* Generic user-supplied functions.  Here the interfaces are expressed   *
* in terms of generic SAMRAIVectors.  These methods are invoked from    *
* within a nonlinear solver package (through an appropriate interface). *
* Since we have no control over the vectors that appear as actual       *
* arguments, each of these methods must create communication algorithms *
* and schedules to fill ghost cells before stencil operations can be    *
* applied.                                                              *
*                                                                       *
*************************************************************************
*/

/*
*************************************************************************
*                                                                       *
* Evaluate nonlinear residual at the vector "x" and place the result in *
* the vector "f".                                                       *
*                                                                       *
*************************************************************************
*/

void ModifiedBratuProblem::evaluateBratuFunction(
   tbox::Pointer< solv::SAMRAIVectorReal<NDIM,double> > x,
   tbox::Pointer< solv::SAMRAIVectorReal<NDIM,double> > f)
{
#ifdef DEBUG_CHECK_ASSERTIONS
   assert(!x.isNull());
   assert(!f.isNull());
#endif

   static tbox::Pointer<tbox::Timer> copy_timer = 
      tbox::TimerManager::getManager()->getTimer("apps::usrFcns::evaluateBratuFunction");
   
   tbox::Pointer<hier::PatchHierarchy<NDIM> > hierarchy = d_solution_vector->getPatchHierarchy();

   /*
    * Create a coarsen algorithm to coarsen data on fine patch interiors.
    * Then, cycle through the levels and coarsen the data.
    */
   CoarsenAlgorithm<NDIM> eval_average;
   eval_average.registerCoarsen(x->getComponentDescriptorIndex(0),
                                x->getComponentDescriptorIndex(0),
                                d_soln_coarsen_op);
   int amr_level = 0;
   for (amr_level = hierarchy->getFinestLevelNumber()-1;
        amr_level >=0;
        amr_level-- ) {
      tbox::Pointer<hier::PatchLevel<NDIM> > level       = hierarchy->getPatchLevel(amr_level);
      tbox::Pointer<hier::PatchLevel<NDIM> > finer_level = hierarchy->getPatchLevel(amr_level+1);
      eval_average.resetSchedule(d_soln_coarsen_schedule[amr_level]);
      d_soln_coarsen_schedule[amr_level]->coarsenData();
      d_soln_coarsen.resetSchedule(d_soln_coarsen_schedule[amr_level]);
   }

   /*
    * Create a refine algorithm to fill ghost cells of solution variable.
    */

   copy_timer->start(); 
   RefineAlgorithm<NDIM> eval_fill;
   eval_fill.registerRefine(d_soln_scratch_id, 
                            x->getComponentDescriptorIndex(0), 
                            d_soln_scratch_id, 
                            d_soln_refine_op);
   copy_timer->stop();

   for (amr_level = hierarchy->getFinestLevelNumber();
        amr_level >= 0;
        amr_level-- ) {
      tbox::Pointer<hier::PatchLevel<NDIM> > level = hierarchy->getPatchLevel(amr_level);

      copy_timer->start();
      eval_fill.resetSchedule( d_soln_fill_schedule[amr_level]);
      d_soln_fill_schedule[amr_level]->fillData(d_new_time);
      d_soln_fill.resetSchedule( d_soln_fill_schedule[amr_level] );
      copy_timer->stop();

      level->allocatePatchData(d_flux_id);
      if (amr_level > 0) {
         level->allocatePatchData(d_coarse_fine_flux_id);
      }

      for (hier::PatchLevel<NDIM>::Iterator p(level); p; p++) {
         tbox::Pointer<hier::Patch<NDIM> > patch = level->getPatch(p());  

         const tbox::Pointer<geom::CartesianPatchGeometry<NDIM> > patch_geom = 
            patch->getPatchGeometry();
         const double* dx   = patch_geom->getDx();
         const hier::Index<NDIM> ifirst = patch->getBox().lower();
         const hier::Index<NDIM> ilast  = patch->getBox().upper();

         tbox::Pointer< pdat::CellData<NDIM,double> > u =
            patch->getPatchData(d_solution, d_scratch);
         tbox::Pointer< pdat::SideData<NDIM,double> > diffusion =
            patch->getPatchData(d_diffusion_coef, d_scratch);
         tbox::Pointer< pdat::SideData<NDIM,double> > flux =
            patch->getPatchData(d_flux_id);
         tbox::Pointer< pdat::OutersideData<NDIM,double> > coarse_fine_flux =
            patch->getPatchData(d_coarse_fine_flux_id);
#if (NDIM == 1)
         FORT_EVALFACEFLUXES( ifirst(0), ilast(0),
                              NUM_GHOSTS_U,
                              diffusion->getPointer(0),
                              u->getPointer(),
                              dx,
                              flux->getPointer(0));
#endif
#if (NDIM == 2)
         FORT_EVALFACEFLUXES( ifirst(0), ilast(0),
                              ifirst(1), ilast(1),
                              NUM_GHOSTS_U,
                              diffusion->getPointer(0),
                              diffusion->getPointer(1),
                              u->getPointer(),
                              dx,
                              flux->getPointer(0),
                              flux->getPointer(1) );
#endif 
#if (NDIM == 3)
         FORT_EVALFACEFLUXES( ifirst(0), ilast(0),
                              ifirst(1), ilast(1),
                              ifirst(2), ilast(2),
                              NUM_GHOSTS_U,
                              diffusion->getPointer(0),
                              diffusion->getPointer(1),
                              diffusion->getPointer(2),
                              u->getPointer(),
                              dx,
                              flux->getPointer(0),
                              flux->getPointer(1),
                              flux->getPointer(2) );
#endif
#if (NDIM == 1) 
         const tbox::Array<hier::BoundaryBox<NDIM> > bdry_faces =
            patch_geom->getNodeBoundary();
#endif
#if (NDIM == 2) 
         const tbox::Array<hier::BoundaryBox<NDIM> > bdry_faces =
            patch_geom->getEdgeBoundary();
#endif
#if (NDIM == 3)
         const tbox::Array<hier::BoundaryBox<NDIM> > bdry_faces =
            patch_geom->getFaceBoundary();
#endif
         for (int i = 0; i < bdry_faces.getSize(); i++) {

            hier::Box<NDIM> bbox = bdry_faces[i].getBox();
            const hier::Index<NDIM> ibeg = bbox.lower();
            const hier::Index<NDIM> iend = bbox.upper(); 
            int face = bdry_faces[i].getLocationIndex();

#if (NDIM == 1)
            FORT_EWBCFLUXFIX( ifirst(0), ilast(0),
                              NUM_GHOSTS_U,
                              dx,
                              u->getPointer(),
                              flux->getPointer(0),
                              bbox.lower(), bbox.upper(),
                              face );
#endif
#if (NDIM == 2)
            FORT_EWBCFLUXFIX( ifirst(0), ilast(0),
                              ifirst(1), ilast(1),
                              NUM_GHOSTS_U,
                              dx,
                              u->getPointer(),
                              flux->getPointer(0),
                              bbox.lower(), bbox.upper(),
                              face );
            FORT_NSBCFLUXFIX( ifirst(0), ilast(0),
                              ifirst(1), ilast(1),
                              NUM_GHOSTS_U,
                              dx,
                              u->getPointer(),
                              flux->getPointer(1),
                              bbox.lower(), bbox.upper(),
                              face );
#endif
#if (NDIM == 3)
            FORT_EWBCFLUXFIX( ifirst(0), ilast(0),
                              ifirst(1), ilast(1),
                              ifirst(2), ilast(2),
                              NUM_GHOSTS_U,
                              dx,
                              u->getPointer(),
                              flux->getPointer(0),
                              bbox.lower(), bbox.upper(),
                              face );
            FORT_NSBCFLUXFIX( ifirst(0), ilast(0),
                              ifirst(1), ilast(1),
                              ifirst(2), ilast(2),
                              NUM_GHOSTS_U,
                              dx,
                              u->getPointer(),
                              flux->getPointer(1),
                              bbox.lower(), bbox.upper(),
                              face );
            FORT_TBBCFLUXFIX( ifirst(0), ilast(0),
                              ifirst(1), ilast(1),
                              ifirst(2), ilast(2),
                              NUM_GHOSTS_U,
                              dx,
                              u->getPointer(),
                              flux->getPointer(2),
                              bbox.lower(), bbox.upper(),
                              face );
#endif

         } // end loop over boundary faces

         // correctPatchFlux( level, patch, u );

         if ( amr_level > 0 ) {

            for (int side = 0; side < 2; side++) {
#if (NDIM == 1)
               FORT_EWFLUXCOPY( ifirst(0), ilast(0),
                                flux->getPointer(0),
                                coarse_fine_flux->getPointer(0,side),
                                side );
#endif
#if (NDIM == 2)
               FORT_EWFLUXCOPY( ifirst(0), ilast(0),
                                ifirst(1), ilast(1),
                                flux->getPointer(0),
                                coarse_fine_flux->getPointer(0,side),
                                side );
               FORT_NSFLUXCOPY( ifirst(0), ilast(0),
                                ifirst(1), ilast(1),
                                flux->getPointer(1),
                                coarse_fine_flux->getPointer(1,side),
                                side );
#endif
#if (NDIM == 3)
               FORT_EWFLUXCOPY( ifirst(0), ilast(0),
                                ifirst(1), ilast(1),
                                ifirst(2), ilast(2),
                                flux->getPointer(0),
                                coarse_fine_flux->getPointer(0,side),
                                side );
               FORT_NSFLUXCOPY( ifirst(0), ilast(0),
                                ifirst(1), ilast(1),
                                ifirst(2), ilast(2),
                                flux->getPointer(1),
                                coarse_fine_flux->getPointer(1,side),
                                side );
               FORT_TBFLUXCOPY( ifirst(0), ilast(0),
                                ifirst(1), ilast(1),
                                ifirst(2), ilast(2),
                                flux->getPointer(2),
                                coarse_fine_flux->getPointer(2,side),
                                side );
#endif


            }
         }
      } // end first pass over patches in level
      
      /*
       * Now that fluxes have been computed in each patch on the level, 
       *  those that lie at coarse/fine interfaces must be replaced by 
       *  the sum of fluxes on coincident faces from the next finer level.
       */

      if (amr_level < hierarchy->getFinestLevelNumber()) {
           
         tbox::Pointer<hier::PatchLevel<NDIM> > finer_level = 
            hierarchy->getPatchLevel(amr_level+1);  

         d_flux_coarsen_schedule[amr_level]->coarsenData();

         finer_level->deallocatePatchData(d_coarse_fine_flux_id);  

      } // end face flux fixup

      /*
       * Now that the right fluxes reside at coarse/fine interfaces,
       *  complete function evaluation by differencing.
       */

      for (hier::PatchLevel<NDIM>::Iterator p(level); p; p++) {
         tbox::Pointer<hier::Patch<NDIM> > patch = level->getPatch(p());

         const tbox::Pointer<geom::CartesianPatchGeometry<NDIM> > patch_geom =
            patch->getPatchGeometry();

         const double* dx   = patch_geom->getDx();
         const double* xlo  = patch_geom->getXLower();
         const double* xhi  = patch_geom->getXUpper();
         const hier::Index<NDIM> ifirst = patch->getBox().lower();
         const hier::Index<NDIM> ilast  = patch->getBox().upper();

         tbox::Pointer< pdat::CellData<NDIM,double> > u =
            patch->getPatchData(d_solution, d_scratch); 
         tbox::Pointer< pdat::CellData<NDIM,double> > u_cur =
            patch->getPatchData(d_solution, d_current); 
         tbox::Pointer< pdat::CellData<NDIM,double> > source = 
            patch->getPatchData(d_source_term, d_scratch);
         tbox::Pointer< pdat::CellData<NDIM,double> > exponential = 
            patch->getPatchData(d_exponential_term, d_scratch);
         tbox::Pointer< pdat::SideData<NDIM,double> > flux =
            patch->getPatchData(d_flux_id);
         tbox::Pointer< pdat::CellData<NDIM,double> > fcur = 
            f->getComponentPatchData(0, *patch);
         tbox::Pointer< pdat::CellData<NDIM,double> > xdat = 
            x->getComponentPatchData(0, *patch);
#if (NDIM == 1)
         FORT_EVALEXPONENTIAL(ifirst(0), ilast(0),
                              xdat->getPointer(),
                              d_lambda,
                              exponential->getPointer());
#endif
#if (NDIM == 2)
         FORT_EVALEXPONENTIAL(ifirst(0), ilast(0),
                              ifirst(1), ilast(1),
                              xdat->getPointer(),
                              d_lambda,
                              exponential->getPointer());
#endif
#if (NDIM == 3)
         FORT_EVALEXPONENTIAL(ifirst(0), ilast(0),
                              ifirst(1), ilast(1),
                              ifirst(2), ilast(2),
                              xdat->getPointer(),
                              d_lambda,
                              exponential->getPointer());
#endif
#if (NDIM == 1)
         FORT_EVALSOURCE( ifirst(0), ilast(0),
                          d_lambda,
                          xlo, xhi, dx, d_new_time,
                          source->getPointer() );
#endif
#if (NDIM == 2)
         FORT_EVALSOURCE( ifirst(0), ilast(0),
                          ifirst(1), ilast(1),
                          d_lambda,
                          xlo, xhi, dx, d_new_time,
                          source->getPointer() );
#endif
#if (NDIM == 3)
         FORT_EVALSOURCE( ifirst(0), ilast(0),
                          ifirst(1), ilast(1),
                          ifirst(2), ilast(2),
                          d_lambda,
                          xlo, xhi, dx, d_new_time,
                          source->getPointer() );
#endif
#if (NDIM == 1)
         FORT_EVALBRATU( ifirst(0), ilast(0),
                         NUM_GHOSTS_U,
                         flux->getPointer(0),
                         source->getPointer(),
                         exponential->getPointer(),
                         u_cur->getPointer(),
                         u->getPointer(),
                         dx, d_current_dt,
                         fcur->getPointer() );
#endif
#if (NDIM == 2)
         FORT_EVALBRATU( ifirst(0), ilast(0),
                         ifirst(1), ilast(1),
                         NUM_GHOSTS_U,
                         flux->getPointer(0),
                         flux->getPointer(1),
                         source->getPointer(),
                         exponential->getPointer(),
                         u_cur->getPointer(),
                         u->getPointer(),
                         dx, d_current_dt,
                         fcur->getPointer() );
#endif
#if (NDIM == 3)
         FORT_EVALBRATU( ifirst(0), ilast(0),
                         ifirst(1), ilast(1),
                         ifirst(2), ilast(2),
                         NUM_GHOSTS_U,
                         flux->getPointer(0),
                         flux->getPointer(1),
                         flux->getPointer(2),
                         source->getPointer(),
                         exponential->getPointer(),
                         u_cur->getPointer(),
                         u->getPointer(),
                         dx, d_current_dt,
                         fcur->getPointer() );
#endif
      } // end second pass over patches in level

      level->deallocatePatchData(d_flux_id);

   } // end loop over levels

   /*
    * Create a coarsen algorithm to coarsen data on fine patch interiors.
    * Then, cycle through the levels and coarsen the data.
    */

   CoarsenAlgorithm<NDIM> f_average;
   f_average.registerCoarsen(f->getComponentDescriptorIndex(0),
                             f->getComponentDescriptorIndex(0),
                             d_soln_coarsen_op);
   for (amr_level = hierarchy->getFinestLevelNumber()-1;
        amr_level >=0;
        amr_level-- ) {
      tbox::Pointer<hier::PatchLevel<NDIM> > level       = hierarchy->getPatchLevel(amr_level);
      tbox::Pointer<hier::PatchLevel<NDIM> > finer_level = hierarchy->getPatchLevel(amr_level+1);
      /*
	We take advantage of the knowlege that f has the same
	structure as the solution and reset the solution coarsening
	schedule instead of creating a new schedule for f.
       */
      f_average.resetSchedule(d_soln_coarsen_schedule[amr_level]);
      d_soln_coarsen_schedule[amr_level]->coarsenData();
      d_soln_coarsen.resetSchedule(d_soln_coarsen_schedule[amr_level]);
   }
}

/*
*************************************************************************
*                                                                       *
* Evaluate Jacobian-vector product on input vector "v" and place the    *
* result in the vector "Jv".                                            *
*                                                                       *
*************************************************************************
*/

int 
ModifiedBratuProblem::jacobianTimesVector(
   tbox::Pointer< solv::SAMRAIVectorReal<NDIM,double> > v,
   tbox::Pointer< solv::SAMRAIVectorReal<NDIM,double> > Jv )
{
#ifdef DEBUG_CHECK_ASSERTIONS
  assert(!(v    == (tbox::Pointer< solv::SAMRAIVectorReal<NDIM,double> >)NULL));
  assert(!(Jv   == (tbox::Pointer< solv::SAMRAIVectorReal<NDIM,double> >)NULL));
#endif

   /*
    * Create a coarsen algorithm to average down fine-to-coarse.
    */

   CoarsenAlgorithm<NDIM> jacv_average;
   jacv_average.registerCoarsen(v->getComponentDescriptorIndex(0),
                                v->getComponentDescriptorIndex(0),
                                d_soln_coarsen_op);

   /*
    * Now cycle through the levels, finest to coarsest, and average 
    * down the data.
    */

   tbox::Pointer<hier::PatchHierarchy<NDIM> > hierarchy = Jv->getPatchHierarchy();

   for (int amr_level = hierarchy->getFinestLevelNumber()-1;
            amr_level >=0;
            amr_level-- ) {

      tbox::Pointer<hier::PatchLevel<NDIM> > level       = hierarchy->getPatchLevel(amr_level);
      tbox::Pointer<hier::PatchLevel<NDIM> > finer_level = hierarchy->getPatchLevel(amr_level+1);
      jacv_average.resetSchedule(d_soln_coarsen_schedule[amr_level]);
      d_soln_coarsen_schedule[amr_level]->coarsenData();
      d_soln_coarsen.resetSchedule(d_soln_coarsen_schedule[amr_level]);

    }

   /*
    * Create a refine algorithm needed to fill ghost cells of vector being
    * multiplied.
    */
   static tbox::Pointer<tbox::Timer> copy_timer = 
      tbox::TimerManager::getManager()->getTimer("apps::usrFcns::jacobianTimesVector");

   copy_timer->start(); 
   RefineAlgorithm<NDIM> jacv_fill;
   jacv_fill.registerRefine(d_soln_scratch_id,
                            v->getComponentDescriptorIndex(0),
                            d_soln_scratch_id,
                            d_soln_refine_op);
   copy_timer->stop(); 

   /*
    * Perform Jacobian-vector product by looping over levels in the hierarchy.
    * The RefineSchedule created by the above RefineAlgorithm will copy
    * the data in the input vector into the interior of the specified scratch
    * space as well as filling the scratch space ghost cells.  Within each
    * level storage locations for each patch are extracted and passed to
    * computation routines.
    *
    * The product is implemented in two passes.  In the first pass, fluxes
    * are computed.  On all levels except the finest, the fluxes that
    * reside at coarse-fine interfaces are replaced by an average of the
    * fluxes from the next finer level.  This ensures proper flux matching
    * at the interfaces.
    *
    * In the second pass, the fluxes that reside at the interfaces are
    * differenced to complete the Jacobian-vector product.  The exponential
    * term, which is a cell-centered quantity, is also added in.
    */

   for (int amr_level = hierarchy->getFinestLevelNumber();
        amr_level >= 0;
        amr_level-- ) {

      tbox::Pointer<hier::PatchLevel<NDIM> > level = hierarchy->getPatchLevel(amr_level);

   /*
    * Allocate scratch space on this level used to store the fluxes.  Also,
    * on all but the coarsest level, allocate space for copies of these
    * fluxes to be averaged down to the next coarser level.
    */

      level->allocatePatchData(d_flux_id);

      if (amr_level > 0) {
         level->allocatePatchData(d_coarse_fine_flux_id);
      }

   /*
    * Fill ghost cell locations on this level.
    */

      copy_timer->start(); 
      jacv_fill.resetSchedule(d_soln_fill_schedule[amr_level]);
      d_soln_fill_schedule[amr_level]->fillData(d_new_time);
      d_soln_fill.resetSchedule(d_soln_fill_schedule[amr_level]);
      copy_timer->stop(); 

   /*
    * Now sweep through patches on a level, evaluating fluxes on faces
    * of cells.
    */

      for (hier::PatchLevel<NDIM>::Iterator p(level); p; p++) {
         tbox::Pointer<hier::Patch<NDIM> > patch = level->getPatch(p());

         const hier::Index<NDIM> ifirst = patch->getBox().lower();
         const hier::Index<NDIM> ilast  = patch->getBox().upper();
         
         tbox::Pointer<geom::CartesianPatchGeometry<NDIM> > patch_geom =
           patch->getPatchGeometry();
         const double* dx = patch_geom->getDx();
         tbox::Pointer< pdat::CellData<NDIM,double> > vdat =
            patch->getPatchData(d_soln_scratch_id);
         tbox::Pointer< pdat::SideData<NDIM,double> > diffusion =
            patch->getPatchData(d_diffusion_coef, d_scratch);
         tbox::Pointer< pdat::SideData<NDIM,double> > flux =
            patch->getPatchData(d_flux_id);
         tbox::Pointer< pdat::OutersideData<NDIM,double> > coarse_fine_flux =
            patch->getPatchData(d_coarse_fine_flux_id);

#if (NDIM == 1)
         FORT_EVALFACEFLUXES( ifirst(0), ilast(0),
                              NUM_GHOSTS_U,
                              diffusion->getPointer(0),
                              vdat->getPointer(),
                              dx,
                              flux->getPointer(0) );
#endif
#if (NDIM == 2)
         FORT_EVALFACEFLUXES( ifirst(0), ilast(0),
                              ifirst(1), ilast(1),
                              NUM_GHOSTS_U,
                              diffusion->getPointer(0),
                              diffusion->getPointer(1),
                              vdat->getPointer(),
                              dx,
                              flux->getPointer(0),
                              flux->getPointer(1) );
#endif
#if (NDIM == 3)
         FORT_EVALFACEFLUXES( ifirst(0), ilast(0),
                              ifirst(1), ilast(1),
                              ifirst(2), ilast(2),
                              NUM_GHOSTS_U,
                              diffusion->getPointer(0),
                              diffusion->getPointer(1),
                              diffusion->getPointer(2),
                              vdat->getPointer(),
                              dx,
                              flux->getPointer(0),
                              flux->getPointer(1),
                              flux->getPointer(2) );
#endif

         /*
	   We employ a convention in which boundary conditions are located on
	   the physical boundaries, but are stored at the centers of ghost
	   cells.  As a result the grid spacing is smaller along the physical
	   boundaries, and the fluxes at these locations have to be modified.
	 */

#if (NDIM == 1)
         const tbox::Array<hier::BoundaryBox<NDIM> > bdry_faces =
            patch_geom->getNodeBoundary();
#endif
#if (NDIM == 2)
         const tbox::Array<hier::BoundaryBox<NDIM> > bdry_faces =
            patch_geom->getEdgeBoundary();
#endif
#if (NDIM == 3)
         const tbox::Array<hier::BoundaryBox<NDIM> > bdry_faces =
            patch_geom->getFaceBoundary();
#endif
         for (int i = 0; i < bdry_faces.getSize(); i++) {

            hier::Box<NDIM> bbox = bdry_faces[i].getBox();
            const hier::Index<NDIM> ibeg = bbox.lower();
            const hier::Index<NDIM> iend = bbox.upper();
            int face = bdry_faces[i].getLocationIndex();

#if (NDIM == 1)
            FORT_EWBCFLUXFIX( ifirst(0), ilast(0),
                              NUM_GHOSTS_U,
                              dx,
                              vdat->getPointer(),
                              flux->getPointer(0),
                              bbox.lower(), bbox.upper(),
                              face );
#endif
#if (NDIM == 2)
            FORT_EWBCFLUXFIX( ifirst(0), ilast(0),
                              ifirst(1), ilast(1),
                              NUM_GHOSTS_U,
                              dx,
                              vdat->getPointer(),
                              flux->getPointer(0),
                              bbox.lower(), bbox.upper(),
                              face );
            FORT_NSBCFLUXFIX( ifirst(0), ilast(0),
                              ifirst(1), ilast(1),
                              NUM_GHOSTS_U,
                              dx,
                              vdat->getPointer(),
                              flux->getPointer(1),
                              bbox.lower(), bbox.upper(),
                              face );
#endif
#if (NDIM == 3)
            FORT_EWBCFLUXFIX( ifirst(0), ilast(0),
                              ifirst(1), ilast(1),
                              ifirst(2), ilast(2),
                              NUM_GHOSTS_U,
                              dx,
                              vdat->getPointer(),
                              flux->getPointer(0),
                              bbox.lower(), bbox.upper(),
                              face );
            FORT_NSBCFLUXFIX( ifirst(0), ilast(0),
                              ifirst(1), ilast(1),
                              ifirst(2), ilast(2),
                              NUM_GHOSTS_U,
                              dx,
                              vdat->getPointer(),
                              flux->getPointer(1),
                              bbox.lower(), bbox.upper(),
                              face );
            FORT_TBBCFLUXFIX( ifirst(0), ilast(0),
                              ifirst(1), ilast(1),
                              ifirst(2), ilast(2),
                              NUM_GHOSTS_U,
                              dx,
                              vdat->getPointer(),
                              flux->getPointer(2),
                              bbox.lower(), bbox.upper(),
                              face );
#endif
         } // end loop over boundary faces

         //         correctPatchFlux( level, patch, vdat );

   /*
    * On all but the coarsest level, we save the face fluxes on the outer edge
    * of the grid level so that they can be averaged down to the next 
    * coarser level.
    */

         if ( amr_level > 0 ) {
            for (int side = 0; side < 2; side++) {
#if (NDIM == 1)
               FORT_EWFLUXCOPY( ifirst(0), ilast(0),
                                flux->getPointer(0),
                                coarse_fine_flux->getPointer(0,side),
                                side );
#endif
#if (NDIM == 2)
               FORT_EWFLUXCOPY( ifirst(0), ilast(0),
                                ifirst(1), ilast(1),
                                flux->getPointer(0),
                                coarse_fine_flux->getPointer(0,side),
                                side );
               FORT_NSFLUXCOPY( ifirst(0), ilast(0),
                                ifirst(1), ilast(1),
                                flux->getPointer(1),
                                coarse_fine_flux->getPointer(1,side),
                                side );
#endif
#if (NDIM == 3)
               FORT_EWFLUXCOPY( ifirst(0), ilast(0),
                                ifirst(1), ilast(1),
                                ifirst(2), ilast(2),
                                flux->getPointer(0),
                                coarse_fine_flux->getPointer(0,side),
                                side );
               FORT_NSFLUXCOPY( ifirst(0), ilast(0),
                                ifirst(1), ilast(1),
                                ifirst(2), ilast(2),
                                flux->getPointer(1),
                                coarse_fine_flux->getPointer(1,side),
                                side );
               FORT_TBFLUXCOPY( ifirst(0), ilast(0),
                                ifirst(1), ilast(1),
                                ifirst(2), ilast(2),
                                flux->getPointer(2),
                                coarse_fine_flux->getPointer(2,side),
                                side );
#endif
            }
         }

      } // end first pass over patches in level

      /*
       * Now that fluxes have been computed in each patch on the level,
       *  those that lie at coarse/fine interfaces must be replaced by
       *  the sum of fluxes on coincident faces from the next finer level.
       */

      if (amr_level < hierarchy->getFinestLevelNumber()) {

         tbox::Pointer<hier::PatchLevel<NDIM> > finer_level =
            hierarchy->getPatchLevel(amr_level+1);

         d_flux_coarsen_schedule[amr_level]->coarsenData();

         finer_level->deallocatePatchData(d_coarse_fine_flux_id);

      } // end face flux fixup

      /*
       * Now that the right fluxes reside at coarse/fine interfaces,
       *  complete function evaluation by differencing.
       */

      for (hier::PatchLevel<NDIM>::Iterator p(level); p; p++) {
         tbox::Pointer<hier::Patch<NDIM> > patch = level->getPatch(p());

         const tbox::Pointer<geom::CartesianPatchGeometry<NDIM> > patch_geom =
            patch->getPatchGeometry();

         const hier::Index<NDIM> ifirst = patch->getBox().lower();
         const hier::Index<NDIM> ilast  = patch->getBox().upper();

         const double* dx   = patch_geom->getDx();

         tbox::Pointer< pdat::CellData<NDIM,double> > jac_a =
            patch->getPatchData(d_jacobian_a_id);
         tbox::Pointer< pdat::CellData<NDIM,double> > Jvdat =
            Jv->getComponentPatchData(0, *patch);
         tbox::Pointer< pdat::CellData<NDIM,double> > vdat =
            patch->getPatchData(d_soln_scratch_id);
         tbox::Pointer< pdat::SideData<NDIM,double> > flux =
            patch->getPatchData(d_flux_id);

#ifdef DEBUG_CHECK_ASSERTIONS
	 assert( vdat->getGhostCellWidth() == hier::IntVector<NDIM>(NUM_GHOSTS_U) );
	 assert( Jvdat->getGhostCellWidth() == hier::IntVector<NDIM>(0) );
	 assert( jac_a->getGhostCellWidth() == hier::IntVector<NDIM>(0) );
#endif

#if (NDIM == 1)
         compjv_( ifirst(0), ilast(0),
		  ifirst(1), ilast(1),
		  NUM_GHOSTS_U,
		  jac_a->getPointer(),
		  flux->getPointer(0),
		  flux->getPointer(1),
		  vdat->getPointer(),
		  dx, d_current_dt,
		  Jvdat->getPointer() );
#if 0
         FORT_BRATUJV( ifirst(0), ilast(0),
                       NUM_GHOSTS_U,
                       flux->getPointer(0),
                       exponential->getPointer(),
                       vdat->getPointer(),
                       dx, d_current_dt,
                       Jvdat->getPointer() );
#endif
#endif
#if (NDIM == 2)
         compjv_( ifirst(0), ilast(0),
		  ifirst(1), ilast(1),
		  NUM_GHOSTS_U,
		  jac_a->getPointer(),
		  flux->getPointer(0),
		  flux->getPointer(1),
		  vdat->getPointer(),
		  dx, d_current_dt,
		  Jvdat->getPointer() );
#if 0
         FORT_BRATUJV( ifirst(0), ilast(0),
                       ifirst(1), ilast(1),
                       NUM_GHOSTS_U,
                       flux->getPointer(0),
                       flux->getPointer(1),
                       exponential->getPointer(),
                       vdat->getPointer(),
                       dx, d_current_dt,
                       Jvdat->getPointer() );
#endif
#endif
#if (NDIM == 3)
         compjv_( ifirst(0), ilast(0),
		  ifirst(1), ilast(1),
		  ifirst(2), ilast(2),
		  NUM_GHOSTS_U,
		  jac_a->getPointer(),
		  flux->getPointer(0),
		  flux->getPointer(1),
		  flux->getPointer(2),
		  vdat->getPointer(),
		  dx, d_current_dt,
		  Jvdat->getPointer() );
#if 0
         FORT_BRATUJV( ifirst(0), ilast(0),
                       ifirst(1), ilast(1),
                       ifirst(2), ilast(2),
                       NUM_GHOSTS_U,
                       flux->getPointer(0),
                       flux->getPointer(1),
                       flux->getPointer(2),
                       exponential->getPointer(),
                       vdat->getPointer(),
                       dx, d_current_dt,
                       Jvdat->getPointer() );
#endif
#endif

       }  // end second pass over patches

      level->deallocatePatchData(d_flux_id);

   } // end pass over levels

   return(0);

}

/*
*************************************************************************
*                                                                       *
* Set up FAC hierarchy preconditioner for Jacobian system.  Here we     *
* use the FAC hierarchy solver in SAMRAI which automatically sets       *
* up the composite grid system and uses hypre as a solver on each       *
* level.  Fortunately, the FAC hierarchy solver suits the problem       *
* we are solving, as the Jacobian matrix arises by discretizing:        *
*                                                                       *
*    div( D(x,t)*grad(u) ) + 1/dt - lambda * exp(u) - df/du             *
*                                                                       *
* evaluated at some iterate of the solution u.                          *
*                                                                       *
*************************************************************************
*/

void ModifiedBratuProblem::setupBratuPreconditioner(
   tbox::Pointer< solv::SAMRAIVectorReal<NDIM,double> > x)
{
#ifdef DEBUG_CHECK_ASSERTIONS
   assert(!x.isNull());
#endif

   tbox::Pointer<hier::PatchHierarchy<NDIM> > hierarchy = d_solution_vector->getPatchHierarchy();

   RefineAlgorithm<NDIM> eval_fill;
   eval_fill.registerRefine(d_soln_scratch_id,
                            x->getComponentDescriptorIndex(0),
                            d_soln_scratch_id,
                            d_soln_refine_op);

   for (int amr_level = hierarchy->getFinestLevelNumber();
        amr_level >= 0;
        amr_level-- ) {
      tbox::Pointer<hier::PatchLevel<NDIM> > level = hierarchy->getPatchLevel(amr_level);

      eval_fill.resetSchedule( d_soln_fill_schedule[amr_level]);
      d_soln_fill_schedule[amr_level]->fillData(d_new_time);
      d_soln_fill.resetSchedule( d_soln_fill_schedule[amr_level]);

      for (hier::PatchLevel<NDIM>::Iterator p(level); p; p++) {
         tbox::Pointer<hier::Patch<NDIM> > patch = level->getPatch(p());

         const tbox::Pointer<geom::CartesianPatchGeometry<NDIM> > patch_geom =
            patch->getPatchGeometry();

         const hier::Index<NDIM> ifirst =patch->getBox().lower();
         const hier::Index<NDIM> ilast  =patch->getBox().upper();

         const double* dx = patch_geom->getDx();
         const double* xlo = patch_geom->getXLower();

         double cell_vol = dx[0];
#if (NDIM > 1)
         cell_vol *= dx[1];
#endif
#if (NDIM > 2)
         cell_vol *= dx[2];
#endif
         tbox::Pointer< pdat::CellData<NDIM,double> > u =
            patch->getPatchData(d_solution, d_scratch);
         tbox::Pointer< pdat::CellData<NDIM,double> > exponential =
            patch->getPatchData(d_exponential_term, d_scratch);
         tbox::Pointer< pdat::CellData<NDIM,double> > source =
            patch->getPatchData(d_source_term, d_scratch); 
         tbox::Pointer< pdat::SideData<NDIM,double> > diffusion =
            patch->getPatchData(d_diffusion_coef, d_scratch);

         tbox::Pointer< pdat::CellData<NDIM,double> > a = patch->getPatchData(d_precond_a_id);
         tbox::Pointer< pdat::FaceData<NDIM,double> > b = patch->getPatchData(d_precond_b_id);

#ifdef DEBUG_CHECK_ASSERTIONS
	 assert( !exponential.isNull() );
	 assert( !source.isNull() );
	 assert( !u.isNull() );
	 assert( !diffusion.isNull() );
	 assert( !a.isNull() );
	 assert( !b.isNull() );
#endif

         /*
          * Compute exponential = lambda * exp(u)
          */
#if (NDIM == 1)
         compexpu_(ifirst(0),ilast(0),
                   NUM_GHOSTS_U,
                   d_lambda,
                   u->getPointer(),
                   exponential->getPointer());

         /*
          * Compute source = df/du
          */
         compsrcderv_(ifirst(0),ilast(0),
                      NUM_GHOSTS_U,
                      xlo, dx,
                      d_new_time,
                      d_lambda,
                      u->getPointer(),
                      source->getPointer());
#endif
#if (NDIM == 2)
         compexpu_(ifirst(0),ilast(0),
                   ifirst(1),ilast(1),
                   NUM_GHOSTS_U,
                   d_lambda,
                   u->getPointer(),
                   exponential->getPointer());

         /*
          * Compute source = df/du
          */
         compsrcderv_(ifirst(0),ilast(0),
                      ifirst(1),ilast(1),
                      NUM_GHOSTS_U,
                      xlo, dx,
                      d_new_time,
                      d_lambda,
                      u->getPointer(),
                      source->getPointer());
#endif
#if (NDIM == 3)
         compexpu_(ifirst(0),ilast(0),
                   ifirst(1),ilast(1),
                   ifirst(2),ilast(2),
                   NUM_GHOSTS_U,
                   d_lambda,
                   u->getPointer(),
                   exponential->getPointer());

         /*
          * Compute source = df/du
          */
         compsrcderv_(ifirst(0),ilast(0),
                      ifirst(1),ilast(1),
                      ifirst(2),ilast(2),
                      NUM_GHOSTS_U,
                      xlo, dx,
                      d_new_time,
                      d_lambda,
                      u->getPointer(),
                      source->getPointer());
#endif
         /*
          * Compute a = cell_vol * ( 1 - d_current_dt*(exponential + source))
          */
#if (NDIM == 1)
         compfacdiag_(ifirst(0),ilast(0),
                      d_current_dt,
                      cell_vol,
                      exponential->getPointer(),
                      source->getPointer(),
                      a->getPointer());
#endif
#if (NDIM == 2)
         compfacdiag_(ifirst(0),ilast(0),
                      ifirst(1),ilast(1),
                      d_current_dt,
                      cell_vol,
                      exponential->getPointer(),
                      source->getPointer(),
                      a->getPointer());
#endif
#if (NDIM == 3)
         compfacdiag_(ifirst(0),ilast(0),
                      ifirst(1),ilast(1),
                      ifirst(2),ilast(2),
                      d_current_dt,
                      cell_vol,
                      exponential->getPointer(),
                      source->getPointer(),
                      a->getPointer());
#endif

         /*
          * Copy side-centered diffusion coefficients to face-centered
          * and multiply by cell_vol*d_current_dt.
          */
#if (NDIM == 1)
         compfacoffdiag_(ifirst(0),ilast(0),
                         d_current_dt,
                         cell_vol,
                         diffusion->getPointer(0),
                         b->getPointer(0));
#endif
#if (NDIM == 2)
         compfacoffdiag_(ifirst(0),ilast(0),
                         ifirst(1),ilast(1),
                         d_current_dt,
                         cell_vol,
                         diffusion->getPointer(0),
                         diffusion->getPointer(1),
                         b->getPointer(0),
                         b->getPointer(1));
#endif
#if (NDIM == 3)
         compfacoffdiag_(ifirst(0),ilast(0),
                         ifirst(1),ilast(1),
                         ifirst(2),ilast(2),
                         d_current_dt,
                         cell_vol,
                         diffusion->getPointer(0),
                         diffusion->getPointer(1),
                         diffusion->getPointer(2),
                         b->getPointer(0),
                         b->getPointer(1),
                         b->getPointer(2));
#endif

      }

   }

   // This will have to generalized if we add in other boundary conditions

   d_FAC_solver->setBoundaries("Dirichlet");

   if ( d_precond_b_id == -1 ) {
      d_FAC_solver->setDConstant(1.0);
   }
   else {
      d_FAC_solver->setDConstant(1.0);
   }
   if ( d_precond_a_id == -1 ) {
      d_FAC_solver->setCConstant(0.0);
   }
   else {
      d_FAC_solver->setCPatchDataId(d_precond_a_id);
   }

}

/*
*************************************************************************
*                                                                       *
* Apply preconditioner where right-hand-side is "r" and "z" is the      *
* solution.   This routine assumes that the preconditioner setup call   *
* has already been invoked.  Return 0 if preconditioner fails;          *
* return 1 otherwise.                                                   *
*                                                                       *
*************************************************************************
*/

int ModifiedBratuProblem::applyBratuPreconditioner(
   tbox::Pointer< solv::SAMRAIVectorReal<NDIM,double> > r,
   tbox::Pointer< solv::SAMRAIVectorReal<NDIM,double> > z)
{
#ifdef DEBUG_CHECK_ASSERTIONS
   assert(!r.isNull());
   assert(!z.isNull());
#endif

   int ret_val = 0;

   tbox::Pointer<hier::PatchHierarchy<NDIM> > hierarchy = d_solution_vector->getPatchHierarchy();

   int r_indx = r->getComponentDescriptorIndex(0);
   int z_indx = z->getComponentDescriptorIndex(0);

   /*
    * Create a coarsen algorithm to coarsen rhs data on fine patch interiors.
    * Then, cycle through the levels and coarsen the data.
    */
   CoarsenAlgorithm<NDIM> pc_rhs_average;
   pc_rhs_average.registerCoarsen(r_indx, r_indx, d_soln_coarsen_op);

   for (int amr_level = hierarchy->getFinestLevelNumber()-1;
        amr_level >=0;
        amr_level-- ) {
      tbox::Pointer<hier::PatchLevel<NDIM> > level       = hierarchy->getPatchLevel(amr_level);
      tbox::Pointer<hier::PatchLevel<NDIM> > finer_level = hierarchy->getPatchLevel(amr_level+1);
      pc_rhs_average.resetSchedule(d_soln_coarsen_schedule[amr_level]);
      d_soln_coarsen_schedule[amr_level]->coarsenData();
      d_soln_coarsen.resetSchedule(d_soln_coarsen_schedule[amr_level]);
   }

   math::HierarchyCellDataOpsReal<NDIM,double>
      math_ops(hierarchy,
               0,
               hierarchy->getFinestLevelNumber());
   math_ops.setToScalar(d_soln_scratch_id,
                        0.0);

   static tbox::Pointer<tbox::Timer> pc_timer = 
      tbox::TimerManager::getManager()->
      getTimer("apps::usrFcns::applyBratuPreconditioner");
   pc_timer->start(); 

   bool converge;
   converge = d_FAC_solver->solveSystem(d_soln_scratch_id,
                                        r_indx,
                                        hierarchy,
                                        0, 
                                        hierarchy->getFinestLevelNumber() );

   pc_timer->stop(); 

   /*
    * Create a coarsen algorithm to coarsen soln data on fine patch interiors.
    * Then, cycle through the levels and coarsen the data.
    */
   CoarsenAlgorithm<NDIM> pc_sol_average;
   pc_sol_average.registerCoarsen(d_soln_scratch_id, 
                                  d_soln_scratch_id, 
                                  d_soln_coarsen_op);

   for (int amr_level = hierarchy->getFinestLevelNumber()-1;
        amr_level >=0;
        amr_level-- ) {
      tbox::Pointer<hier::PatchLevel<NDIM> > level       = hierarchy->getPatchLevel(amr_level);
      tbox::Pointer<hier::PatchLevel<NDIM> > finer_level = hierarchy->getPatchLevel(amr_level+1);
      pc_sol_average.resetSchedule(d_scratch_soln_coarsen_schedule[amr_level]);
      d_scratch_soln_coarsen_schedule[amr_level]->coarsenData();
      d_scratch_soln_coarsen.resetSchedule(d_scratch_soln_coarsen_schedule[amr_level]);
   }

   for (int amr_level = hierarchy->getFinestLevelNumber();
        amr_level >= 0;
        amr_level-- ) {
      tbox::Pointer<hier::PatchLevel<NDIM> > level = hierarchy->getPatchLevel(amr_level);

      for (hier::PatchLevel<NDIM>::Iterator p(level); p; p++) {
         tbox::Pointer<hier::Patch<NDIM> > patch = level->getPatch(p());

         tbox::Pointer< pdat::CellData<NDIM,double> > src_data =
            patch->getPatchData(d_soln_scratch_id);
         tbox::Pointer< pdat::CellData<NDIM,double> > dst_data = patch->getPatchData(z_indx);

         dst_data->copy(*src_data);
      }
   }

   return(ret_val);
}

/*
*************************************************************************
*                                                                       *
* Evaluate the Jacobian matrix A(x) for a given solution x.             *
*                                                                       *
*************************************************************************
*/

void ModifiedBratuProblem::evaluateBratuJacobian(
   tbox::Pointer< solv::SAMRAIVectorReal<NDIM,double> > x)
{
   tbox::Pointer<hier::PatchHierarchy<NDIM> > hierarchy = d_solution_vector->getPatchHierarchy();

   RefineAlgorithm<NDIM> eval_fill;
   eval_fill.registerRefine(d_soln_scratch_id,
                            x->getComponentDescriptorIndex(0),
                            d_soln_scratch_id,
                            d_soln_refine_op);

   for (int amr_level = hierarchy->getFinestLevelNumber();
        amr_level >= 0;
        amr_level-- ) {
      tbox::Pointer<hier::PatchLevel<NDIM> > level = hierarchy->getPatchLevel(amr_level);

      eval_fill.resetSchedule( d_soln_fill_schedule[amr_level]);
      d_soln_fill_schedule[amr_level]->fillData(d_new_time);
      d_soln_fill.resetSchedule( d_soln_fill_schedule[amr_level]);

      for (hier::PatchLevel<NDIM>::Iterator p(level); p; p++) {
         tbox::Pointer<hier::Patch<NDIM> > patch = level->getPatch(p());

         const tbox::Pointer<geom::CartesianPatchGeometry<NDIM> > patch_geom =
            patch->getPatchGeometry();

         const hier::Index<NDIM> ifirst = patch->getBox().lower();
         const hier::Index<NDIM> ilast  = patch->getBox().upper();

         const double* dx = patch_geom->getDx();
         const double* xlo = patch_geom->getXLower();

         double cell_vol = dx[0];
#if (NDIM > 1)
         cell_vol *= dx[1];
#endif
#if (NDIM > 2)
         cell_vol *= dx[2];
#endif

         tbox::Pointer< pdat::CellData<NDIM,double> > u =
            patch->getPatchData(d_solution, d_scratch);
         tbox::Pointer< pdat::CellData<NDIM,double> > exponential =
            patch->getPatchData(d_exponential_term, d_scratch);
         tbox::Pointer< pdat::CellData<NDIM,double> > source =
            patch->getPatchData(d_source_term, d_scratch); 
         tbox::Pointer< pdat::SideData<NDIM,double> > diffusion =
            patch->getPatchData(d_diffusion_coef, d_scratch);

         tbox::Pointer< pdat::CellData<NDIM,double> > a = patch->getPatchData(d_jacobian_a_id);
         tbox::Pointer< pdat::FaceData<NDIM,double> > b = patch->getPatchData(d_jacobian_b_id);

#if (NDIM == 1)
         /*
          * Compute exponential = lambda * exp(u)
          */
         compexpu_(ifirst(0),ilast(0),
                   NUM_GHOSTS_U,
                   d_lambda,
                   u->getPointer(),
                   exponential->getPointer());

         /*
          * Compute source = df/du
          */
         compsrcderv_(ifirst(0),ilast(0),
                      NUM_GHOSTS_U,
                      xlo, dx,
                      d_new_time,
                      d_lambda,
                      u->getPointer(),
                      source->getPointer());
#endif
#if (NDIM == 2)
         /*
          * Compute exponential = lambda * exp(u)
          */
         compexpu_(ifirst(0),ilast(0),
                   ifirst(1),ilast(1),
                   NUM_GHOSTS_U,
                   d_lambda,
                   u->getPointer(),
                   exponential->getPointer());

         /*
          * Compute source = df/du
          */
         compsrcderv_(ifirst(0),ilast(0),
                      ifirst(1),ilast(1),
                      NUM_GHOSTS_U,
                      xlo, dx,
                      d_new_time,
                      d_lambda,
                      u->getPointer(),
                      source->getPointer());
#endif
#if (NDIM == 3)
         /*
          * Compute exponential = lambda * exp(u)
          */
         compexpu_(ifirst(0),ilast(0),
                   ifirst(1),ilast(1),
                   ifirst(2),ilast(2),
                   NUM_GHOSTS_U,
                   d_lambda,
                   u->getPointer(),
                   exponential->getPointer());

         /*
          * Compute source = df/du
          */
         compsrcderv_(ifirst(0),ilast(0),
                      ifirst(1),ilast(1),
                      ifirst(2),ilast(2),
                      NUM_GHOSTS_U,
                      xlo, dx,
                      d_new_time,
                      d_lambda,
                      u->getPointer(),
                      source->getPointer());
#endif
         /*
          * Compute a = lambda*exp(u) + df/du
          */
	 math::PatchCellDataOpsReal<NDIM,double> cell_basic_ops;
	 cell_basic_ops.add( a,
			     exponential,
			     source,
			     a->getBox() );

      }
   }
}
    
/*
*************************************************************************
*                                                                       *
* Set physical boundary conditions for patch at a given time.           *
*                                                                       *
*************************************************************************
*/

void ModifiedBratuProblem::setPhysicalBoundaryConditions(
   hier::Patch<NDIM>& patch, 
   const double time,
   const hier::IntVector<NDIM>& ghost_width_to_fill )
{

  (void) ghost_width_to_fill;

  /* 
   * Grab data to operate on.
   */

   tbox::Pointer< pdat::CellData<NDIM,double> > u = patch.getPatchData(d_soln_scratch_id);

   const hier::Index<NDIM> ifirst = patch.getBox().lower();
   const hier::Index<NDIM> ilast =  patch.getBox().upper();

   /*
    * Determine boxes that touch the physical boundary.
    */

   const tbox::Pointer<geom::CartesianPatchGeometry<NDIM> > patch_geom = patch.getPatchGeometry();
#if (NDIM == 1) 
   const tbox::Array<hier::BoundaryBox<NDIM> > boundary = patch_geom->getNodeBoundary();
#endif
#if (NDIM == 2) 
   const tbox::Array<hier::BoundaryBox<NDIM> > boundary = patch_geom->getEdgeBoundary();
#endif
#if (NDIM == 3)
   const tbox::Array<hier::BoundaryBox<NDIM> > boundary = patch_geom->getFaceBoundary();
#endif

   /* 
    * Walk the list of boxes that describe the boundary, and apply
    * boundary conditions to each segment.
    */

   int face;
   for (int i=0; i<boundary.getSize(); i++) {
      hier::Box<NDIM> bbox = hier::Box<NDIM>(boundary[i].getBox());
      face = boundary[i].getLocationIndex();
#if (NDIM == 1)
      FORT_SETBC( ifirst(0), ilast(0),
                  NUM_GHOSTS_U,
                  u->getPointer(),
                  bbox.lower(), bbox.upper(),
                  face );
#endif
#if (NDIM == 2)
      FORT_SETBC( ifirst(0), ilast(0),
                  ifirst(1), ilast(1),
                  NUM_GHOSTS_U,
                  u->getPointer(),
                  bbox.lower(), bbox.upper(),
                  face );
#endif
#if (NDIM == 3)
      FORT_SETBC( ifirst(0), ilast(0),
                  ifirst(1), ilast(1),
                  ifirst(2), ilast(2),
                  NUM_GHOSTS_U,
                  u->getPointer(),
                  bbox.lower(), bbox.upper(),
                  face );
#endif
   }
}

/*
*************************************************************************
*                                                                       *
* Read data from input database.                                        *
*                                                                       *
*************************************************************************
*/

void ModifiedBratuProblem::getFromInput(
   tbox::Pointer<tbox::Database> db,
   bool is_from_restart)
{
#ifdef DEBUG_CHECK_ASSERTIONS
   assert(!db.isNull());
#endif

   if (db->keyExists("lambda")) {
      d_lambda = db->getDouble("lambda");
   } else {
      TBOX_ERROR(d_object_name << " -- Key data `lambda'"
                 << " missing in input.");
   }

   if (db->keyExists("timestep")) {
       d_input_dt = db->getDouble("timestep");
       if (d_input_dt < 0.0) {
          TBOX_ERROR(d_object_name << " Input error: timestep < 0.0");
       }
   } else {
      if (!is_from_restart) {
         TBOX_ERROR(d_object_name << " -- Key data `timestep'"
                                  << " missing in input.");
      }
   }

   if (db->keyExists("max_precond_its")) {
       d_max_precond_its = db->getInteger("max_precond_its");
       if (d_max_precond_its < 0) {
          TBOX_ERROR(d_object_name << " Input error: max_precond_its < 0");
       }
   } else {
      if (!is_from_restart) {
         TBOX_ERROR(d_object_name << " -- Key data `max_precond_its'"
                                  << " missing in input.");
      }
   }

   if (db->keyExists("precond_tol")) {
       d_precond_tol = db->getDouble("precond_tol");
       if (d_precond_tol <= 0.0) {
          TBOX_ERROR(d_object_name << " Input error: precond_tol <= 0.0");
       }
   } else {
      if (!is_from_restart) {
         TBOX_ERROR(d_object_name << " -- Key data `precond_tol'"
                                  << " missing in input.");
      }
   }

}

/*
*************************************************************************
*                                                                       *
* Put class version number and data members in database.                *
*                                                                       *
*************************************************************************
*/

void ModifiedBratuProblem::putToDatabase(tbox::Pointer<tbox::Database> db)
{
#ifdef DEBUG_CHECK_ASSERTIONS
   assert(!db.isNull());
#endif

   db->putInteger("MODIFIED_BRATU_PROBLEM", MODIFIED_BRATU_PROBLEM);
   
   db->putDouble("d_lambda", d_lambda);
   db->putDouble("d_input_dt", d_input_dt);
   db->putInteger("d_max_precond_its", d_max_precond_its);
   db->putDouble("d_precond_tol", d_precond_tol);

}

/*
*************************************************************************
*                                                                       *
* Print all class data members to given output stream.                  *
*                                                                       *
*************************************************************************
*/

void ModifiedBratuProblem::printClassData(ostream& os) const
{
   os << "\nModifiedBratuProblem::printClassData..." << endl;
   os << "ModifiedBratuProblem: this = " << (ModifiedBratuProblem*)this << endl;
   os << "d_object_name = " << d_object_name << endl;
   os << "d_grid_geometry = "
      << (geom::CartesianGridGeometry<NDIM>*)d_grid_geometry << endl;
   os << "d_grid_geometry = "
      << (geom::CartesianGridGeometry<NDIM>*)d_grid_geometry << endl;

   os << "d_soln_scratch_id =   " << d_soln_scratch_id << endl;
   os << "d_flux_id =   " << d_flux_id << endl;
   os << "d_coarse_fine_flux_id =   " << d_coarse_fine_flux_id << endl;
   os << "d_jacobian_a_id =   " << d_jacobian_a_id << endl;
   os << "d_jacobian_b_id =   " << d_jacobian_b_id << endl;
   os << "d_weight_id =   " << d_weight_id << endl;
   os << "d_nghosts =   " << d_nghosts << endl;

   os << "d_lambda =   " << d_lambda << endl;
   os << "d_input_dt = " << d_input_dt << endl;
   
   os << "d_current_time =   " << d_current_time << endl;
   os << "d_new_time = " << d_new_time << endl;
   os << "d_current_dt =   " << d_current_dt << endl;

   os << "d_max_precond_its =   " << d_max_precond_its << endl;
   os << "d_precond_tol = " << d_precond_tol << endl;

}

void ModifiedBratuProblem::getLevelEdges(hier::BoxList<NDIM>& boxes, 
                                         tbox::Pointer<hier::Patch<NDIM> > patch, 
                                         tbox::Pointer<hier::PatchLevel<NDIM> > level, 
                                         const int dim, 
                                         const int face)
{

  /* 
   * Shift the box associated with the patch on the side indicated by
   * (dim,face).
   */

   boxes.clearItems();
   hier::Box<NDIM> box = patch->getBox();
   hier::Box<NDIM> boundary = box;
   if (face == 0) {
      boundary.lower(dim) = box.lower(dim) - 1;
      boundary.upper(dim) = box.lower(dim) - 1;
   } else {
      boundary.lower(dim) = box.upper(dim) + 1;
      boundary.upper(dim) = box.upper(dim) + 1;
   }

   /* 
    * Initialize the list of boxes that constitute the edge of the
    * level, and delete the portion of the box that lies interior to
    * the level.
    */

   boxes.unionBoxes(boundary);
   boxes.removeIntersections(level->getBoxes());

   /* 
    * Finally delete the part of the level edge that meets the
    * physical boundary.
    */

   tbox::Pointer<geom::CartesianPatchGeometry<NDIM> > geometry = patch->getPatchGeometry();
   tbox::Array<hier::BoundaryBox<NDIM> > boundary_boxes = geometry->getCodimensionBoundary(1);
   for (int i=0; i < boundary_boxes.getSize(); i++) {
      boxes.removeIntersections(boundary_boxes[i].getBox());
   }

   /*
    * Put the result into canonical order.
    */

   boxes.simplifyBoxes();
}

void ModifiedBratuProblem::correctLevelFlux(tbox::Pointer<hier::PatchLevel<NDIM> > level)
{
   for (hier::PatchLevel<NDIM>::Iterator p(level); p; p++) {
      tbox::Pointer<hier::Patch<NDIM> > patch = level->getPatch(p());
      const hier::Box<NDIM> box = patch->getBox();
      tbox::Pointer< pdat::SideData<NDIM,double> > flux_data = patch->getPatchData(d_flux_id);

      /*
       * For each dimension, for each side:  compute the index space that
       * describes the part of the patch that lies on the outer edge of
       * the refinement level.  Then, scale the fluxes along that edge
       * to account for the different grid spacing when CONSTANT_REFINE
       * is used to fill the ghost cells.
       */

      for (int d = 0; d < NDIM; d++) {
         for (int s = 0; s <= 1; s++) {
            hier::Index<NDIM> delta(0);
            delta(d) = ( (s == 0) ? 1 : -1 );
            hier::Index<NDIM> twodelta(0);
            twodelta(d) = ( (s == 0) ? 2 : -2 ); 
            hier::BoxList<NDIM> level_edges;
            getLevelEdges(level_edges, patch, level, d, s);
            for (hier::BoxList<NDIM>::Iterator l(level_edges); l; l++) {
               for (pdat::CellIterator<NDIM> ic(l()); ic; ic++) {
                  pdat::SideIndex<NDIM> iside(ic() + delta, d, s);
                  (*flux_data)(iside) = 2.0*(*flux_data)(iside)/3.0;
               }   // cell loop
            }   // box loop
         }   // side loop
      }   // dimension loop
   }   // patch loop
}

void ModifiedBratuProblem::correctPatchFlux(
                      tbox::Pointer<hier::PatchLevel<NDIM> > level, 
                      tbox::Pointer<hier::Patch<NDIM> > patch, 
                      tbox::Pointer< pdat::CellData<NDIM,double> > u )
{
   const hier::Box<NDIM> box = patch->getBox();
   tbox::Pointer< pdat::SideData<NDIM,double> > flux_data = patch->getPatchData(d_flux_id);
   const tbox::Pointer<geom::CartesianPatchGeometry<NDIM> > geometry =  patch->getPatchGeometry();
   const double* dx = geometry->getDx();

   for (int d = 0; d < NDIM; d++) {
      for (int s = 0; s <= 1; s++) {
         hier::Index<NDIM> delta1(0);
         delta1(d) = ( (s == 0) ? 1 : -1 );
         hier::Index<NDIM> delta2(0);
         delta2(d) = ( (s == 0) ? 2 : -2 );
         double factor = ( (s == 0) ? 1.0/dx[d] : -1.0/dx[d] );
         hier::BoxList<NDIM> level_edges;
         getLevelEdges(level_edges, patch, level, d, s);
         for (hier::BoxList<NDIM>::Iterator l(level_edges); l; l++) {
            for (pdat::CellIterator<NDIM> ic(l()); ic; ic++) {
               pdat::SideIndex<NDIM> iside(ic() + delta1, d, s);
               (*flux_data)(iside) = factor*( - 8.0*(*u)(ic())/15.0
                                              + (*u)(ic()+delta1)/3.0
                                              + (*u)(ic()+delta2)/5.0 );
            } // cell loop
         } // box loop
      } // side loop
   } // dimension loop
}

#endif
