//
// File:	HierarchyDataOpsManager.h
// Package:	SAMRAI mathops
// Copyright:	(c) 1997-2005 The Regents of the University of California
// Revision:	$Revision: 601 $
// Modified:	$Date: 2005-09-06 11:23:15 -0700 (Tue, 06 Sep 2005) $
// Description:	Singleton manager for hierarchy data operation objects.
//

#ifndef included_math_HierarchyDataOpsManager
#define included_math_HierarchyDataOpsManager

#ifndef included_SAMRAI_config
#include "SAMRAI_config.h"
#endif
#ifdef HAVE_DCOMPLEX
#ifndef included_math_HierarchyDataOpsComplex
#include "HierarchyDataOpsComplex.h"
#endif
#endif
#ifndef included_math_HierarchyDataOpsInteger
#include "HierarchyDataOpsInteger.h"
#endif
#ifndef included_math_HierarchyDataOpsReal
#include "HierarchyDataOpsReal.h"
#endif
#ifndef included_hier_PatchHierarchy
#include "PatchHierarchy.h"
#endif
#ifndef included_hier_Variable
#include "Variable.h"
#endif
#ifndef included_tbox_Array
#include "tbox/Array.h"
#endif
#ifndef included_tbox_Pointer
#include "tbox/Pointer.h"
#endif

namespace SAMRAI {
    namespace math {

/**
 * Class HierarchyDataOpsManager<DIM> is a Singleton class that serves as
 * a single point of control for gaining access to operation objects which
 * to treat patch data over multiple levels in an AMR hierarchy. Specifically, 
 * this manager returns a pointer to a hierarchy data operation object given 
 * a variable and a hierarchy.  Currently, this manager class supports patch 
 * data operation objects for cell-, face-, and node-centered data of type 
 * double, float, integer and dcomplex.  Although it may be used to manage
 * data operations for multiple patch hierarchies, it supports only one data 
 * operations per variable centering and data type per hierarchy.  The manager 
 * will create and return a default hierarchy operation object when queried 
 * for an operation object if the operation does not already exist.  If a
 * matching operation object exists and is known to the manager, that operation
 * object will be returned.  This manager class may be extended using class 
 * inheritance and implementing a manager subclass.  So, for example, if
 * more than one operation object per variable type is needed on a hierarchy,
 * such inheritance should be used to add this capability to the manager.
 * 
 * Important note: This manager class is incomplete. Specifically, one cannot
 *                 set an operation object to over-ride the default .  This 
 *                 feature will be added in the near future.  
 *
 * Important note: If the manager must construct a new hierarchy data
 *                 operation object, the range of levels used in the
 *                 operation object must be set explicitly before the 
 *                 operations can be used.
 *
 * See the Design Patterns book by Gamma {\em et al.} for more information
 * about the singleton pattern.
 *
 * @see math::HierarchyDataOpsComplex
 * @see math::HierarchyDataOpsInteger
 * @see math::HierarchyDataOpsReal
 */

template<int DIM>
class HierarchyDataOpsManager
{
public:
   /**
    * Return a pointer to the single instance of the patch data operation 
    * manager.  All access to the HierarchyDataOpsManager<DIM> object is 
    * through the getManager() function.  For example, to obtain a pointer 
    * to a hierarchy data operation object appropriate for a variable of type
    * double one makes the following call: 
    * HierarchyDataOpsManager<DIM>::getManager()->getOperationsDouble(
    *                                               var, hierarchy),
    * where ``var'' is a pointer to the variable, and hierarchy is a
    * pointer to the AMR hierarchy. 
    *
    * Note that when the manager is accessed for the first time, the
    * Singleton instance is registered with the ShutdownRegistry
    * class which destroys such objects at program completion.  Thus,
    * users of this class do not explicitly allocate or deallocate the
    * Singleton instance.
    */
   static HierarchyDataOpsManager<DIM>* getManager();

   /**
    * Deallocate the HierarchyDataOpsManager instance.  It is not 
    * necessary to explicitly call freeManager() at program termination, 
    * since it is automatically called by the ShutdownRegistry class.
    */
   static void freeManager();

   //@{
   /*!
     \name Getting operators.

     If a unique operator object is not requested, and if one already
     exists for the hierarchy and variable specified, the existing one
     will be created and returned.  Otherwise, a new one is created.
     Objects created created for unique requests will not be used later
     when an equivalent request is made.
   */
   /*!
     \brief Return pointer to operation object for a double variable
     on the given hierarchy.
   */
   virtual tbox::Pointer< HierarchyDataOpsReal<DIM,double> > 
   getOperationsDouble
   ( /*! operation should correspond to this variable */
     const tbox::Pointer< hier::Variable<DIM> >& variable
   , /*! operation should correspond to this hierarchy */
     tbox::Pointer< hier::PatchHierarchy<DIM> >& hierarchy
   , /*! Whether a unique operator is requested */
     bool get_unique=false
   );

#ifdef HAVE_FLOAT
   /*!
     \brief Return pointer to operation object for a float variable
     on the given hierarchy.
   */
   virtual tbox::Pointer< HierarchyDataOpsReal<DIM,float> > 
   getOperationsFloat
   ( /*! operation should correspond to this variable */
     const tbox::Pointer< hier::Variable<DIM> >& variable
   , /*! operation should correspond to this hierarchy */
     tbox::Pointer< hier::PatchHierarchy<DIM> >& hierarchy
   , /*! Whether a unique operator is requested */
     bool get_unique=false
   );
#endif

#ifdef HAVE_DCOMPLEX
   /*!
     \brief Return pointer to operation object for a complex variable
     on the given hierarchy.
   */
   virtual tbox::Pointer< HierarchyDataOpsComplex<DIM> > 
   getOperationsComplex
   ( /*! operation should correspond to this variable */
     const tbox::Pointer< hier::Variable<DIM> >& variable
   , /*! operation should correspond to this hierarchy */
     tbox::Pointer< hier::PatchHierarchy<DIM> >& hierarchy
   , /*! Whether a unique operator is requested */
     bool get_unique=false
   );
#endif

   /*!
     \brief Return pointer to operation object for a integer variable
     on the given hierarchy.
   */
   virtual tbox::Pointer< HierarchyDataOpsInteger<DIM> > 
   getOperationsInteger
   ( /*! operation should correspond to this variable */
     const tbox::Pointer< hier::Variable<DIM> >& variable
   , /*! operation should correspond to this hierarchy */
     tbox::Pointer< hier::PatchHierarchy<DIM> >& hierarchy
   , /*! Whether a unique operator is requested */
     bool get_unique=false
   );
   //@}


protected:
   /**
    * The constructor for HierarchyDataOpsManager<DIM> is protected.  
    * Consistent with the definition of a Singleton class, only subclasses 
    * have access to the constructor for the class.
    */
   HierarchyDataOpsManager();

   /**
    * The destructor for HierarchyDataOpsManager<DIM> is protected.  
    * See the comments for the constructor.
    */
   virtual ~HierarchyDataOpsManager<DIM>();

   /**
    * Initialize Singleton instance with instance of subclass.  This function
    * is used to make the singleton object unique when inheriting from this
    * base class.
    */
   void registerSingletonSubclassInstance(
      HierarchyDataOpsManager<DIM>* subclass_instance);

private:
   static HierarchyDataOpsManager<DIM>* s_pdat_op_manager_instance;
   static bool s_registered_callback;

   //@{ \name Operations for data of various types.
   tbox::Array< tbox::Pointer< HierarchyDataOpsReal<DIM,double> > > 
      d_cell_ops_double;
   tbox::Array< tbox::Pointer< HierarchyDataOpsReal<DIM,double> > >
      d_face_ops_double;
   tbox::Array< tbox::Pointer< HierarchyDataOpsReal<DIM,double> > >
      d_node_ops_double;
   tbox::Array< tbox::Pointer< HierarchyDataOpsReal<DIM,double> > >
      d_side_ops_double;
   tbox::Array< tbox::Pointer< HierarchyDataOpsReal<DIM,double> > >
      d_edge_ops_double;

#ifdef HAVE_FLOAT
   tbox::Array< tbox::Pointer< HierarchyDataOpsReal<DIM,float> > >
      d_cell_ops_float;
   tbox::Array< tbox::Pointer< HierarchyDataOpsReal<DIM,float> > >
      d_face_ops_float;
   tbox::Array< tbox::Pointer< HierarchyDataOpsReal<DIM,float> > >
      d_side_ops_float;
   tbox::Array< tbox::Pointer< HierarchyDataOpsReal<DIM,float> > >
      d_node_ops_float;
   tbox::Array< tbox::Pointer< HierarchyDataOpsReal<DIM,float> > >
      d_edge_ops_float;
#endif

#ifdef HAVE_DCOMPLEX
   tbox::Array< tbox::Pointer< HierarchyDataOpsComplex<DIM> > >
      d_cell_ops_complex;
   tbox::Array< tbox::Pointer< HierarchyDataOpsComplex<DIM> > >
      d_face_ops_complex;
   tbox::Array< tbox::Pointer< HierarchyDataOpsComplex<DIM> > >
      d_side_ops_complex;
   tbox::Array< tbox::Pointer< HierarchyDataOpsComplex<DIM> > >
      d_node_ops_complex;
   tbox::Array< tbox::Pointer< HierarchyDataOpsComplex<DIM> > >
      d_edge_ops_complex;
#endif

   tbox::Array< tbox::Pointer< HierarchyDataOpsInteger<DIM> > >
      d_cell_ops_int;
   tbox::Array< tbox::Pointer< HierarchyDataOpsInteger<DIM> > >
      d_face_ops_int;
   tbox::Array< tbox::Pointer< HierarchyDataOpsInteger<DIM> > >
      d_side_ops_int;
   tbox::Array< tbox::Pointer< HierarchyDataOpsInteger<DIM> > >
      d_node_ops_int;
   tbox::Array< tbox::Pointer< HierarchyDataOpsInteger<DIM> > >
      d_edge_ops_int;
   //@}

};

}
}
#endif

#ifdef INCLUDE_TEMPLATE_IMPLEMENTATION
#include "HierarchyDataOpsManager.C"
#endif
