//
// File:        TimeRefinementIntegrator.C
// Package:     SAMRAI algorithms
// Copyright:   (c) 1997-2005 The Regents of the University of California
// Revision:    $Revision: 173 $
// Modified:    $Date: 2005-01-19 09:09:04 -0800 (Wed, 19 Jan 2005) $
// Description: Time integration manager for AMR with local time stepping.
//

#ifndef included_algs_TimeRefinementIntegrator_C
#define included_algs_TimeRefinementIntegrator_C

#include "TimeRefinementIntegrator.h"
#include <stdlib.h>
#include <fstream>
using namespace std;
#ifdef DEBUG_CHECK_ASSERTIONS
#ifndef included_assert
#define included_assert
#include <assert.h>
#endif
#endif
#include "tbox/IEEE.h"
#include "tbox/PIO.h"
#include "tbox/RestartManager.h"
#include "tbox/TimerManager.h"
#include "tbox/Timer.h"
#include "tbox/Utilities.h"

#define DEBUG_TIMES
//#undef DEBUG_TIMES

#ifndef NULL
#define NULL (0)
#endif

#define ALGS_TIME_REFINEMENT_INTEGRATOR_VERSION (2)

#ifdef DEBUG_NO_INLINE
#include "TimeRefinementIntegrator.I"
#endif

namespace SAMRAI {
    namespace algs {

/*
*************************************************************************
*                                                                       *
* The constructor for TimeRefinementIntegrator<DIM> checks for valid   *
* input data, initializes time stepping data to undefined values,       *
* and forces certain parameters in the level strategy, regridding       *
* algorithm to be consistent with this object's data members.  This     *
* constructor also invokes the variable registration process in the     *
* level strategy object.                                                *
*                                                                       *
*************************************************************************
*/

template<int DIM> TimeRefinementIntegrator<DIM>::TimeRefinementIntegrator(
   const string& object_name,
   tbox::Pointer<tbox::Database> input_db,
   tbox::Pointer< hier::BasePatchHierarchy<DIM> > hierarchy,
   TimeRefinementLevelStrategy<DIM>* level_integrator,
   tbox::Pointer< mesh::BaseGriddingAlgorithm<DIM> > gridding_algorithm,
   bool register_for_restart)
{
#ifdef DEBUG_CHECK_ASSERTIONS
   assert(!object_name.empty());
   assert(!input_db.isNull());
   assert(!hierarchy.isNull());
   assert(level_integrator != ((TimeRefinementLevelStrategy<DIM>*)NULL));
   assert(!gridding_algorithm.isNull());
#endif

   t_initialize_hier = tbox::TimerManager::getManager()-> 
      getTimer("algs::TimeRefinementIntegrator::initializeHierarchy()");
   t_advance_hier = tbox::TimerManager::getManager()->
      getTimer("algs::TimeRefinementIntegrator::advanceHierarchy()");


   d_object_name = object_name;
   d_registered_for_restart = register_for_restart;

   if (d_registered_for_restart) {
      tbox::RestartManager::getManager()->
         registerRestartItem(d_object_name, this);
   }
  
   d_patch_hierarchy = hierarchy;
   d_refine_level_integrator = level_integrator;
   d_gridding_algorithm = gridding_algorithm;

   d_use_refined_timestepping = level_integrator->usingRefinedTimestepping();

   /*
    * Set default values for debugging.
    */

   d_start_time = tbox::IEEE::getSignalingNaN();
   d_end_time = tbox::IEEE::getSignalingNaN();
   d_integrator_time = tbox::IEEE::getSignalingNaN();
   d_integrator_step = tbox::IEEE::getINT_MAX();
   d_grow_dt = 1.0;
   d_max_integrator_steps = tbox::IEEE::getINT_MAX();

   const int max_levels = d_gridding_algorithm->getMaxLevels();

   d_regrid_interval.resizeArray(max_levels);

   d_level_old_old_time.resizeArray(max_levels);
   d_level_old_time.resizeArray(max_levels);
   d_level_sim_time.resizeArray(max_levels);
   d_dt_max_level.resizeArray(max_levels);
   d_dt_actual_level.resizeArray(max_levels);
   d_step_level.resizeArray(max_levels);
   d_max_steps_level.resizeArray(max_levels);

   int level_number;

   for (level_number=0; level_number < max_levels; level_number++) {
      d_regrid_interval[level_number] = tbox::IEEE::getINT_MAX();
      d_level_old_old_time[level_number] = tbox::IEEE::getSignalingNaN();
      d_level_old_time[level_number] = tbox::IEEE::getSignalingNaN();
      d_level_sim_time[level_number] = tbox::IEEE::getSignalingNaN();
      d_dt_max_level[level_number] = tbox::IEEE::getSignalingNaN();
      d_dt_actual_level[level_number] = tbox::IEEE::getSignalingNaN();
      d_step_level[level_number] = tbox::IEEE::getINT_MAX();
      d_max_steps_level[level_number] = tbox::IEEE::getINT_MAX();
   }

   /*
    * Set regrid interval data based on ratios between levels or input value.
    */

   if (d_use_refined_timestepping) {
      if (max_levels > 1) {

         for (level_number = 1; level_number < max_levels; level_number++) {
            const hier::IntVector<DIM> ratio(d_gridding_algorithm->
                                        getRatioToCoarserLevel(level_number));

            if ( ((ratio.max() % 2) == 0) || (ratio.max() == 1) ) {
               d_regrid_interval[level_number] = 2;
            } else {
               if ((ratio.max() % 3) == 0) {
                  d_regrid_interval[level_number] = 3;
               } else {
                  TBOX_ERROR(d_object_name << ":  "
                             << " integrator cannot set regrid interval"
                             << " based on ratios between levels.");
               }
            }
         }

         d_regrid_interval[0] = d_regrid_interval[1];

      } else {
         for (level_number = 0; level_number < max_levels; level_number++) {
            d_regrid_interval[level_number] = 1;
         }
      }
   } else {
      int regrid_interval =
         input_db->getIntegerWithDefault("regrid_interval",1);

      setRegridInterval(regrid_interval);
   }

   /*
    * Initialize this time refinement integration object with data read 
    * from input and restart databases.
    */
   bool is_from_restart = tbox::RestartManager::getManager()->isFromRestart();
   if (is_from_restart) {
      getFromRestart(); 
   }
   getFromInput(input_db,is_from_restart);

   /*
    * Initialize remaining integrator data members.
    */

   d_just_regridded = false; 

   if (!is_from_restart) {
      d_integrator_time = d_start_time;
      d_integrator_step = 0;
      d_last_finest_level = 0;
   }

}

/*
*************************************************************************
*                                                                       *
* Destructor tells tbox::RestartManager to remove this object from the   *
* list of restart items.                                                *
*                                                                       *
*************************************************************************
*/

template<int DIM> TimeRefinementIntegrator<DIM>::~TimeRefinementIntegrator() 
{
   if (d_registered_for_restart) {
      tbox::RestartManager::getManager()->unregisterRestartItem(d_object_name);
   }
}

/*
*************************************************************************
*                                                                       *
* Create patch level and patches for coarsest level in AMR hierarchy    *
* at initial simulation time.  Then, set time stepping data and set     *
* initial patch data on the coarsest level.  The actual level data      *
* initialization process is invoked by the recursive private member     *
* function initialize().  This function will create and initialize      * 
* successively finer hierarchy levels until either the maximum number   *
* allowable level is reached or no further refinement is needed.        *
*                                                                       *
*************************************************************************
*/

template<int DIM> double TimeRefinementIntegrator<DIM>::initializeHierarchy()
{

   t_initialize_hier->start();

   d_refine_level_integrator->initializeLevelIntegrator(d_gridding_algorithm);

   if (tbox::RestartManager::getManager()->isFromRestart()) {

      d_patch_hierarchy->getFromRestart(d_gridding_algorithm->getMaxLevels());
   
      d_gridding_algorithm->
         getTagAndInitializeStrategy()->
         resetHierarchyConfiguration(d_patch_hierarchy,
                                     0,
                                     d_patch_hierarchy->
                                     getFinestLevelNumber());

   } else {

      d_gridding_algorithm->makeCoarsestLevel(d_patch_hierarchy,
                                              d_start_time);

      if (d_use_refined_timestepping) {
         initializeRefinedTimesteppingLevelData(0);
      } else {
         initializeSynchronizedTimesteppingLevelData(0);
      }
   
      /*
       * After data on each level is initialized at simulation start time,
       * coarser levels are synchronized with finer levels that didn't exist
       * when the coarser level initial data was set.  This synchronization
       * process is defined by the integration algorithm.
       */
      if (d_patch_hierarchy->getFinestLevelNumber() > 0) {
         // "true" argument: const bool initial_time = true;
         d_refine_level_integrator->synchronizeNewLevels(d_patch_hierarchy,
                                                         0,
                                                         d_patch_hierarchy->
                                                         getFinestLevelNumber(),
                                                         d_start_time,
                                                         true);
      }
   }

   t_initialize_hier->stop();

   return(d_dt_max_level[0]);

}

/*
*************************************************************************
*                                                                       *
* Advance all levels in hierarchy through specified time interval dt    * 
* by invoking the recursive timestepping process at the coarsest        *
* hierarchy level (level 0).  If synchronized timestepping is used,     *
* then loop through each level advance and then synchronize all levels. *
* The return value is the proper time increment for a subsequent        * 
* advance of level 0.                                                   *
*                                                                       *
*************************************************************************
*/

template<int DIM> double TimeRefinementIntegrator<DIM>::advanceHierarchy(
   const double dt,
   const bool rebalance_coarsest)
{
#ifdef DEBUG_CHECK_ASSERTIONS
   assert(dt >= 0.0);
#endif

   t_advance_hier->start();

   d_level_sim_time[0] = d_integrator_time;

   if (rebalance_coarsest) {
      d_gridding_algorithm->makeCoarsestLevel(d_patch_hierarchy,
                                              d_level_sim_time[0]);
   }

   double dt_new; 

   if (!d_use_refined_timestepping) {

      dt_new = advanceForSynchronizedTimestepping(d_level_sim_time[0]+dt); 

      if (d_integrator_time + dt_new > d_end_time) {
         dt_new = d_end_time - d_integrator_time;
      } 
      
   } else { 
      advanceRecursivelyForRefinedTimestepping(0, d_level_sim_time[0]+dt);
      d_integrator_time += dt;
      dt_new = tbox::Utilities::dmin( d_dt_actual_level[0],
                                     d_end_time-d_integrator_time );
   }

   t_advance_hier->stop();

   return dt_new;
}

/*
*************************************************************************
*                                                                       *
* Initialize data on given level.  Then, invoke regridding and          *
* recursively initialize finer levels, if necessary.  The coarsest      *
* level (i.e., level 0) is created by initializeHierarchy().  Each      *
* finer level is constructed by calls this function.  The gridding      *
* algorithm data member build finer each level and its patches.         *
*                                                                       *
* This function assumes that the patch level and patches exist before   *
* it is called.  Upon leaving this routine, initial data is set         *
* according to the methods in the level integration routines, which     *
* are generally invoked from the gridding algorithm class.  Also,       *
* basic time increment data is set for each level.                      *
*                                                                       *
*************************************************************************
*/

template<int DIM> void
TimeRefinementIntegrator<DIM>::initializeRefinedTimesteppingLevelData(
   const int level_number)
{
#ifdef DEBUG_CHECK_ASSERTIONS
   assert( (level_number >= 0) &&
           (level_number <= d_patch_hierarchy->getFinestLevelNumber()) );
#endif

   const tbox::Pointer< hier::BasePatchLevel<DIM> > 
      patch_level = d_patch_hierarchy->getPatchLevel(level_number); 

   /* 
    * Initialize step count and start time for current level.
    * Then, query level strategy for proper time step. 
    */

   d_step_level[level_number] = 0;
   d_max_steps_level[level_number] = 1; 
   d_level_sim_time[level_number] = d_start_time;

   bool initial_time = true;
   double dt_level;

   dt_level = d_refine_level_integrator->getLevelDt(patch_level,
                                            d_level_sim_time[level_number],
                                            initial_time);

   /*
    * If a coarser level exists, we require that the time increment on
    * the current level be consistent with the needs of the numerical 
    * integration algorithm (i.e., the level strategy).  Futhermore, 
    * if the current level can be refined and the level is not the coarsest
    * in the hierarchy, we require that the time increment be adjusted so 
    * that the number of steps taken on the level before synchronization 
    * with the next coarser level is an integer multiple of the grid 
    * refinement ratio between the two levels.
    */

   if ( level_number > 0 ) {
      dt_level = 
      tbox::Utilities::dmin( dt_level, 
                            d_refine_level_integrator->getMaxFinerLevelDt(
                               level_number,
                               d_dt_actual_level[level_number-1],
                               d_gridding_algorithm->
                               getRatioToCoarserLevel(level_number)) );
   }
 
   d_dt_max_level[level_number] = d_dt_actual_level[level_number] = dt_level;

   /*
    * Create a finer level if appropriate.
    */

   if ( d_gridding_algorithm->levelCanBeRefined(level_number) ) {
 
      int buffer;
      if ( d_gridding_algorithm->errorEstimationUsesTimeIntegration() ) {
         buffer = d_regrid_interval[level_number]; 
      } else {
         buffer = d_tag_buffer[level_number];
      }

      double regrid_start_time = 
         d_level_sim_time[level_number] - d_dt_actual_level[level_number];
      // "true" argument: const bool initial_time = true;
      d_gridding_algorithm->makeFinerLevel(d_patch_hierarchy,
                                           d_level_sim_time[level_number],
                                           true,
                                           buffer,
                                           regrid_start_time);
       
      /*
       * If new finer level is made, data on its patches is initialized.
       * Also, if new level can be refined and time integration is used
       * during regridding process, data on the current level is advanced
       * through a single time increment to provide boundary data for the
       * regridding process on finer levels.
       */

      if ( d_patch_hierarchy->finerLevelExists(level_number) ) {

         /*
          * If time integration is used in the refinement process and
          * the newly created finer level can be further refined,
          * data is advanced on the current level through one time 
          * increment to provide time interpolated boundary data for
          * regridding on the finer level.
          */ 
         if ( d_gridding_algorithm->levelCanBeRefined(level_number+1) &&
              d_gridding_algorithm->errorEstimationUsesTimeIntegration() ) {

            // "false" argument: bool last_step = false;
            // "true" argument: bool regrid_advance = true;
            d_refine_level_integrator->
               advanceLevel(patch_level,
                            d_patch_hierarchy,
                            d_level_sim_time[level_number],
                            d_level_sim_time[level_number]
                            + d_dt_actual_level[level_number],
                            firstLevelStep(level_number),
                            false,
                            true);
         }
    
         /* 
          * RECURSIVE invocation of initialization on next finer level.
          */

         initializeRefinedTimesteppingLevelData(level_number+1);
       
         /*
          * If current level patch data was advanced above, all new
          * data is discarded so that advance routines all start with the
          * same patch data on each level, and because AMR synchronization 
          * may require a different initial time increment size.
          */

         if ( d_gridding_algorithm->levelCanBeRefined(level_number+1) &&
              d_gridding_algorithm->errorEstimationUsesTimeIntegration() ) {
            d_refine_level_integrator->
               resetDataToPreadvanceState(patch_level);
         }
      }
   }
}

template<int DIM> void
TimeRefinementIntegrator<DIM>::initializeSynchronizedTimesteppingLevelData(
   const int level_number)
{
#ifdef DEBUG_CHECK_ASSERTIONS
   assert( (level_number >= 0) &&
           (level_number <= d_patch_hierarchy->getFinestLevelNumber()) );
#endif

   const tbox::Pointer< hier::BasePatchLevel<DIM> > 
      patch_level = d_patch_hierarchy->getPatchLevel(level_number); 

   /* 
    * Initialize step count and start time for current level.
    * Then, query level strategy for proper time step. 
    */

   d_step_level[level_number] = 0;
   d_max_steps_level[level_number] = 1; 
   d_level_sim_time[level_number] = d_start_time;
   d_level_old_time[level_number] = d_start_time;

   bool initial_time = true;
   double dt_level;

   dt_level = d_refine_level_integrator->getLevelDt(patch_level,
                                             d_level_sim_time[level_number],
                                             initial_time);

   if (level_number == 0) {
      d_dt = dt_level;
   } else {
      d_dt = tbox::Utilities::dmin(d_dt, dt_level);
   }
   d_dt_max_level[level_number] = d_dt_actual_level[level_number] = d_dt;

   for (int i = 0; i < level_number; i++) {
      if (d_dt_max_level[i] > d_dt_max_level[level_number]) {
         d_dt_max_level[i] = d_dt_actual_level[i] = 
            d_dt_max_level[level_number];
      }
   }

   /*
    * Create a finer level if appropriate.
    */

   if ( d_gridding_algorithm->levelCanBeRefined(level_number) ) {
 
      int buffer;
      if ( d_gridding_algorithm->errorEstimationUsesTimeIntegration() ) {
         buffer = d_regrid_interval[level_number]; 
      } else {
         buffer = d_tag_buffer[level_number];
      }

      double regrid_start_time = 
         d_level_sim_time[level_number] - d_dt_actual_level[level_number];
      // "true" argument: const bool initial_time = true;
      d_gridding_algorithm->makeFinerLevel(d_patch_hierarchy,
                                           d_level_sim_time[level_number],
                                           true,
                                           buffer,
                                           regrid_start_time);
       
      /*
       * If new finer level is made, data on its patches is initialized.
       * Also, if new level can be refined and time integration is used
       * during regridding process, data on the current level is advanced
       * through a single time increment to provide boundary data for the
       * regridding process on finer levels.
       */

      if ( d_patch_hierarchy->finerLevelExists(level_number) ) {

         /*
          * If time integration is used in the refinement process and
          * the newly created finer level can be further refined,
          * data is advanced on the current level through one time 
          * increment to provide time interpolated boundary data for
          * regridding on the finer level.
          */ 
         if ( d_gridding_algorithm->levelCanBeRefined(level_number+1) &&
              d_gridding_algorithm->errorEstimationUsesTimeIntegration() ) {

            // "false" argument: bool last_step = false;
            // "true" argument: bool regrid_advance = true;
            d_refine_level_integrator->
               advanceLevel(patch_level,
                            d_patch_hierarchy,
                            d_level_sim_time[level_number],
                            d_level_sim_time[level_number]
                            + d_dt_actual_level[level_number],
                            firstLevelStep(level_number),
                            false,
                            true);
         }
    
         /* 
          * RECURSIVE invocation of initialization on next finer level.
          */

         initializeSynchronizedTimesteppingLevelData(level_number+1);
       
         /*
          * If current level patch data was advanced above, all new
          * data is discarded so that advance routines all start with the
          * same patch data on each level, and because AMR synchronization 
          * may require a different initial time increment size.
          */
         if ( d_gridding_algorithm->levelCanBeRefined(level_number+1) &&
              d_gridding_algorithm->errorEstimationUsesTimeIntegration() ) {
            d_refine_level_integrator->
               resetDataToPreadvanceState(patch_level);
         }
      }
   }
}

/*
*************************************************************************
*                                                                       *
* Advance data on level to specified time.  Then, advance each finer    *
* hierarchy level to the same end time using recursive function calls.  *
* It is assumed that when this function is called, only the data needed *
* for initialization exists on the level.  Also, the solutions on       *
* specified level and all finer levels are synchronized at the the      *
* current simulation time on level level_number-1.  The integration     *
* process implemented in this function is outlined as follows:          *
*                                                                       *
*    1) Initialize timestep count, simulation time, time step increment *
*       for level level_number.                                         *
*    2) Adjust time first time increment if necessary and estimate      *
*       number of timesteps required to advance the level to end time.  *
*    3) Iterate over sequence of timesteps for level level_number:      *
*       a) Set end time for current step, record current time for       *
*          synchronization.                                             *
*       b) Advance solution on level using given time increment.        *
*          Note that the level strategy performs the advance.           * 
*       c) Increment step count information.                            *
*       d) If finer level exists, advance finer level to end time of    *
*          most recent time step using recursive function call.         *
*       e) Synchronize data between levels as necessary.  Note that     *
*          new solution on level level_number as well as each finer     *
*          level corresponds to the same time.  Also, synchronization   *
*          involves several levels, typically, including all levels     *
*          finer than level_number, level level_number itself, and the  *
*          possibly level level_number-1.  Which levels synchronize     *
*          here depends on when regridding occurs.  Note that the       *
*          level strategy performs the data synchronization.            *
*       f) Increment the simulation time and determine the size of the  *
*          next timestep for the level and estimate the number of       *
*          timesteps needed to advance the level to end time.           *
*       g) If appropriate, regrid all finer levels.                     *
*          Note that the gridding algorithm regrids the mesh.           *
*       h) Synchronize all levels involved in the regridding process,   *
*          if it occurred.  The sync is performed by level strategy.    *
*       i) If regridding did not occur, then reset current and new      *
*          solution data on the level.                                  *
*                                                                       *
*************************************************************************
*/

template<int DIM> void TimeRefinementIntegrator<DIM>::advanceRecursivelyForRefinedTimestepping(
   const int level_number, 
   const double end_time)
{
#ifdef DEBUG_CHECK_ASSERTIONS
   assert( (level_number >= 0) &&
           (level_number <= d_patch_hierarchy->getFinestLevelNumber()) );
   assert(end_time >= d_integrator_time);
#endif

   const tbox::Pointer< hier::BasePatchLevel<DIM> > 
      patch_level = d_patch_hierarchy->getPatchLevel(level_number);

   /*
    * Initialize step count, start time for current level. 
    * Determine time remaining on level and, if necessary, set
    * the max time increment for the level.
    */

   d_step_level[level_number] = 0;
   d_max_steps_level[level_number] = 1;
   double time_remaining = 0.0; 

   if (level_number > 0) {
      d_level_sim_time[level_number] = d_level_sim_time[level_number-1];
      time_remaining = end_time - d_level_sim_time[level_number];

      /*
       * If this level was created during last regrid, compute max time 
       * increment for level. 
       */

      if ( d_last_finest_level < level_number ) {
         d_dt_max_level[level_number] = 
         tbox::Utilities::dmin( d_dt_actual_level[level_number-1],
                               d_refine_level_integrator->getMaxFinerLevelDt(
                               level_number,
                               d_dt_actual_level[level_number-1],
                               d_gridding_algorithm->
                               getRatioToCoarserLevel(level_number)) );

         if ( d_gridding_algorithm->levelCanBeRefined(level_number) ) {
            d_dt_max_level[level_number] = 
            tbox::Utilities::dmin( d_dt_max_level[level_number],
                                  time_remaining/
                                  double(d_regrid_interval[level_number]) );
         }
         d_last_finest_level = level_number;
      }

   } else {
      time_remaining = end_time - d_level_sim_time[level_number];
   }

   /*
    * Determine time increment for first step in integration sequence,
    * and estimate the number of steps before synchronization with the
    * next coarser level occurs.
    */

   bool sync_after_step = 
   findNextDtAndStepsRemaining(level_number, 
                               time_remaining, 
                               d_dt_max_level[level_number]);

   /* 
    * Loop over a dynamically determined sequence of timesteps on the 
    * current level (level_number).  Note that if level is coarsest in 
    * AMR hierarchy, we must determine whether there will only be a 
    * single advance step before the function returns.
    */

   double new_level_time, dt_new;
   while ( !lastLevelStep(level_number) ) {

      /*
       * Determine new time after this step, set old time for synchronization.
       */

      new_level_time = (  sync_after_step ? end_time
                        : d_level_sim_time[level_number] 
                          + d_dt_actual_level[level_number] );
      d_level_old_old_time[level_number] = d_level_old_time[level_number];
      d_level_old_time[level_number] = d_level_sim_time[level_number];
      d_just_regridded = false;

#ifdef DEBUG_TIMES
      tbox::plog << "\nAdvancing level number = " << level_number << endl;
      tbox::plog << "step number = " << d_step_level[level_number] << endl;
      tbox::plog << "max steps = " << d_max_steps_level[level_number] << endl;
      tbox::plog << "current time = " << d_level_sim_time[level_number] << endl;
      tbox::plog << "dt used = " << d_dt_actual_level[level_number] << endl;
      tbox::plog << "new level time = " << new_level_time << endl;
      tbox::plog << "dt max = " << d_dt_max_level[level_number] << endl;
      tbox::plog << "end time = " << end_time << endl;
      tbox::plog << "sync_after_step = " << sync_after_step << endl;
#endif

      /*
       * Advance level from current simulation time to new_level_time 
       * using a single time advance step.  Note that the level strategy
       * returns the next time increment for the level.  Also, we keep both
       * new and previous data on level so that time interpolation can be
       * used to set boundary conditions for finer levels and for proper
       * data synchronization once all finer levels have been advanced.
       */

      // "sync_after_step" is same as "last_step" in level strategy.
      dt_new = d_refine_level_integrator->advanceLevel(patch_level,
                                                d_patch_hierarchy,
                                                d_level_sim_time[level_number], 
                                                new_level_time,
                                                firstLevelStep(level_number),
                                                sync_after_step);

      /*
       * Update step count information.  Then, advance all finer levels.
       */

      if (level_number == 0) d_integrator_step++;
      d_step_level[level_number]++;

      if (d_patch_hierarchy->finerLevelExists(level_number)) {
         advanceRecursivelyForRefinedTimestepping(level_number+1,
                                                  new_level_time);
      }

      /*
       * Synchronize data between levels in the hierarchy as necessary.
       * Note that this process synchronizes data between this level,
       * several finer levels, and the next coarser level, potentially.
       */

      int coarsest_sync_level = -1;
      int finest_level_number = d_patch_hierarchy->getFinestLevelNumber();

      if ( atRegridPoint(level_number) ) {

         if ( !lastLevelStep(level_number) 
              || !coarserLevelRegridsToo(level_number) ) {

            coarsest_sync_level = ( ( (level_number > 0)
                                       && lastLevelStep(level_number) )
                                          ? level_number-1 : level_number );

            if ( coarsest_sync_level < finest_level_number ) {
#ifdef DEBUG_TIMES
               tbox::plog << "\nSynchronizing levels " << coarsest_sync_level
                                                 << " to "
                                                 << finest_level_number << endl;
#endif
               d_refine_level_integrator->
                  standardLevelSynchronization(d_patch_hierarchy,
                                               coarsest_sync_level,
                                               finest_level_number,
                                               new_level_time,
                                               d_level_old_time);
            }

         }

      } else {

         if ( level_number < finest_level_number ) {
            if ( (!lastLevelStep(level_number) 
                 || (level_number == 0)) && !d_just_regridded ) {
#ifdef DEBUG_TIMES
               tbox::plog << "\nSynchronizing levels " << level_number << " to "
                                                 << finest_level_number << endl;
#endif
               d_refine_level_integrator->
                  standardLevelSynchronization(d_patch_hierarchy,
                                               level_number,
                                               finest_level_number,
                                               new_level_time,
                                               d_level_old_time);

               d_refine_level_integrator->
                  resetTimeDependentData(d_patch_hierarchy->
                                        getPatchLevel(finest_level_number),
                                        new_level_time,
                                        d_gridding_algorithm->
                                        levelCanBeRefined(finest_level_number));
            }
         }

      }
 
      /*
       * Update level simulation time and time remaining until 
       * synchronization with next coarser level.  Then, adjust 
       * time increment and step sequence for current level.
       */

      d_level_sim_time[level_number] = new_level_time;
      time_remaining = end_time - new_level_time; 

      sync_after_step = findNextDtAndStepsRemaining(level_number, 
                                                    time_remaining, 
                                                    dt_new);

      /*
       * All finer levels are synchronized with this level now.
       * If appropriate, we regrid finer levels and re-synchronize 
       * levels as needed.  Otherwise, we reset time-dependent data and 
       * re-synchronize levels as needed.  Note that the regridding
       * process resets the data on each level involved in the regridding.
       */

      if ( atRegridPoint(level_number) ) {

         if ( !lastLevelStep(level_number) 
              || !coarserLevelRegridsToo(level_number) ) {
#ifdef DEBUG_TIMES
              tbox::plog << "\nRegridding from level number = " 
                   << level_number << endl;
#endif
            /*
             * Reset time dependent data.  If the gridding algorithm uses
             * time integration for error estimation, it will have already
             * reset time dependent data on all levels regridded, so
             * only reset data on levels that are not regridded.  If the
             * gridding algorithm does not used time integration, reset data
             * on all levels.
             */
            if (d_gridding_algorithm->errorEstimationUsesTimeIntegration() ) {
               if ( !d_gridding_algorithm->
                         levelCanBeRefined(finest_level_number) ) {
                  d_refine_level_integrator->resetTimeDependentData(
                     d_patch_hierarchy->getPatchLevel(finest_level_number), 
                     new_level_time, 
                     d_gridding_algorithm->
                     levelCanBeRefined(finest_level_number));
               }
            } else {
               for (int ln = level_number; ln <= finest_level_number; ln++) {
                  d_refine_level_integrator->resetTimeDependentData(
                     d_patch_hierarchy->getPatchLevel(ln),
                     d_level_sim_time[ln],
                     d_gridding_algorithm->levelCanBeRefined(ln));
               }
            }

            d_last_finest_level = finest_level_number;
 
            /*
             * Regrid finer levels.  If the error estimation procedure
             * uses time integration (e.g. Richardson extrapolation) then
             * we must supply the oldest time at which data is stored.
             *
             * If the error coarsen ratio is two, data will be stored
             * from the previous timestep (at d_level_old_time).  If the 
             * error coarsen ratio is three, data will be stored
             * from two previous timesteps (at d_level_old_old_time).  
             * 
             * If we are not using time integration, the oldest time
             * information should not be used, so it is set to NaNs 
             * to throw an assertion if it is accessed.
             */

            tbox::Array<double> regrid_start_time;
            if (!d_gridding_algorithm->errorEstimationUsesTimeIntegration()) {

               int max_levels = d_gridding_algorithm->getMaxLevels();
               regrid_start_time.resizeArray(max_levels);
               for (int i = 0; i < regrid_start_time.getSize(); i++) {
                  regrid_start_time[i] = 0.;
               }
               
            } else {

               if (d_gridding_algorithm->getErrorCoarsenRatio() == 2) {
                  regrid_start_time = d_level_old_time;
               } else if (d_gridding_algorithm->getErrorCoarsenRatio() == 3) {
                  regrid_start_time = d_level_old_old_time;
               } else {
                  TBOX_ERROR(d_object_name << ": the supplied gridding "
                             << "algorithm uses an error coarsen ratio of "
                             << d_gridding_algorithm->getErrorCoarsenRatio()
                             << " which is not supported in this class" 
                             << endl);
               }

            }

            d_gridding_algorithm->
               regridAllFinerLevels(d_patch_hierarchy,
                                    level_number,
                                    d_level_sim_time[level_number],
                                    d_tag_buffer,
                                    regrid_start_time,
                                    (coarsest_sync_level>=level_number));

            d_just_regridded = true; 

            if (level_number < d_patch_hierarchy->getFinestLevelNumber()) {
#ifdef DEBUG_TIMES
               tbox::plog << "\nSynchronizing levels after regrid : "
                    << level_number << " to "
                    << d_patch_hierarchy->getFinestLevelNumber() << endl;
#endif
              
               // "false" argument: const bool initial_time = false;
               d_refine_level_integrator->
               synchronizeNewLevels(d_patch_hierarchy,
                                    level_number,
                                    d_patch_hierarchy->getFinestLevelNumber(),
                                    d_level_sim_time[level_number],
                                    false);
            }

         }

      } else {

         if ( !lastLevelStep(level_number) || (level_number == 0) ) {
            d_refine_level_integrator->resetTimeDependentData(
                                patch_level,
                                d_level_sim_time[level_number],
                                d_gridding_algorithm->
                                levelCanBeRefined(level_number));
         }

         if ( d_just_regridded ) {
#ifdef DEBUG_TIMES
            tbox::plog << "\nSynchronizing levels after regrid : "
                 << level_number << " to "
                 << d_patch_hierarchy->getFinestLevelNumber() << endl;
#endif
            // "false" argument: const bool initial_time = false;
            d_refine_level_integrator->
            synchronizeNewLevels(d_patch_hierarchy,
                                 level_number,
                                 d_patch_hierarchy->getFinestLevelNumber(),
                                 d_level_sim_time[level_number],
                                 false);
         }

      }

   }

}

/*
*************************************************************************
*                                                                       *
* Advance data on all hierarchy levels through specified time increment *
* using the same timesteps on each level.  Synchronization and          *
* regridding are performed as needed.                                   *
*                                                                       *
*************************************************************************
*/

template<int DIM> double TimeRefinementIntegrator<DIM>::advanceForSynchronizedTimestepping(
   const double end_time)
{
#ifdef DEBUG_CHECK_ASSERTIONS
   assert(end_time >= d_integrator_time);
#endif

   double dt = end_time - d_integrator_time;

   int finest_level_number = d_patch_hierarchy->getFinestLevelNumber();
   double dt_new = tbox::IEEE::getDBL_MAX();

   int level_num;
   for (level_num = 0; level_num <= finest_level_number; level_num++) {

      tbox::Pointer< hier::BasePatchLevel<DIM> > patch_level =
         d_patch_hierarchy->getPatchLevel(level_num);

      d_step_level[level_num] = 1;
      d_max_steps_level[level_num] = 1;
      d_dt_max_level[level_num] = dt;
      d_level_sim_time[level_num] = d_integrator_time;

#ifdef DEBUG_TIMES
      tbox::plog << "\nAdvancing level number = " << level_num << endl;
      tbox::plog << "step number = " << d_step_level[level_num] << endl;
      tbox::plog << "max steps = " << d_max_steps_level[level_num] << endl;
      tbox::plog << "current time = " << d_integrator_time << endl;
      tbox::plog << "dt used = " << dt << endl;
      tbox::plog << "new level time = " << d_integrator_time + dt << endl;
#endif

      // "true" argument: bool first_step = true;
      // "false" argument: bool last_step = false;
      double dt_next_level =
         d_refine_level_integrator->advanceLevel(patch_level,
                                                 d_patch_hierarchy,
                                                 d_integrator_time,
                                                 d_integrator_time + dt,
                                                 true, 
                                                 false);

      dt_new = tbox::Utilities::dmin(dt_new, dt_next_level);
   }

   dt_new = tbox::Utilities::dmin(dt_new,d_grow_dt*dt);

   for (level_num = 0; level_num <= finest_level_number; level_num++) {
      d_dt_max_level[level_num] = d_dt_actual_level[level_num] = dt_new;
   }

   d_integrator_time += dt;
   d_integrator_step++;

   int coarse_level_number = 0;


   if (finest_level_number > 0) {

#ifdef DEBUG_TIMES
      tbox::plog << "\nSynchronizing levels " << coarse_level_number
           << " to " << finest_level_number << endl;
#endif

      d_refine_level_integrator->standardLevelSynchronization(
                                 d_patch_hierarchy,
                                 0,
                                 finest_level_number,
                                 d_integrator_time,
                                 d_integrator_time - dt);
   }

   /*
    * Store the time from the previous step(s) in the "old_time"
    * array.  This information may be used during the re-gridding
    * process, if time integration is used during error estimation.
    */
   for (level_num = 0; level_num <= finest_level_number; level_num++) {
      d_level_old_old_time[level_num] = d_level_old_time[level_num];
      d_level_old_time[level_num] = d_level_sim_time[level_num];
   }

   /*
    * Are we ready to re-grid??
    */ 
   bool regrid_now = (d_integrator_step % d_regrid_interval[0] == 0);
   
   if (!regrid_now) {

      /*
       * We are not re-gridding, so all we need to do is simply
       * reset the time dependent data on all the levels to 
       * prepare for the next advance.
       */
      for (int ln = 0; ln <= finest_level_number; ln++) {
         d_refine_level_integrator->resetTimeDependentData(
            d_patch_hierarchy->getPatchLevel(ln),
            d_integrator_time,
            d_gridding_algorithm->levelCanBeRefined(ln));
      }

   } else {

      /* 
       * We are re-gridding.  Reset the time dependent data.  If 
       * the error estimation procedure does NOT use time 
       * integration (e.g. gradient detection) then simply go 
       * through all levels of the hierarchy and reset.  
       *
       * If the error estimation does use time integration 
       * (e.g. Richardson extrapolation) the time dependent
       * data will be accessed during the tagging phase. 
       * Hence, we only reset time dependent data on levels 
       * that are not tagged (i.e. the finest level) and
       * defer resetting time dependent data on levels that
       * are tagged to the level integrator performing the 
       * re-gridding operations. NOTE:  One should assure 
       * this is properly done in the level integrator. 
       */

      if (d_gridding_algorithm->errorEstimationUsesTimeIntegration() ) {
         if ( !d_gridding_algorithm->
               levelCanBeRefined(finest_level_number) ) {
            d_refine_level_integrator->resetTimeDependentData(
               d_patch_hierarchy->getPatchLevel(finest_level_number),
               d_integrator_time,
               d_gridding_algorithm->
               levelCanBeRefined(finest_level_number));
         }
      } else {
         for (int ln = 0; ln <= finest_level_number; ln++) {
            d_refine_level_integrator->resetTimeDependentData(
               d_patch_hierarchy->getPatchLevel(ln),
               d_integrator_time,
               d_gridding_algorithm->levelCanBeRefined(ln));
         }
      }

      /*
       * Regrid all levels, from coarsest to finest.  If the error 
       * estimation procedure uses time integration (e.g. Richardson 
       * extrapolation) then we must supply the oldest time at which 
       * data is stored.
       *
       * If the error coarsen ratio is two, data will be stored
       * from the previous timestep (at d_level_old_time).  If the 
       * error coarsen ratio is three, data will be stored
       * from two previous timesteps (at d_level_old_old_time).  
       * 
       * If we are not using time integration, the oldest time
       * information should not be used, so it is set to NaNs 
       * to throw an assertion if it is accessed.
       */
      
      tbox::Array<double> regrid_start_time;
      if (!d_gridding_algorithm->errorEstimationUsesTimeIntegration()) {

         int max_levels = d_gridding_algorithm->getMaxLevels();
         regrid_start_time.resizeArray(max_levels);
         for (int i = 0; i < regrid_start_time.getSize(); i++) {
            regrid_start_time[i] = 0.;
         }

      } else {

         if (d_gridding_algorithm->getErrorCoarsenRatio() == 2) {
            regrid_start_time = d_level_old_time;
         } else if (d_gridding_algorithm->getErrorCoarsenRatio() == 3) {
            regrid_start_time = d_level_old_old_time;
         } else {
            TBOX_ERROR(d_object_name << ": the supplied gridding "
                       << "algorithm uses an error coarsen ratio of "
                       << d_gridding_algorithm->getErrorCoarsenRatio()
                       << " which is not supported in this class" 
                       << endl);
         }

      }
      
      d_gridding_algorithm->
         regridAllFinerLevels(d_patch_hierarchy,
                              coarse_level_number,
                              d_integrator_time,
                              d_tag_buffer,
                              regrid_start_time);
      
      /*
       * Synchronize data on new levels.
       */
      if (d_patch_hierarchy->getFinestLevelNumber() > 0) {

#ifdef DEBUG_TIMES
         tbox::plog << "\nSynchronizing levels after regrid : "
              << coarse_level_number << " to "
              << d_patch_hierarchy->getFinestLevelNumber() << endl;
#endif
         const bool initial_time = false;
         d_refine_level_integrator->synchronizeNewLevels(
            d_patch_hierarchy,
            coarse_level_number,
            d_patch_hierarchy->getFinestLevelNumber(),
            d_integrator_time,
            initial_time);
      }

   }
   
   return (dt_new);
}


/*
************************************************************************* 
*                                                                       *
* Compute time step data for given level and estimate the number of     *
* time steps needed in current step sequence on that level.  The        *
* boolean return value indicates whether the next step taken will       *
* be the last in the step sequence on the level.  The outline of the    *
* procedure is as follows:                                              *
*                                                                       *
*    1) Determine maximum time increment allowed, permitting time step  *
*       growth if appropriate.                                          *
*    2) If time steps remain in current step sequence:                  *
*       a) Estimate the number of timesteps left in the sequence based  *
*          on the time remaining.                                       *
*       b) Possibly increase the number of time steps remaining so      *
*          that total step count remains consistent with regridding     *
*          sequence.  Note that the total number of steps must be an    *
*          integer multiple of the regrid interval for the level.       *
*       c) Compute time increment by dividing number of steps left      *
*          into remaining time.                                         *
*    3) If no steps remain on the level in the current sequence, the    *
*       next time increment is set to the current maximum increment.    * 
*                                                                       *
*************************************************************************
*/

template<int DIM> bool TimeRefinementIntegrator<DIM>::findNextDtAndStepsRemaining(
   const int level_number, 
   const double time_remaining, 
   const double dt_bound) 
{
#ifdef DEBUG_CHECK_ASSERTIONS
   assert( (level_number >= 0) &&
           (level_number <= d_patch_hierarchy->getFinestLevelNumber()) );
   assert(time_remaining >= 0.0);
   assert(dt_bound >= 0.0);
#endif

   /*
    * Grow time increment from previous if possible, but time step size 
    * cannot be greater than dt_bound. also, we cannot take any more 
    * steps if we have exceeded number allowable.
    */

   d_dt_max_level[level_number] = 
   tbox::Utilities::dmin( dt_bound, 
                         d_dt_max_level[level_number] * d_grow_dt );

   if ( d_step_level[level_number] < d_max_steps_level[level_number] ) {

      /*
       * If we have not exceeded the max number of steps, the max number
       * of steps can be adjusted according to the amount of time remaining
       * of the level.
       */

      int number_steps_remaining = 
      int(time_remaining/d_dt_max_level[level_number]);

      double dt_temp = 
      d_dt_max_level[level_number]*double(number_steps_remaining);

      if (time_remaining-dt_temp > 
          sqrt(tbox::IEEE::getDBL_EPSILON())*time_remaining) {
         number_steps_remaining++;
      }

      d_max_steps_level[level_number] = 
      d_step_level[level_number]+number_steps_remaining;

      /*
       * If we are not on the coarsest hierarchy level, there must be
       * an integer multiple of the regrid interval of steps in the
       * step sequence.
       */

      if ( level_number > 0
           && d_gridding_algorithm->levelCanBeRefined(level_number) ) {

         d_max_steps_level[level_number] = 
         tbox::Utilities::imax( d_max_steps_level[level_number],
                               d_regrid_interval[level_number] );

         int number_regrids = 
         d_max_steps_level[level_number]/d_regrid_interval[level_number];

         if ( d_max_steps_level[level_number] 
              % d_regrid_interval[level_number] ) number_regrids++;

         d_max_steps_level[level_number] =
         number_regrids * d_regrid_interval[level_number];

      }

      if ( d_step_level[level_number] >= d_max_steps_level[level_number] ) {
         TBOX_ERROR(d_object_name << ":  "
                    << "no steps left to divide remaining time ...\n" 
                    << "level_number = " << level_number << endl
                    << "time_remaining = " << time_remaining
                    << "\ndt_bound = " << dt_bound
                    << "\nnumber_steps_remaining = " << number_steps_remaining
                    << endl);
      }

      /*
       * Adjust current time increment so that amount of time remaining
       * on the level is divided approximately evenly among the time
       * steps remaining.
       */

      d_dt_actual_level[level_number] = 
      time_remaining / double( d_max_steps_level[level_number]
                               - d_step_level[level_number] );

   } else {
      d_dt_actual_level[level_number] = d_dt_max_level[level_number];
   }

   return( (d_max_steps_level[level_number] 
            - d_step_level[level_number]) <= 1 ); 

}

/*
*************************************************************************
*                                                                       *
* Return true if the level can be remeshed at the current step.         * 
* regrid step interval.  Otherwise, return false.                       *
*                                                                       *
*************************************************************************
*/

template<int DIM> bool TimeRefinementIntegrator<DIM>::atRegridPoint(
   const int level_number) const
{
#ifdef DEBUG_CHECK_ASSERTIONS
   assert( (level_number >= 0) && 
           (level_number <= d_patch_hierarchy->getFinestLevelNumber()) );
#endif

   int step_number = ( (level_number == 0) ? d_integrator_step 
                                           : d_step_level[level_number] );

   return( (step_number>0) 
            && d_gridding_algorithm->levelCanBeRefined(level_number)
            && (step_number%d_regrid_interval[level_number] == 0) );
}

/*
*************************************************************************
*                                                                       *
* Return true if the specified level can be regridded and the the next  *
* coarser level can be regridded too.  Otherwise, false is returned.    *
*                                                                       *
*************************************************************************
*/

template<int DIM> bool TimeRefinementIntegrator<DIM>::coarserLevelRegridsToo(
   const int level_number) const 
{
#ifdef DEBUG_CHECK_ASSERTIONS
   assert( (level_number >= 0) &&
           (level_number <= d_patch_hierarchy->getFinestLevelNumber()) );
#endif
   return( (level_number>0) ? atRegridPoint(level_number-1) : false );
}

/*
*************************************************************************
*                                                                       *
* Print all data member for TimeRefinementIntegrator object.            *
*                                                                       *
*************************************************************************
*/

template<int DIM> void TimeRefinementIntegrator<DIM>::printClassData(ostream& os) const 
{
   os << "\nTimeRefinementIntegrator<DIM>::printClassData..." << endl;
   os << "\nTimeRefinementIntegrator<DIM>: this = "
      << (TimeRefinementIntegrator<DIM>*)this << endl;
   os << "d_object_name = " << d_object_name << endl;
   os << "d_integrator_time = " << d_integrator_time << "\n"
      << "d_start_time = " << d_start_time << "\n"
      << "d_end_time = " << d_end_time << "\n"
      << "d_integrator_step = " << d_integrator_step << "\n"
      << "d_max_integrator_steps = " << d_max_integrator_steps << "\n"
      << "d_grow_dt = " << d_grow_dt << endl;
   os << "d_just_regridded = " << d_just_regridded << endl;
   os << "d_last_finest_level = " << d_last_finest_level << endl;
   os << "d_patch_hierarchy = " << d_patch_hierarchy.getPointer() << endl;
   os << "d_refine_level_integrator = " 
      << (TimeRefinementLevelStrategy<DIM>*)d_refine_level_integrator << endl;
   os << "d_gridding_algorithm = " 
      << (mesh::BaseGriddingAlgorithm<DIM>*)d_gridding_algorithm << endl;

   const int max_levels = d_gridding_algorithm->getMaxLevels();
   for (int level_number=0; level_number<max_levels; level_number++) {
      printDataForLevel(os, level_number);
   }
}

/*
*************************************************************************
*                                                                       *
* Print all level-specific data for TimeRefinementIntegrator object.    *
*                                                                       *
*************************************************************************
*/

template<int DIM> void TimeRefinementIntegrator<DIM>::printDataForLevel(
   ostream& os, 
   const int level_number) const 
{
#ifdef DEBUG_CHECK_ASSERTIONS
   assert( (level_number >= 0) &&
           (level_number <= d_patch_hierarchy->getFinestLevelNumber()) );
#endif
   os << "\nTimeRefinementIntegrator<DIM>::printDataForLevel..." << endl;
   os << "\nd_level_sim_time[" << level_number << "] = " 
      << d_level_sim_time[level_number] << endl;
   os << "\nd_level_old_time[" << level_number << "] = " 
      << d_level_old_time[level_number] << endl;
   os << "\nd_level_old_old_time[" << level_number << "] = " 
      << d_level_old_old_time[level_number] << endl;
   os << "\nd_dt_max_level[" << level_number << "] = " 
      << d_dt_max_level[level_number] << endl;
   os << "\nd_dt_actual_level[" << level_number << "] = " 
      << d_dt_actual_level[level_number] << endl;
   os << "\nd_step_level[" << level_number << "] = " 
      << d_step_level[level_number] << endl;
   os << "\nd_max_steps_level[" << level_number << "] = " 
      << d_max_steps_level[level_number] << endl;
}

/*
*************************************************************************
*                                                                       *
* Write the class version number and data members to database object.   *
*                                                                       *
*************************************************************************
*/

template<int DIM> void TimeRefinementIntegrator<DIM>::putToDatabase(
   tbox::Pointer<tbox::Database> db)
{
#ifdef DEBUG_CHECK_ASSERTIONS
   assert(!db.isNull());
#endif

   db->putInteger("ALGS_TIME_REFINEMENT_INTEGRATOR_VERSION",
                   ALGS_TIME_REFINEMENT_INTEGRATOR_VERSION);

   db->putDouble("d_start_time", d_start_time);
   db->putDouble("d_end_time", d_end_time);
   db->putDouble("d_grow_dt", d_grow_dt);
   db->putInteger("d_max_integrator_steps", d_max_integrator_steps);
   db->putIntegerArray("d_regrid_interval", d_regrid_interval);
   db->putIntegerArray("d_tag_buffer", d_tag_buffer);
   db->putDouble("d_integrator_time", d_integrator_time);
   db->putInteger("d_integrator_step", d_integrator_step);
   db->putInteger("d_last_finest_level", d_last_finest_level);
   db->putDoubleArray("d_dt_max_level", d_dt_max_level);
   db->putDoubleArray("d_dt_actual_level", d_dt_actual_level);
}

/*
*************************************************************************
*                                                                       *
* If simulation is not from restart, read in all data members from      *
* the input database.  Otherwise, only override end_time, grow_dt       *
* max_integrator_steps, and tag_buffer from the input database.		*
*                                                                       *
*************************************************************************
*/

template<int DIM> void TimeRefinementIntegrator<DIM>::getFromInput(
   tbox::Pointer<tbox::Database> db,
   bool is_from_restart)
{
#ifdef DEBUG_CHECK_ASSERTIONS
   assert(!db.isNull());
#endif
   
   if (is_from_restart) {
      if (db->keyExists("end_time")) {
         d_end_time = db->getDouble("end_time");
      } 

      if (db->keyExists("grow_dt")) {
         d_grow_dt = db->getDouble("grow_dt");
      } 

      if (db->keyExists("max_integrator_steps")) {
         d_max_integrator_steps = db->getInteger("max_integrator_steps");
      } 

      if (db->keyExists("tag_buffer")) {
         tbox::Array<int> temp_tag_buffer = db->getIntegerArray("tag_buffer");
         if (temp_tag_buffer.getSize() <
             (d_gridding_algorithm->getMaxLevels() - 1)) {
            int tsize = temp_tag_buffer.getSize();
            temp_tag_buffer.resizeArray(
               d_gridding_algorithm->getMaxLevels() - 1);
            for (int i = tsize;
                 i < d_gridding_algorithm->getMaxLevels(); i++) {
               temp_tag_buffer[i] = temp_tag_buffer[tsize-1];
            }
         }
         d_tag_buffer = temp_tag_buffer;
      }

   } else { 
      
      /*
       * If not from restart, read in all data members from input database.
       */

      if (db->keyExists("start_time")) {
         d_start_time = db->getDouble("start_time");
      } else {
         TBOX_ERROR(d_object_name << ":  "
                    << "Key data `start_time' not found in input.");
      }

      if (db->keyExists("end_time")) {
         d_end_time = db->getDouble("end_time");
      } else {
         TBOX_ERROR(d_object_name << ":  "
                    << "Key data `end_time' not found in input.");
      }

      if (db->keyExists("grow_dt")) {
         d_grow_dt = db->getDouble("grow_dt");
      } else {
         TBOX_ERROR(d_object_name << ":  "
                    << "Key data `grow_dt' not found in input.");
      }

      if (db->keyExists("max_integrator_steps")) {
         d_max_integrator_steps = db->getInteger("max_integrator_steps");
      } else {
         TBOX_ERROR(d_object_name << ":  "
                    << "Key data `max_integrator_steps' not found in input.");
      }

      if (db->keyExists("tag_buffer")) {
         d_tag_buffer = db->getIntegerArray("tag_buffer");
         if (d_tag_buffer.getSize() <
             (d_gridding_algorithm->getMaxLevels() - 1)) {
            int tsize = d_tag_buffer.getSize();
            d_tag_buffer.resizeArray(
               d_gridding_algorithm->getMaxLevels() - 1);
            for (int i = tsize;
                 i < d_gridding_algorithm->getMaxLevels(); i++) {
               d_tag_buffer[i] = d_tag_buffer[tsize-1];
            }
         }
      } else {
         int level_number;

         d_tag_buffer.resizeArray(d_gridding_algorithm->getMaxLevels());
         for (level_number = 0; 
              level_number < d_gridding_algorithm->getMaxLevels();
              level_number++) {
            d_tag_buffer[level_number] = d_regrid_interval[level_number];
         }

         TBOX_WARNING(d_object_name << ":  "
                   << "Key data `tag_buffer' not found in input.  "
                   << "Default values used.  See class header for details.");
      }
   } 
}

/*
*************************************************************************
*                                                                       *
* Gets the database in the restart root database that corresponds to    * 
* the object name.  This method then checks to make sure that the class *
* version number and the restart version number are the same.  If they  *
* are, then reads in the objects data members from the restart          *
* database.                                                             *
*                                                                       *
* Data read from restart database: d_start_time, d_end_time, d_grow_dt, *
* d_max_integrator_step, d_regrid_interval, d_tag_buffer,               *
* d_integrator_step, d_dt_max_level, d_dt_actual_level.                 *
*                                                                       *
*************************************************************************
*/

template<int DIM> void TimeRefinementIntegrator<DIM>::getFromRestart()
{

   tbox::Pointer<tbox::Database> restart_db =
      tbox::RestartManager::getManager()->getRootDatabase();

   tbox::Pointer<tbox::Database> db;
   if ( restart_db->isDatabase(d_object_name) ) {
      db = restart_db->getDatabase(d_object_name);
   } else {
      TBOX_ERROR("Restart database corresponding to "
                 << d_object_name << " not found in restart file.");
   }

   int ver = db->getInteger("ALGS_TIME_REFINEMENT_INTEGRATOR_VERSION");
   if (ver != ALGS_TIME_REFINEMENT_INTEGRATOR_VERSION) {
      TBOX_ERROR(d_object_name << ":  "
                 << "Restart file version different than class version.");
   }

   d_start_time = db->getDouble("d_start_time");
   d_end_time = db->getDouble("d_end_time");
   d_grow_dt = db->getDouble("d_grow_dt");
   d_max_integrator_steps = db->getInteger("d_max_integrator_steps");
   d_regrid_interval = db->getIntegerArray("d_regrid_interval");
   d_tag_buffer = db->getIntegerArray("d_tag_buffer");
   d_integrator_time = db->getDouble("d_integrator_time");
   d_integrator_step = db->getInteger("d_integrator_step");
   d_last_finest_level = db->getInteger("d_last_finest_level");
   d_dt_max_level = db->getDoubleArray("d_dt_max_level");
   d_dt_actual_level = db->getDoubleArray("d_dt_actual_level");
}

}
}
#endif
