/*
  Id:		$Id: JacobiPoissonX.C.sed,v 1.1.2.2 2002/03/20 00:43:45 gunney Exp $
  Copyright:	(c) 1997-2000 The Regents of the University of California
  Release:	$Name: v_0_1_7 $
  Revision:	$Revision: 1.1.2.2 $
  Modified:	$Date: 2002/03/20 00:43:45 $
  Description:	JacobiPoissonX class implementation
*/


#include <setArrayData.h>
#include <printObjectX.h>
#include <MultiDimArrayAccess.h>
#include <arrayConversionX.h>
#include <setVectorWeightsX.h>
#include "miscFcnsX.h"
#include "JacobiPoissonX.h"

#include <hier_VariableDatabase2.h>
#include <pdat_CellData2.h>
#include <pdat_SideData2.h>
#include <hier_Patch2.h>
#include <hier_Variable2.h>
#include <geom_CartesianPatchGeometry2.h>
#include <xfer_Geometry2.h>
#include <xfer_CoarsenAlgorithm2.h>
#include <xfer_CoarsenOperator2.h>
#include <xfer_CoarsenSchedule2.h>
#include <xfer_RefineAlgorithm2.h>
#include <xfer_RefineOperator2.h>
#include <xfer_RefineSchedule2.h>
#include <hier_Index2.h>
#include <geom_CartesianCellDoubleLinearRefine2.h>
#include <geom_CartesianCellDoubleConservativeLinearRefine2.h>
#include <geom_CartesianCellDoubleWeightedAverage2.h>


using namespace SAMRAI;


JacobiPoissonX::JacobiPoissonX(
  const string &object_name
, tbox_Pointer<tbox_Database> database
)
: d_context_persistent("PERSISTENT")
, d_context_scratch("SCRATCH")
, d_diffcoef("solution:diffcoef",1)
, d_scalar("solution:scalar",1)
, d_source("source",1)
, d_exact("solution:exact",1)
, d_weight("vector weight",1)
{

  /*
    Register variables with VariableDatabase
    and get the descriptor indices for those variables.
    It is not necessary to save the indices obtained from
    the registration, because they can always be retrieved
    from the mapVariableAndContextToIndex, but we do it
    because we refer to the indices often.
  */
  {
    tbox_Pointer<hier_VariableContext>
      context_persistent_ptr(&d_context_persistent,false);
    tbox_Pointer<hier_VariableContext>
      context_scratch_ptr(&d_context_scratch,false);
    hier_VariableDatabaseX *variable_db = hier_VariableDatabaseX::getDatabase();

    /*
	Persistent data.
    */
    d_diffcoef_persistent
      = variable_db->registerVariableAndContext(
	tbox_Pointer<hier_VariableX>(&d_diffcoef,false)
	, context_persistent_ptr );
    d_scalar_persistent
      = variable_db->registerVariableAndContext(
	tbox_Pointer<hier_VariableX>(&d_scalar,false)
	, context_persistent_ptr );
    d_source_persistent
      = variable_db->registerVariableAndContext(
	tbox_Pointer<hier_VariableX>(&d_source,false)
	, context_persistent_ptr
	, hier_IntVectorX(1) /* ghost cell width is 1 */
	);
    d_exact_persistent
      = variable_db->registerVariableAndContext(
	tbox_Pointer<hier_VariableX>(&d_exact,false)
	, context_persistent_ptr );
    d_weight_persistent
      = variable_db->registerVariableAndContext(
	tbox_Pointer<hier_VariableX>(&d_weight,false)
	, context_persistent_ptr );
    /*
      Scratch data.
    */
    d_scalar_scratch
      = variable_db->registerVariableAndContext(
	tbox_Pointer<hier_VariableX>(&d_scalar,false)
	, context_scratch_ptr
	, hier_IntVectorX(1) /* ghost cell width is 1 */
	);
  }

  return;
}





void JacobiPoissonX::initializeLevelData (
    /*! Hierarchy to initialize */
    const tbox_Pointer< hier_PatchHierarchyX > hierarchy
  , /*! Level to initialize */
    const int level_number
  , const double init_data_time
  , const bool can_be_refined
  , /*! Whether level is being introduced for the first time */
    const bool initial_time
  , /*! Level to copy data from */
    const tbox_Pointer< hier_PatchLevelX > old_level
  , const bool allocate_data
  ) {

  /*
    Reference the level object with the given index from the hierarchy.
  */
  tbox_Pointer<hier_PatchLevelX> level
    = hierarchy->getPatchLevel(level_number);

  /*
    If instructed, allocate all patch data on the level.
    Allocate only persistent data.  Scratch data will
    generally be allocated and deallocated as needed.
  */
  if ( allocate_data ) {
    level->allocatePatchData( d_diffcoef_persistent );
    level->allocatePatchData( d_scalar_persistent );
    level->allocatePatchData( d_source_persistent );
    level->allocatePatchData( d_exact_persistent );
    level->allocatePatchData( d_weight_persistent );
  }

  /*
    Initialize data in all patches in the level.
  */
  int npatch = level->getNumberOfPatches();
  for ( int j=0; j<npatch; j++ ) {

    tbox_Pointer<hier_PatchX> patch
      = level->getPatch(j);
    tbox_Pointer<geom_CartesianPatchGeometryX> patch_geom
      = patch->getPatchGeometry();

    tbox_Pointer<pdat_SideDataX<double> > diffcoef_data
      = patch->getPatchData(d_diffcoef_persistent);
    tbox_Pointer<pdat_CellDataX<double> > scalar_data
      = patch->getPatchData(d_scalar_persistent);
    tbox_Pointer<pdat_CellDataX<double> > exact_data
      = patch->getPatchData(d_exact_persistent);
    tbox_Pointer<pdat_CellDataX<double> > source_data
      = patch->getPatchData(d_source_persistent);

    /* Set initial guess of solution. */
    MultiDimArrayAccess<double,2> t0
      = arrayData2ArrayAccess( scalar_data->getArrayData() );
    setArrayDataToConstant(
        t0 // arrayData2ArrayAccess( scalar_data->getArrayData() )
      , scalar_data->getGhostBox().lower()
      , scalar_data->getGhostBox().upper()
      , patch_geom->getXLower()
      , patch_geom->getXUpper()
      , patch_geom->getDx()
      , 0.0
      );
    /* Set exact solution. */
    MultiDimArrayAccess<double,2> t4
      = arrayData2ArrayAccess( exact_data->getArrayData() );
    setArrayDataToSinusoidal(
        t4 // arrayData2ArrayAccess( exact_data->getArrayData() )
      , exact_data->getBox().lower()
      , exact_data->getBox().upper()
      , patch_geom->getXLower()
      , patch_geom->getXUpper()
      , patch_geom->getDx()
      );
    setArrayDataToScaled(
        t4 // arrayData2ArrayAccess( scratch_solution->getArrayData() )
      , exact_data->getGhostBox().lower()
      , exact_data->getGhostBox().upper()
      , -1.0/(8*M_PI*M_PI)
      );
    /* Set source function. */
    MultiDimArrayAccess<double,2> t1
      = arrayData2ArrayAccess( source_data->getArrayData() );
    setArrayDataToSinusoidal(
        t1 // arrayData2ArrayAccess( source_data->getArrayData() )
      , source_data->getBox().lower()
      , source_data->getBox().upper()
      , patch_geom->getXLower()
      , patch_geom->getXUpper()
      , patch_geom->getDx()
      );
    printObjectX(plog,source_data->getArrayData());
    /* Set diffusion coefficients. */
    MultiDimArrayAccess<double,2> t2
      = arrayData2ArrayAccess( diffcoef_data->getArrayData(0) );
    setArrayDataToConstant(
        t2
      , diffcoef_data->getGhostBox().lower()
      , diffcoef_data->getGhostBox().upper()
      , patch_geom->getXLower()
      , patch_geom->getXUpper()
      , patch_geom->getDx()
      , 1.0
      );
    MultiDimArrayAccess<double,2> t3
      = arrayData2ArrayAccess( diffcoef_data->getArrayData(1) );
    setArrayDataToConstant(
        t3
      , diffcoef_data->getGhostBox().lower()
      , diffcoef_data->getGhostBox().upper()
      , patch_geom->getXLower()
      , patch_geom->getXUpper()
      , patch_geom->getDx()
      , 1.0
      );
  }
  /* Set vector weight. */
  setVectorWeightsX( hierarchy, d_weight_persistent );
}




int JacobiPoissonX::jacobiIteration(
  const tbox_Pointer< SAMRAI::hier_PatchHierarchyX > hierarchy
, double relax_coef
) {

  /*
    Jacobi iteration requires temporarily copying solution into scratch space.
    In this loop, we allocate the scratch space.  It will be deallocated
    before leaving in this function.
  */
  for ( int ln=0; ln < hierarchy->getNumberLevels(); ++ln ) {
    tbox_Pointer<hier_PatchLevelX> level = hierarchy->getPatchLevel(ln);
    level->allocatePatchData(d_scalar_scratch);
  }



  /*
    Refine solution into scratch space to fill the fine level
    ghost cells in the scratch space.  The interiors will be
    filled also, but then they will be overwritten by copying
    the actual fine grid solution into the fine coverage area.

    1. Create a refine algorithm.

    2. Associate a refinement operator with source-desination pairs of data
    (specified by indices).

    3. Cycle through the levels and refine the data.
  */
  // 1.
  xfer_RefineAlgorithmX refiner;
  // 2.
  {

    tbox_Pointer<geom_CartesianCellDoubleConservativeLinearRefineX>
      conservative_linear_refine_operator
      = new geom_CartesianCellDoubleConservativeLinearRefineX;
    assert( conservative_linear_refine_operator->getOperatorName() == "CONSERVATIVE_LINEAR_REFINE" );

    tbox_Pointer<geom_CartesianCellDoubleLinearRefineX>
      linear_refine_operator
      = new geom_CartesianCellDoubleLinearRefineX;
    assert( linear_refine_operator->getOperatorName() == "LINEAR_REFINE" );

    refiner.registerRefine( d_scalar_scratch, d_scalar_persistent
			  , d_scalar_scratch, conservative_linear_refine_operator );
  }
  // 3.
  for ( int src_ln=0; src_ln < hierarchy->getNumberLevels()-1; ++src_ln ) {
    int dst_ln = src_ln+1;
    /*
      Source level is the the current level (number ln).
      Destination level is the next finer level (ln+1).
    */
    tbox_Pointer<hier_PatchLevelX> src_level
      = hierarchy->getPatchLevel(src_ln);
    tbox_Pointer<hier_PatchLevelX> dst_level
      = hierarchy->getPatchLevel(dst_ln);
    assert(src_level);
    assert(dst_level);

    {
      // refiner.createSchedule( dst_level , src_level )->fillData(0.0);
      // tbox_Pointer<xfer_RefineScheduleX> refine_schedule =
      // refiner.createSchedule( dst_level , src_level );
      tbox_Pointer<xfer_RefineScheduleX> refine_schedule =
        refiner.createSchedule( dst_level, src_ln, hierarchy );
        // refiner.createSchedule( dst_level, src_level, hierarchy );
      refine_schedule->fillData(0.0);
    }

  }



  /*
    Copy solution so that the interiors of scratch patches get set
    to the current solution.
  */
  for ( int ln=0; ln < hierarchy->getNumberLevels(); ++ln ) {
    tbox_Pointer<hier_PatchLevelX> level = hierarchy->getPatchLevel(ln);

    for ( hier_PatchLevelX::Iterator p(level); p; p++ ) {
      tbox_Pointer<hier_PatchX> patch = level->getPatch(p());

      /* Get the patch data. */
      tbox_Pointer<hier_PatchDataX> current_solution
	= patch->getPatchData(d_scalar_persistent);
      tbox_Pointer<hier_PatchDataX> scratch_solution
	= patch->getPatchData(d_scalar_scratch);

      /* Copy the current solution into the scratch space. */
      // plog << "current before copying: level " << ln << " patch " << p() ;
      // printObjectX( plog, (hier_PatchDataX&)current_solution );
      // printObjectX( plog, current_solution->getArrayData() );
      scratch_solution->copy(*current_solution);
      // plog << "scratch after copying: level " << ln << " patch " << p() ;
      // printObjectX( plog, (hier_PatchDataX&)scratch_solution );
      // printObjectX( plog, scratch_solution->getArrayData() );

    }

  }




  /*
    Perform jacobi relaxation on all levels, all patches.
  */
  for ( int ln = hierarchy->getFinestLevelNumber(); ln >= 0; ln -- ) {

    tbox_Pointer<hier_PatchLevelX> level = hierarchy->getPatchLevel(ln);

    for ( hier_PatchLevelX::Iterator p(level); p; p++ ) {

      tbox_Pointer<hier_PatchX> patch = level->getPatch(p());

      /*
	Get information about the patch geometry.
      */
      tbox_Pointer<geom_CartesianPatchGeometryX> patch_geom
	= patch->getPatchGeometry();

      /*
	Get the patch data.
      */
      tbox_Pointer<pdat_CellDataX<double> > current_solution
	= patch->getPatchData(d_scalar_persistent);
      tbox_Pointer<pdat_CellDataX<double> > scratch_solution
	= patch->getPatchData(d_scalar_scratch);
      tbox_Pointer<pdat_CellDataX<double> > source
	= patch->getPatchData(d_source_persistent);

      if ( ln == 0 ) {
	/*
	  Set physical boundary condition in the scratch space
	  for coarsest level.
	*/
	jacobiPoissonBoundary(
            arrayData2ArrayAccess( scratch_solution->getArrayData(), 0 )
          , current_solution->getBox().lower()
          , current_solution->getBox().upper()
          , patch_geom->getXLower()
          , patch_geom->getXUpper()
          , patch_geom->getDx()
          );
      }
      if( 0 ) jacobiPoissonExactBC(
          arrayData2ArrayAccess( scratch_solution->getArrayData() )
        , current_solution->getBox().lower()
        , current_solution->getBox().upper()
        , patch_geom->getXLower()
        , patch_geom->getXUpper()
        , patch_geom->getDx()
        );

      /*
	Relax solution in scratch and assign result to the solution.
      */
      // plog << "scratch before: level " << ln << " patch " << p() ;
      // printObjectX( plog, (hier_PatchDataX&)scratch_solution );
      // printObjectX( plog, scratch_solution->getArrayData() );
      // plog << "current before: level " << ln << " patch " << p() ;
      // printObjectX( plog, (hier_PatchDataX&)current_solution );
      // printObjectX( plog, current_solution->getArrayData() );
      jacobiPoissonSmooth(
          relax_coef
        , arrayData2ArrayAccess( current_solution->getArrayData(), 0 )
        , arrayData2ArrayAccess( scratch_solution->getArrayData(), 0 )
        , arrayData2ArrayAccess( source->getArrayData(), 0 )
        , current_solution->getBox().lower()
        , current_solution->getBox().upper()
        , patch_geom->getXLower()
        , patch_geom->getXUpper()
        , patch_geom->getDx()
        );
      // plog << "current after: level " << ln << " patch " << p() ;
      // printObjectX( plog, (hier_PatchDataX&)current_solution );
      // printObjectX( plog, current_solution->getArrayData() );

    }

  }



  /*
    The computational stencils will be extending into the ghost cells
    of the scratch space, so for each level (except the finest) get
    the data into the ghost cells using data from the next finest level.

    1. Create a coarsen algorithm to coarsen data on fine patch interiors.

    2.
    Associate a coarsening operator with source-desination pairs of data
    (specified by indices).  The indices are the same because we transfer
    data from between levels, but it is the same variable in the same
    context.

    3. Cycle through the levels and coarsen the data.
  */

  // 1.
  xfer_CoarsenAlgorithmX coarsener;
  // 2.
  {
    tbox_Pointer<geom_CartesianCellDoubleWeightedAverageX>
      conservative_coarsen_operator
      = new geom_CartesianCellDoubleWeightedAverageX;
    coarsener.registerCoarsen(
			       d_scalar_persistent
			     , d_scalar_persistent
			     , conservative_coarsen_operator
			     );
  }
  // 3.
  for ( int ln = hierarchy->getFinestLevelNumber(); ln > 0; --ln ) {
    if(1) coarsener.createSchedule(
        hierarchy->getPatchLevel(ln-1)
      , hierarchy->getPatchLevel(ln)
      )->coarsenData();
  }

  /*
    Deallocate scratch data.
  */
  for ( int ln=0; ln < hierarchy->getNumberLevels(); ++ln ) {
    tbox_Pointer<hier_PatchLevelX> level = hierarchy->getPatchLevel(ln);
    level->deallocatePatchData(d_scalar_scratch);
  }

  return 0;
}






int JacobiPoissonX::registerVariablesWithPlotter(
  tbox_Pointer<plot_CartesianVizamraiDataWriterX> viz_writer
) const {

  /*
    Register variables with plotter.
  */
  viz_writer->registerPlotVariable("Computed solution", d_scalar_persistent);
  viz_writer->registerPlotVariable("Exact solution", d_exact_persistent);
  viz_writer->registerDerivedPlotVariable("Error"
    , (plot_VizamraiDerivedDataStrategyX *)this );
  viz_writer->registerPlotVariable("Poisson source", d_source_persistent);
  viz_writer->registerPlotVariable("Vector weight", d_weight_persistent);

  return 0;
}




void JacobiPoissonX::writeDerivedDataToStream(
  tbox_AbstractStream &stream
, const hier_PatchX &patch
, int patch_level_number
, const hier_BoxX &region
, const string &variable_name
, int plot_type
) {

  assert(plot_type==1);	// Writing float data.
  assert( variable_name == "Error" );
  /*
    Get the patch data.
  */
  tbox_Pointer<pdat_CellDataX<double> > current_solution
    = patch.getPatchData(d_scalar_persistent);
  tbox_Pointer<pdat_CellDataX<double> > exact_solution
    = patch.getPatchData(d_exact_persistent);

  {
    ConstMultiDimArrayAccess<double,2> ex =
      arrayData2ArrayAccess( exact_solution->getArrayData() );
    ConstMultiDimArrayAccess<double,2> co =
      arrayData2ArrayAccess( current_solution->getArrayData() );
    const int *lower = region.lower();
    const int *upper = region.upper();
    for ( int j=lower[1]; j<=upper[1]; ++j ) {
      for ( int i=lower[0]; i<=upper[0]; ++i ) {
	float diff = ( co(j,i) - ex(j,i) );
	stream.pack(&diff);
      }
    }
  }
  return;
}





int JacobiPoissonX::computeError(
  const tbox_Pointer< SAMRAI::hier_PatchHierarchyX > hierarchy
, double *l2norm
, double *linorm
) const {

  /*
    Compute error on all levels, all patches.
    This can be done using the data operator classes in SAMRAI.
    But we do it manually here to combine the operations into
    the least number of loopings and to illustrate how it is done.
  */
  double diff=0, difflinf=0, diffl2=0, weightsum=0;
  for ( int ln = hierarchy->getFinestLevelNumber(); ln >= 0; ln -- ) {
    tbox_Pointer<hier_PatchLevelX> level = hierarchy->getPatchLevel(ln);

    for ( hier_PatchLevelX::Iterator p(level); p; p++ ) {
      tbox_Pointer<hier_PatchX> patch = level->getPatch(p());

      /*
	Get the patch data.
      */
      tbox_Pointer<pdat_CellDataX<double> > current_solution
	= patch->getPatchData(d_scalar_persistent);
      tbox_Pointer<pdat_CellDataX<double> > exact_solution
	= patch->getPatchData(d_exact_persistent);
      tbox_Pointer<pdat_CellDataX<double> > weight
	= patch->getPatchData(d_weight_persistent);

      {
	ConstMultiDimArrayAccess<double,2> ex =
          arrayData2ArrayAccess( exact_solution->getArrayData() );
	ConstMultiDimArrayAccess<double,2> co =
          arrayData2ArrayAccess( current_solution->getArrayData() );
	ConstMultiDimArrayAccess<double,2> wt =
          arrayData2ArrayAccess( weight->getArrayData() );
        const int *lower = current_solution->getBox().lower();
        const int *upper = current_solution->getBox().upper();
	for ( int j=lower[1]; j<=upper[1]; ++j ) {
	  for ( int i=lower[0]; i<=upper[0]; ++i ) {
	    /*
	      Disregard zero weights in error computations
	      because they are on coarse grids covered by finer grids.
	    */
	    if ( wt(j,i) > 0 ) {
	      diff = fabs( co(j,i) - ex(j,i) );
	      if ( difflinf < diff ) difflinf = diff;
	      diffl2 += wt(j,i)*diff*diff;
	      weightsum += wt(j,i);
	    }
	  }
	}
      }

    }

  }
  diffl2 = sqrt( diffl2/weightsum );

  *l2norm = diffl2;
  *linorm = difflinf;

  return 0;
}
