/*
  Id:		$Id: JacobiPoissonX.h.sed,v 1.1.2.1 2002/03/16 02:07:00 gunney Exp $
  Copyright:	(c) 1997-2000 The Regents of the University of California
  Release:	$Name: v_0_1_7 $
  Revision:	$Revision: 1.1.2.1 $
  Modified:	$Date: 2002/03/16 02:07:00 $
  Description:	JacobiPoissonX class declaration
*/

#ifndef included_JacobiPoissonX_h
#define included_JacobiPoissonX_h


#include <string.h>
#include <tbox_Pointer.h>
#include <tbox_Database.h>


/*
  SAMRAI classes
*/
#include <mesh_StandardTagAndInitStrategy2.h>
#include <plot_CartesianVizamraiDataWriter2.h>
#include <pdat_CellVariable2.h>
#include <pdat_SideVariable2.h>
#include <hier_Patch2.h>
#include <hier_PatchHierarchy2.h>
#include <hier_PatchLevel2.h>
#include <hier_IntVector2.h>
#include <mesh_StandardTagAndInitStrategy2.h>
#include <hier_Box2.h>
#include <plot_VizamraiDerivedDataStrategy2.h>


using namespace SAMRAI;


/*!
  \brief Class to solve JacobiPoissonX's equation on a SAMR grid.

  This class demonstrates how use the FAC solver class to solve
  JacobiPoissonX's equation on a SAMR grid.

  This class inherits and implements virtual functions from
  SAMRAI::mesh_StandardTagAndInitStrategyX to initialize data
  on the SAMR grid.
*/
class JacobiPoissonX
  : public mesh_StandardTagAndInitStrategyX
  , public plot_VizamraiDerivedDataStrategyX
{

public:

  /*!
    \brief Constructor.

    Requirements:
    - the referenced objects

    Actions:
    - Initialize the viz_writer to serve this object
  */
  JacobiPoissonX(
    /*! Ojbect name */
    const string &object_name
  , /*! Input database */
    tbox_Pointer<SAMRAI::tbox_Database> database
  );



  //@{ \name SAMRAI::mesh_StandardTagAndInitStrategyX virtuals

public:

  /*!
    \brief Allocate and initialize data for a new level in the patch hierarchy.

    This is where you implement the code for initialize data on the
    grid.  Nevermind when it is called or where in the program that
    happens.  All the information you need to initialize the grid
    are in the arguments.

    \see SAMRAI::mesh_StandardTagAndInitStrategyX::initializeLevelData()
  */
  virtual void initializeLevelData (
      /*! Hierarchy to initialize */
      const tbox_Pointer< SAMRAI::hier_PatchHierarchyX > hierarchy
    , /*! Level to initialize */
      const int level_number
    , const double init_data_time
    , const bool can_be_refined
    , /*! Whether level is being introduced for the first time */
      const bool initial_time
    , /*! Level to copy data from */
      const tbox_Pointer< SAMRAI::hier_PatchLevelX > old_level
        = tbox_Pointer< SAMRAI::hier_PatchLevelX >((0))
    , /*! Whether data on new patch needs to be allocated */
      const bool allocate_data = true
    );

  virtual void resetHierarchyConfiguration (
      const tbox_Pointer< SAMRAI::hier_PatchHierarchyX > hierarchy
    , const int coarsest_level
    , const int finest_level
      ) { /* Nothing is needed. */ }

  //@}

  //@{ \name SAMRAI::plot_VizamraiDerivedDataStrategyX virtuals

  /*!
    \brief Function to support writing vizamrai derived data.
  */
  void JacobiPoissonX::writeDerivedDataToStream(
    tbox_AbstractStream &stream
  , const hier_PatchX &patch
  , int patch_level_number
  , const hier_BoxX &region
  , const string &variable_name
  , int plot_type
  );

  //@}



public:

  /*!
    \brief Tell a plotter which data to write for this class.
  */
  int registerVariablesWithPlotter(
    /*! Vizramrai writer */
    tbox_Pointer<plot_CartesianVizamraiDataWriterX> viz_writer
    ) const;

  /*!
    \brief Smooth the solution using one Jacobi iteration
  */
  int jacobiIteration(
      /*! The hierarchy containing the grid. */
      const tbox_Pointer< SAMRAI::hier_PatchHierarchyX > hierarchy
    , /*! The relaxation coefficient in the iteration. */
      double relax_coef = 1
    );


  /*!
    \brief Compute the error of the current solution.
  */
  int JacobiPoissonX::computeError(
    const tbox_Pointer< SAMRAI::hier_PatchHierarchyX > hierarchy
  , double *l2norm
  , double *linorm
  ) const;




  //@{
private:
  /*!
    \name Private state variables for solution.
  */

  /*!
    \brief Context for persistent data.
  */
  SAMRAI::hier_VariableContext d_context_persistent;

  /*!
    \brief Context for scratch data.
  */
  SAMRAI::hier_VariableContext d_context_scratch;

  /*!
    \brief Diffusion coefficient.
  */
  SAMRAI::pdat_SideVariableX<double> d_diffcoef;

  /*!
    \brief Scalar solution of Poisson's equation.
  */
  SAMRAI::pdat_CellVariableX<double> d_scalar;

  /*!
    \brief Source for Poisson's equation.
  */
  SAMRAI::pdat_CellVariableX<double> d_source;

  /*!
    \brief Exact solution.
  */
  SAMRAI::pdat_CellVariableX<double> d_exact;

  /*!
    \brief Vector weights.

    For cells not covered by a finer cell, the weight
    is the volume.  For cells that are, the weight is zero.
    This is used in computing norms on the AMR grid.
  */
  SAMRAI::pdat_CellVariableX<double> d_weight;


  /*!
    \brief Saved variable-context index.

    Because we refer to them often, variable-context indices are saved.
    They are initialized in the constructor and never change.
    Each index represents a variable-context pair in this class.
    Thus the indices are \em not independent state variables.
    If we had to set them, we need the associated variable and
    context (and the variable database which manages the mapping).
    See the SAMRAI::hier_VariableDatabaseX class for more into.
  */
  int d_scalar_persistent, d_scalar_scratch, d_diffcoef_persistent
    , d_source_persistent, d_weight_persistent, d_exact_persistent;

  //@}

};


#endif	// included_JacobiPoissonX_h
