/*
  Id:		$Id: jpX.C.sed,v 1.1.2.2 2002/03/20 00:43:45 gunney Exp $
  Copyright:	(c) 1997-2000 The Regents of the University of California
  Release:	$Name: v_0_1_7 $
  Revision:	$Revision: 1.1.2.2 $
  Modified:	$Date: 2002/03/20 00:43:45 $
  Description:	Tutorial program for poisson solver using jacobi iteration
*/

#include "SAMRAI_config.h"

#include <iostream>
#include <string>

#ifdef HAVE_ISO_HEADERS
#include <fstream>
using namespace std;
#else
#include <fstream.h>
#endif

#include <printObjectX.h>
#include "JacobiPoissonX.h"

/*
  Headers for basic SAMRAI objects used in this code.
*/
#include <tbox_SAMRAIManager.h>
#include <tbox_Array.h>
#include <tbox_Boolean.h>
#include <tbox_Database.h>
#include <tbox_InputManager.h>
#include <tbox_MPI.h>
#include <tbox_PIO.h>
#include <tbox_Pointer.h>
#include <tbox_String.h>
#include <tbox_TimerManager.h>
#include <tbox_Utilities.h>
#include <plot_CartesianVizamraiDataWriter2.h>

/*
  Headers for major algorithm/data structure objects from SAMRAI
*/
#include <hier_GridGeometry2.h>
#include <hier_VariableDatabase2.h>
#include <mesh_BergerRigoutsos2.h>
#include <geom_CartesianGridGeometry2.h>
#include <mesh_GriddingAlgorithm2.h>
#include <hier_PatchHierarchy2.h>
#include <mesh_StandardTagAndInitialize2.h>
#include <mesh_UniformLoadBalance2.h>

#ifndef LACKS_NAMESPACE
using namespace SAMRAI;
#endif




/*!
  \brief Tutorial example for solving Poisson's equation using
  Jacobi iterations.

  Learn about interpolation and transfer of data between
  different patches and levels.
*/
int main( int argc, char *argv[] )
{
  /*
    Initialize MPI, SAMRAI, and enable logging.
  */

  tbox_MPI::init(&argc, &argv);
  tbox_SAMRAIManager::startup();

  string input_filename = "input";
  if ( argc > 1 ) input_filename = argv[1];


  /*
    Create input database and parse all data in input file into it.
  */

  tbox_Pointer<tbox_Database> input_db = new tbox_InputDatabase("input_db");
  tbox_InputManager::getManager()->parseInputFile(input_filename, input_db);


  /*
    Get the Main database part of the input database.
    This database contains information relevant to main.
  */

  tbox_Pointer<tbox_Database> main_db = input_db->getDatabase("Main");
  plog << "Main database:" << endl; main_db->printClassData(plog);


  /*
    Base filename info.
  */

  string base_filename = main_db->getStringWithDefault("base_filename", "jp");


  /*
    Log file info.
  */

  string log_filename
    = main_db->getStringWithDefault("log_filename", base_filename+".log");
  tbox_PIO::logOnlyNodeZero(log_filename);


  /*
    Create a patch hierarchy for use later.
    This object is a required input for these objects: jacobi_poisson.
  */
  tbox_Pointer<hier_PatchHierarchy2> patch_hierarchy;
  {
    /*
      Create a grid geometry required for the hier_patchHierarchyX object.
    */
    tbox_Pointer<geom_CartesianGridGeometryX> grid_geometry =
      new geom_CartesianGridGeometryX("CartesianGridGeometry",
				      input_db->getDatabase("CartesianGridGeometry"));
    plog << "Grid Geometry:" << endl;
    grid_geometry->printClassData(plog);
    patch_hierarchy =
      new hier_PatchHierarchy2("Patch Hierarchy", grid_geometry);
  }
  printObjectX( plog, *patch_hierarchy );



  /*
    Create the problem-specific object implementing the required
    SAMRAI virtual functions.
    This object is a required input for these objects: hyp_level_integrator.
  */
  JacobiPoissonX* jacobi_poisson = new JacobiPoissonX(
				 "JacobiPoissonX"
				 , input_db->getDatabase("JacobiPoissonX")
				 );
  printObjectX( plog, *patch_hierarchy );




  /*
    Create the gridding algorithm used to generate the SAMR grid.
  */
  tbox_Pointer<mesh_GriddingAlgorithmX> gridding_algorithm;
  {
    /*
      Create the tag-and-initializer, box-generator and load-balancer
      object references required by the gridding_algorithm object.
    */
    tbox_Pointer<mesh_StandardTagAndInitializeX> tag_and_initializer =
      new mesh_StandardTagAndInitializeX(
          "CellTaggingMethod"
        , jacobi_poisson
        , input_db->getDatabase("StandardTagAndInitialize")
        );
    tbox_Pointer<mesh_BergerRigoutsosX> box_generator =
      new mesh_BergerRigoutsosX();
    tbox_Pointer<mesh_UniformLoadBalanceX> load_balancer =
      new mesh_UniformLoadBalanceX(input_db->getDatabase("UniformLoadBalance"));

    gridding_algorithm =
      new mesh_GriddingAlgorithmX("Gridding Algorithm",
				  input_db->getDatabase("GriddingAlgorithm"),
				  tag_and_initializer,
				  box_generator,
				  load_balancer);
    plog << "Gridding algorithm:" << endl;
    gridding_algorithm->printClassData(plog);
  }



  /*
    Make the patch levels.
  */
  gridding_algorithm->makeCoarsestLevel(patch_hierarchy,0.0);
  bool done=false;
  for (int lnum = 0;
       gridding_algorithm->levelCanBeRefined(lnum) && !done; lnum++) {
    plog << "Adding finner levels with lnum = " << lnum << endl;
    gridding_algorithm->makeFinerLevel( patch_hierarchy
					, /* simulation time */ 0.0
					, /* whether initial time */ true
					, /* tag buffer size */ 0
					);
    plog << "Just added finner levels with lnum = " << lnum << endl;
    done = !(patch_hierarchy->finerLevelExists(lnum));
  }



  /*
    Set up the plotter.
  */
  tbox_Pointer<plot_CartesianVizamraiDataWriter2> viz_data_writer;
  /* Get the output filename. */
  string viz_filename
    = main_db->getStringWithDefault("viz_filename", base_filename);
  int viz_dump_interval
    = main_db->getIntegerWithDefault("viz_dump_interval", 0 );
  if ( viz_dump_interval ) {
    /* Create the vizamrai data writer. */
    viz_data_writer = new plot_CartesianVizamraiDataWriter2("Viz Writer");
    /* Register variables with plotter. */
    jacobi_poisson->registerVariablesWithPlotter(viz_data_writer);
    /*
      Tell the plotter about the refinement ratios.
      This must be done once (and again each time the data changes).
    */
    for ( int ln=1; ln<gridding_algorithm->getMaxLevels(); ln++ ) {
      const hier_IntVectorX &lratio =
	gridding_algorithm->getRatioToCoarserLevel(ln);
      viz_data_writer->setRatioToCoarserLevel(ln, lratio);
    }
    /*
      Set the finest level to plot (optional).
    */
    int finest_plot_level
      = main_db->getIntegerWithDefault( "finest_plot_level"
	, patch_hierarchy->getFinestLevelNumber() );
    viz_data_writer->setFinestLevelToPlot(finest_plot_level);
  }



  /*
    After creating all objects and initializing their state,
    we print the input database and variable database contents
    to the log file.
  */
  plog << "\nCheck input data and variables before simulation:" << endl;
  plog << "Input database..." << endl;
  input_db->printClassData(plog);
  plog << "\nVariable database..." << endl;
  hier_VariableDatabaseX::getDatabase()->printClassData(plog);




  /*
    Loop through a fixed number of iterations.
  */
  pout.setf(ios::scientific);
  int number_of_iterations =
    main_db->getIntegerWithDefault("number_of_iterations",1);
  for ( int itnum=0; itnum<number_of_iterations; ++itnum ) {
    pout << "Iteration " << itnum
	 ;
    if ( viz_dump_interval && !(itnum%viz_dump_interval) ) {
      /* Write the plot file. */
      viz_data_writer->writePlotData( patch_hierarchy , viz_filename , itnum );
    }
    /* Perform iteration. */
    jacobi_poisson->jacobiIteration( patch_hierarchy , 1.0 );
    /* Per iteration output */
    double l2norm, linorm;
    jacobi_poisson->computeError( patch_hierarchy, &l2norm, &linorm );
    pout << " l2norm " << l2norm << " linorm " << linorm << endl;
  }




  /* Write the final plot file. */
  if ( viz_dump_interval )
    viz_data_writer->writePlotData( patch_hierarchy , viz_filename
				  , number_of_iterations );



  /*
    Exit properly by shutting down services in correct order.
  */
  tbox_SAMRAIManager::shutdown();
  tbox_MPI::finalize();


  return(0);
}
