
#include <MultiDimArrayAccess.h>
#include <tbox_PIO.h>
#include <math.h>

/*
  Relax current solution and put result in scratch space.
*/
int jacobiPoissonSmooth(
  double relax_coef
, MultiDimArrayAccess<double,2> c
, MultiDimArrayAccess<double,2> s
, MultiDimArrayAccess<double,2> f
, const int *lower
, const int *upper
, const double *xlo
, const double *xhi
, const double *dx
) {
  int i, j;
  const double a1=dx[0]*dx[0];
  for ( j=lower[1]; j<=upper[1]; ++j ) {
    for ( i=lower[0]; i<=upper[0]; ++i ) {
      c(j,i) = 0.25*( s(j-1,i) + s(j+1,i) + s(j,i-1) + s(j,i+1) - a1*f(j,i) );
    }
  }
  return 0;
}


/*
  Set the boundary condition.
  Set the data in ghost cell to the reverse of the interior data
  to force a Dirichlet boundary condition.
*/
int jacobiPoissonBoundary(
  MultiDimArrayAccess<double,2> d
, const int *lower
, const int *upper
, const double *xlo
, const double *xhi
, const double *dx
) {

  int i, j;

  j = lower[1]-1;
  for ( i=lower[0]; i<=upper[0]; ++i ) {
    d(j,i) = -d(j+1,i);
  }
  j = upper[1]+1;
  for ( i=lower[0]; i<=upper[0]; ++i ) {
    d(j,i) = -d(j-1,i);
  }

  i = lower[0]-1;
  for ( j=lower[1]; j<=upper[1]; ++j ) {
    d(j,i) = -d(j,i+1);
  }
  i = upper[0]+1;
  for ( j=lower[1]; j<=upper[1]; ++j ) {
    d(j,i) = -d(j,i-1);
  }

  return 0;
}




/*
  Set boundary condition to exact solution.
  Which is -source/(8*pi*pi).
*/
int jacobiPoissonExactBC(
  MultiDimArrayAccess<double,2> s
, const int *lower
, const int *upper
, const double *xlo
, const double *xhi
, const double *dx
) {

  int i, j;
  double a1 = -1.0/(8*M_PI*M_PI);
  double x, y, sinx, siny;

  j = lower[1]-1;
  y = xlo[1] - 0.5*dx[1];
  siny = sin(2*M_PI*y);
  for ( i=lower[0]; i<=upper[0]; ++i ) {
    x = xlo[0] + (i-lower[0]+0.5)*dx[0];
    sinx = sin(2*M_PI*x);
    s(j,i) = a1*sinx*siny;
  }

  j = upper[1]+1;
  y = xhi[1] + 0.5*dx[1];
  siny = sin(2*M_PI*y);
  for ( i=lower[0]; i<=upper[0]; ++i ) {
    x = xlo[0] + (i-lower[0]+0.5)*dx[0];
    sinx = sin(2*M_PI*x);
    s(j,i) = a1*sinx*siny;
  }

  i = lower[0]-1;
  x = xlo[0] - 0.5*dx[0];
  sinx = sin(2*M_PI*x);
  for ( j=lower[1]; j<=upper[1]; ++j ) {
    y = xlo[1] + (j-lower[1]+0.5)*dx[1];
    siny = sin(2*M_PI*y);
    s(j,i) = a1*sinx*siny;
  }

  i = upper[0]+1;
  x = xhi[0] + 0.5*dx[0];
  sinx = sin(2*M_PI*x);
  for ( j=lower[1]; j<=upper[1]; ++j ) {
    y = xlo[1] + (j-lower[1]+0.5)*dx[1];
    siny = sin(2*M_PI*y);
    s(j,i) = a1*sinx*siny;
  }

  return 0;
}
