#ifndef include_MultiDimArrayAccess_h
#define include_MultiDimArrayAccess_h

#include <sys/types.h>
#include <iostream.h>


/*!
  \brief Handles index ranges for multidimensional arrays

  The member functions should all be inlined for better
  performance.
*/
template< int DIM >
class MultiDimArrayIndexRange {

  //! Array of starting indices
protected: int d_start[DIM];
  //! Array of sizes
protected: size_t d_size[DIM];

  /*!
    \brief Constructor for setting index data using size and
    initial values.

    Any pointers that are NULL are not used.  The resulting default
    settings are:
    - Array sizes are 0
    - Array starting indices are 0
    - Dimension indices are not reversed

    There is another constructor which accepts the first and final
    indices instead of the sizes and first indices.
    \b NOTE: the place of the initial indices is different than
    it is for the constructor taking final indices instead of sizes.
  */
public: MultiDimArrayIndexRange
( /*! Array sizes		*/ const size_t *sz=((size_t*)0)
, /*! Array starting indices	*/ const int *st=((int*)0)
, /*! Reverse indices flag	*/ bool reverse=false
) {
  int i;
  if ( st )	{ for ( i=0; i<DIM; i++ ) d_start[i] = st[i]; }
  else		{ for ( i=0; i<DIM; i++ ) d_start[i] = 0; }
  if ( sz )	{ for ( i=0; i<DIM; i++ ) d_size[i] = sz[i]; }
  else		{ for ( i=0; i<DIM; i++ ) d_size[i] = 0; }
  if ( reverse ) reverseDim();
}

  /*!
    \brief Constructor for setting index data to a range.

    Any pointers that are NULL are not used.  The resulting default
    settings are:
    - Array sizes are 0
    - Array starting indices are 0
    - Dimension indices are not reversed

    This version takes two \c int* arguments, for the initial
    and final indices.  It does not support default arguments
    until after the indices argument.  \b NOTE: the place of the
    initial indices is different than it is for the constructor
    taking sizes instead of final indices.

    If \c si is \c NULL, starting indices are set to 0.
    If \c sf is \c NULL, sizes are set to zero.
  */
public: MultiDimArrayIndexRange (
  /*! Array of initial indices	*/ const int *si
, /*! Array of final indices	*/ const int *sf
, /*! Reverse indices flag	*/ bool reverse=false
) {
  int i;
  if ( si )	{ for ( i=0; i<DIM; i++ ) d_start[i] = si[i]; }
  else		{ for ( i=0; i<DIM; i++ ) d_start[i] = 0; }
  if ( sf )	{ for ( i=0; i<DIM; i++ ) d_size[i] = 1 + sf[i] - d_start[i]; }
  else		{ for ( i=0; i<DIM; i++ ) d_size[i] = 0; }
  if ( reverse ) reverseDim();
}

  //@{ \name Functions to set indices
  
  /*!
    Set size and starting indices.
  */
public: void setSizeAndStart (
  /*! Array sizes (NULL for no change)		*/ const size_t *sz=((size_t*)0)
, /*! Starting indices (NULL for no change)	*/ const int *st=((int*)0)
, /*! Reverse indices flag			*/ bool reverse=false
) {
  if ( reverse ) reverseDim();
  if (sz) for ( int i=0; i<DIM; i++ ) d_size[i] = sz[i];
  if (st) for ( int i=0; i<DIM; i++ ) d_start[i] = st[i];
  if ( reverse ) reverseDim();
}
  
  /*!
    Set first and final indices (inclusive).
  */
public: void setInclusiveRange (
  /*! First valid indices (NULL for no change)	*/ const int first[DIM]
, /*! Final valid indices (NULL for no change)	*/ const int final[DIM]
, /*! Reverse indices flag	*/ bool reverse=false
) {
  if ( reverse ) reverseDim();
  if (first) for ( int i=0; i<DIM; i++ ) d_start[i] = first[i];
  if (final) for ( int i=0; i<DIM; i++ ) d_size[i] = final[i] - first[i] + 1;
  if ( reverse ) reverseDim();
}

  /*!
    \brief Reverse the ordering of the dimensions.

    This tells the object that the indices you have set
    If you do this, you should also reverse your indices
    to the offset functions.

    Use this to accomodate Fortran ordering.

    Example, the 2x3 arrays declared in C with \c ca[2][3]
    and FORTRAN with \c fa(2,3) have different memory
    arangements.  But because you think of them as 2x3 arrays,
    you pass in {2,3} for their dimensions.  In this case,
    dereferencing the C array with this class works, but
    dereferencing the FORTRAN array will not work.
    To use with FORTRAN-declared arrays, you should
    use reverseDim() to tell the object that the array is
    really a 3x2 array.  Then don't forget to reverse the
    order of the arguments to offset also.
  */
public: void reverseDim() {
  for ( int i=0; i<DIM/2; i++ ) {
    int itmp; size_t stmp;
    itmp = d_start[i];	d_start[i] = d_start[DIM-1-i];	d_start[DIM-1-i] = itmp;
    stmp = d_size[i];	d_size[i] = d_size[DIM-1-i];	d_size[DIM-1-i] = stmp;
  }
}

  /*!
    \brief Adjust the dimensions

    Adjust the first and final indices.
    Set the dimension to adjust to < 0 or >= DIM to adjust \em all dimensions.
    \b Note: The third argument is the increment to the final index
    and not the size.

    \b Note: No error checking is done, for example, to make sure that the
    resulting size is non-negative.

    \return Adjusted MultiDimArrayIndexRange object
  */
public: const MultiDimArrayIndexRange &adjustDim
( /*! Dimension to adjust	*/ int d
, /*! Increment to first index	*/ int first
, /*! Increment to final index	*/ int final
) {
  if ( d < 0 || d >= DIM ) {
    int i;
    for ( i=0; i<DIM; i++ ) {
      d_start[i] += first;
      d_size[i] += final - first;
    }
  }
  else {
    d_start[d] += first;
    d_size[d] += final - first;
  }
  return *this;
}

  //@}



  //@{ \name IO functions
  /*!
    \brief Output to ostream.
  */
  friend ostream &operator<<( ostream &os, const MultiDimArrayIndexRange<DIM> &r ) {
    os << DIM;
    for ( int i=0; i<DIM; i++ )
      os << ' ' << r.d_start[i] << ' ' << r.d_size[i];
    return os;
  }
  /*!
    \brief Input from istream.
  */
  friend istream &operator<<( istream &is, MultiDimArrayIndexRange<DIM> &r ) {
    int dim;
    is >> dim;
    assert(dim == DIM);
    for ( int i=0; i<DIM; i++ )
      is >> r.d_start[i] >> r.d_size[i];
    return is;
  }
  //@}



  //@{ \name Functions to facilitate looping

  //! Give starting index.
public: int beg( /*! index of dimension */ size_t i ) const {
  return d_start[i];
}

  //! Give ending index (one more than the last valid index).
public: int end( /*! index of dimension */ size_t i ) const {
  return d_start[i] + d_size[i];
}

  //! Give size.
public: size_t size( /*! index of dimension */ size_t i ) const {
  return d_size[i];
}

  //@}



  //@{ \name Functions computing offsets

  //! Offset for a 1D array (not for use with any other type arrays).
public: int offset (
  /*! Index of item	*/ int i0
) const {
  return (i0-d_start[0]);
}

  //! Offset for a 2D array (not for use with any other type arrays).
public: int offset (
  /*! First index of item	*/ int i0
, /*! Second index of item	*/ int i1
) const {
  return (i1-d_start[1]) + d_size[1]*(i0-d_start[0]);
}

  //! Offset for a 3D array (not for use with any other type arrays).
public: int offset (
  /*! First index of item	*/ int i0
, /*! Second index of item	*/ int i1
, /*! Third index of item	*/ int i2
) const {
  return (i2-d_start[2]) + d_size[2]*( (i1-d_start[1]) + d_size[1]*(i0-d_start[0]) );
}

  //! Offset for a 4D array (not for use with any other type arrays).
public: int offset (
  /*! First index of item	*/ int i0
, /*! Second index of item	*/ int i1
, /*! Third index of item	*/ int i2
, /*! Fourth index of item	*/ int i3
) const {
  return (i3-d_start[3]) + d_size[3]*( (i2-d_start[2]) + d_size[2]*( (i1-d_start[1]) + d_size[1]*(i0-d_start[0]) ) );
}

  //@}


};






/*!
  \brief Non-const multidimensional array access.

  This class \em never allocates or deallocates data.
  It takes pointers to preallocated data and prodes an
  interface to that data.  Member functions are used
  to give that interface.

  This class provides functions for explicit index checking,
  but it does \em NO implicit error checking on either the
  dimensionality of the array or it size.
  Such should be done through subclassing.

  The member functions should all be inlined for better
  performance.
*/
template < class TYPE, int DIM >
class MultiDimArrayAccess {

  //! Pointer to array.
private: TYPE *d_ptr;
  //! Index range object
private: MultiDimArrayIndexRange<DIM> d_range;

  /*!
    \brief Constructor for setting all data, with default values.

    Any pointers that are NULL are not used.  The resulting default
    settings are:
    - Data pointer is NULL
    - Array sizes are 0
    - Array starting indices are 0
    - Dimension indices are not reversed

    There is another constructor which accepts the first and final
    indices instead of the sizes and first indices.
    \b NOTE: the place of the initial indices is different than
    it is for the constructor taking final indices instead of sizes.
  */
public: MultiDimArrayAccess (
  /*! Pointer to data		*/ TYPE *p=((TYPE*)0)
, /*! Array sizes		*/ const size_t *sz=((size_t*)0)
, /*! Array starting indices	*/ const int *st=((int*)0)
, /*! Reverse indices flag	*/ bool reverse=false
)
: d_ptr(p), d_range(sz,st,reverse) {
}

  /*!
    \brief Constructor for setting all data, with default values.

    Any pointers that are NULL are not used.  The resulting default
    settings are:
    - Data pointer is NULL
    - Array sizes are 0
    - Array starting indices are 0
    - Dimension indices are not reversed

    This version takes two \c int* arguments, for the initial
    and final indices.  It does not support default arguments
    until after the indices argument.  \b NOTE: the place of the
    initial indices is different than it is for the constructor
    taking sizes instead of final indices.

    If \c si is \c NULL, starting indices are set to 0.
    If \c sf is \c NULL, sizes are set to zero.
  */
public: MultiDimArrayAccess (
  /*! Pointer to data		*/ TYPE *p
, /*! Array of initial indices	*/ const int *si
, /*! Array of final indices	*/ const int *sf
, /*! Reverse indices flag	*/ bool reverse=false
)
: d_ptr(p), d_range(si,sf,reverse) {
}

  /*!
    \brief Constructor for specifying pointer and index range object

    There is no default argument in this constructor.
  */
public: MultiDimArrayAccess (
  /*! Pointer to data		*/ TYPE *p
, /*! Array index object	*/ const MultiDimArrayIndexRange<DIM> &r
)
: d_ptr(p), d_range(r) {
}

  /*!
    \brief Copy constructor

    There is no default argument in this constructor.
  */
public: MultiDimArrayAccess (
  /*! Copyee object		*/ const MultiDimArrayAccess &r
)
: d_ptr(r.d_ptr), d_range(r.d_range) {
}

  /*!
    \brief Conversion into boolean.

    \return true iff data pointer is not NULL.
  */
public: operator bool() const {
  return ( d_ptr != (TYPE*)0 );
}

  /*!
    \brief Conversion into pointer.

    \return the data pointer.
  */
public: operator TYPE*() const {
  return d_ptr;
}




  /*!
    \brief Set the data pointer.
  */
public: void setPointer (
  /*! Pointer value	*/ TYPE *p
) {
  d_ptr = p;
}

  /*!
    \brief Adjust the dimensions

    \see MultiDimArrayIndexRange::adjustDim.
  */
public: const MultiDimArrayIndexRange<DIM> &adjustDim (
  /*! Dimension to adjust	*/ int d
, /*! Increment to first index	*/ int first
, /*! Increment to final index	*/ int final
) {
  d_range.adjustDim(d,first,final);
  return d_range;
}

  /*!
    \brief Reverse the ordering of the dimensions.

    \see MultiDimArrayIndexRange::reverseDim.
  */
public: void reverseDim() {
  d_range.reverseDim();
}



  //@{ \name Functions for accessing items

public: const MultiDimArrayIndexRange<DIM> &range() { return d_range; };
public: int beg(size_t i) const { return d_range.beg(i); }
public: int end(size_t i) const { return d_range.end(i); }
public: int size(size_t i) const { return d_range.size(i); }

  //! Grant general access to item in a 1D array.
public: TYPE &operator()
( /*! Index	*/ int i0
) const {
  return *( d_ptr + d_range.offset(i0) );
}

  //! Grant general access to item in a 2D array.
public: TYPE &operator()
( /*! First index	*/ int i0
, /*! Second index	*/ int i1
) const {
  return *( d_ptr + d_range.offset(i0,i1) );
}

  //! Grant general access to item in a 3D array.
public: TYPE &operator()
( /*! First index	*/ int i0
, /*! Second index	*/ int i1
, /*! Third index	*/ int i2
) const {
  return *( d_ptr + d_range.offset(i0,i1,i2) );
}

  //! Grant general access to item in a 4D array.
public: TYPE &operator()
( /*! First index	*/ int i0
, /*! Second index	*/ int i1
, /*! Third index	*/ int i2
, /*! Fourth index	*/ int i3
) const {
  return *( d_ptr + d_range.offset(i0,i1,i2,i3) );
}

  //@}

};	// class MultiDimArrayAccess







/*!
  \brief Const multidimensional array access.

  This class \em never allocates or deallocates data.
  It takes pointers to preallocated data and prodes an
  interface to that data.  Member functions are used
  to give that interface.

  This class provides functions for explicit index checking,
  but it does \em NO implicit error checking on either the
  dimensionality of the array or it size.
  Such should be done through subclassing.

  The member functions should all be inlined for better
  performance.
*/
template < class TYPE, int DIM >
class ConstMultiDimArrayAccess {

  //! Pointer to array.
private: const TYPE *d_ptr;
  //! Index range object
private: MultiDimArrayIndexRange<DIM> d_range;

  /*!
    \brief Constructor for setting all data, with default values.

    Any pointers that are NULL are not used.  The resulting default
    settings are:
    - Data pointer is NULL
    - Array sizes are 0
    - Array starting indices are 0
    - Dimension indices are not reversed

    There is another constructor which accepts the first and final
    indices instead of the sizes and first indices.
    \b NOTE: the place of the initial indices is different than
    it is for the constructor taking final indices instead of sizes.
  */
public: ConstMultiDimArrayAccess (
  /*! Pointer to data		*/ const TYPE *p=((TYPE*)0)
, /*! Array sizes		*/ const size_t *sz=((size_t*)0)
, /*! Array starting indices	*/ const int *st=((int*)0)
, /*! Reverse indices flag	*/ bool reverse=false
)
: d_ptr(p), d_range(sz,st,reverse) {
}

  /*!
    \brief Constructor for setting all data, with default values.

    Any pointers that are NULL are not used.  The resulting default
    settings are:
    - Data pointer is NULL
    - Array sizes are 0
    - Array starting indices are 0
    - Dimension indices are not reversed

    This version takes two \c int* arguments, for the initial
    and final indices.  It does not support default arguments
    until after the indices argument.  \b NOTE: the place of the
    initial indices is different than it is for the constructor
    taking sizes instead of final indices.

    If \c si is \c NULL, starting indices are set to 0.
    If \c sf is \c NULL, sizes are set to zero.
  */
public: ConstMultiDimArrayAccess (
  /*! Pointer to data		*/ const TYPE *p
, /*! Array of initial indices	*/ const int *si
, /*! Array of final indices	*/ const int *sf
, /*! Reverse indices flag	*/ bool reverse=false
)
: d_ptr(p), d_range(si,sf,reverse) {
}

  /*!
    \brief Constructor for specifying pointer and index range object

    There is no default argument in this constructor.
  */
public: ConstMultiDimArrayAccess (
  /*! Pointer to data		*/ const TYPE *p
, /*! Array index object	*/ const MultiDimArrayIndexRange<DIM> &r
)
: d_ptr(p), d_range(r) {
}

  /*!
    \brief Copy constructor

    There is no default argument in this constructor.
  */
public: ConstMultiDimArrayAccess (
  /*! Copyee object		*/ const ConstMultiDimArrayAccess &r
)
: d_ptr(r.d_ptr), d_range(r.d_range) {
}

  /*!
    \brief Conversion into boolean.

    \return true iff data pointer is not NULL.
  */
public: operator bool() const {
  return ( d_ptr != (TYPE*)0 );
}

  /*!
    \brief Conversion into pointer.

    \return the data pointer.
  */
public: operator const TYPE*() const {
  return d_ptr;
}

  /*!
    \brief Constructor referencing non-const array.

    Attempt to type-cast the argument to TYPE.  If the type-cast fails,
    then the data pointer is set to NULL.
  */
public: ConstMultiDimArrayAccess (
  /*! Copyee object		*/ const MultiDimArrayAccess<TYPE,DIM> &r
)
: d_ptr((TYPE*)0), d_range() {
  if ( r ) {
    d_ptr = r;
    int beg[DIM]; size_t size[DIM];
    for ( int i=0; i<DIM; ++i ) {
      beg[i] = r.beg(i);
      size[i] = r.size(i);
    }
    d_range.setSizeAndStart(size,beg);
  }
}


  /*!
    Set the data pointer.
  */
public: void setPointer (
  /*! Pointer value	*/ const TYPE *p
) {
  d_ptr = p;
}

  /*!
    \brief Adjust the dimensions

    @see MultiDimArrayIndexRange::adjustDim
  */
public: const MultiDimArrayIndexRange<DIM> &adjustDim (
  /*! Dimension to adjust	*/ int d
, /*! Increment to first index	*/ int first
, /*! Increment to final index	*/ int final
) {
  d_range.adjustDim(d,first,final);
  return d_range;
}

  /*!
    \brief Reverse the ordering of the dimensions.

    \see MultiDimArrayIndexRange::reverseDim.
  */
public: void reverseDim() {
  d_range.reverseDim();
}



  //@{ \name Functions for accessing items

public: const MultiDimArrayIndexRange<DIM> &range() { return d_range; };
public: int beg(size_t i) const { return d_range.beg(i); }
public: int end(size_t i) const { return d_range.end(i); }
public: int size(size_t i) const { return d_range.size(i); }

  //! Grant general access to item in a 1D array.
public: const TYPE &operator()
( /*! Index	*/ int i0
) const {
  return *( d_ptr + d_range.offset(i0) );
}

  //! Grant general access to item in a 2D array.
public: const TYPE &operator()
( /*! First index	*/ int i0
, /*! Second index	*/ int i1
) const {
  return *( d_ptr + d_range.offset(i0,i1) );
}

  //! Grant general access to item in a 3D array.
public: const TYPE &operator()
( /*! First index	*/ int i0
, /*! Second index	*/ int i1
, /*! Third index	*/ int i2
) const {
  return *( d_ptr + d_range.offset(i0,i1,i2) );
}

  //! Grant general access to item in a 4D array.
public: const TYPE &operator()
( /*! First index	*/ int i0
, /*! Second index	*/ int i1
, /*! Third index	*/ int i2
, /*! Fourth index	*/ int i3
) const {
  return *( d_ptr + d_range.offset(i0,i1,i2,i3) );
}

  //@}

};	// class ConstMultiDimArrayAccess





#endif	// include_MultiDimArrayAccess_h
